/* By Antti-Pekka.Tuovinen@cs.helsinki.fi

   Two additional methods to class Geom for computing
   points on the perimeter of a diamond.
*/

package CH.ifa.draw.contrib;

import CH.ifa.draw.util.*;
import java.awt.Rectangle;
import java.awt.Point;


public class DiamondGeom {

    private DiamondGeom() {}; // never instantiated

    /**
     * Converts a polar to a point on a diamond's perimeter.
     */
    static public Point polarToPointOnDiamond(double angle, double fx, double fy) {
        double si = Math.sin(angle);
        double co = Math.cos(angle);
	Point ref = new Point((int)((fx+fy)*co+0.5),(int)((fx+fy)*si+0.5));
	if (ref.x > 0)
		if (ref.y >0)
	   		return Geom.intersect(0,0,ref.x,ref.y,(int)fx,0,0,(int)fy);
		else
	   		return Geom.intersect(0,0,ref.x,ref.y,(int)fx,0,0,-(int)fy);
	else
		if (ref.y >0)
	   		return Geom.intersect(0,0,ref.x,ref.y,-(int)fx,0,0,(int)fy);
		else
	   		return Geom.intersect(0,0,ref.x,ref.y,-(int)fx,0,0,-(int)fy);
    }

    /**
     * Gets the point on a diamond that corresponds to the given angle.
     */
    static public Point diamondAngleToPoint(Rectangle r, double angle) {
        Point center = Geom.center(r);
        Point p = DiamondGeom.polarToPointOnDiamond(angle, r.width/2,r.height/2);
        return new Point(center.x + p.x, center.y + p.y);
    }

    static public Point polarToPointOnUpTriangle(double angle, double fx, double fy) {
        double si = Math.sin(angle);
        double co = Math.cos(angle);
		Point ref = new Point((int)((fx+fy)*co+0.5),(int)((fx+fy)*si+0.5));
		Point intersection = null;
		intersection = 
			Geom.intersect(0,0,ref.x,ref.y,(int)fx,(int)fy,0,-(int)fy);
		if (intersection == null)
			intersection = 
				Geom.intersect(0,0,ref.x,ref.y,-(int)fx,(int)fy,0,-(int)fy);
		if (intersection == null)
			intersection = 
				Geom.intersect(0,0,ref.x,ref.y,-(int)fx,(int)fy,(int)fx,(int)fy);
		if (intersection == null)
			intersection = new Point(0,0);
		return intersection;
    }

    static public Point polarToPointOnDownTriangle(double angle, double fx, double fy) {
        double si = Math.sin(angle);
        double co = Math.cos(angle);
		Point ref = new Point((int)((fx+fy)*co+0.5),(int)((fx+fy)*si+0.5));
		Point intersection = null;
		intersection = 
			Geom.intersect(0,0,ref.x,ref.y,(int)fx,-(int)fy,0,(int)fy);
		if (intersection == null)
			intersection = 
				Geom.intersect(0,0,ref.x,ref.y,-(int)fx,-(int)fy,0,(int)fy);
		if (intersection == null)
			intersection = 
				Geom.intersect(0,0,ref.x,ref.y,-(int)fx,-(int)fy,(int)fx,-(int)fy);
		if (intersection == null)
			intersection = new Point(0,0);
		return intersection;
    }

    static public Point upTriangleAngleToPoint(Rectangle r, double angle) {
    	//System.err.println(angle);
        Point center = Geom.center(r);
        Point p = DiamondGeom.polarToPointOnUpTriangle(angle, r.width/2,r.height/2);
        //System.err.println(center);
        //System.err.println(p);
        return new Point(center.x + p.x, center.y + p.y);
    }

    static public Point downTriangleAngleToPoint(Rectangle r, double angle) {
    	//System.err.println(angle);
        Point center = Geom.center(r);
        Point p = DiamondGeom.polarToPointOnDownTriangle(angle, r.width/2,r.height/2);
        //System.err.println(center);
        //System.err.println(p);
        return new Point(center.x + p.x, center.y + p.y);
    }
}