package CH.ifa.draw.contrib.xcomp;

import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import java.awt.*;
import java.util.*;
import java.io.*;

/**
 * The basic implementation for list-like layout strategies.
 * Like BagLayout this layout manager positions the contained
 * figures 'as they were'. That is, no positioning is performed.
 * This is an example implementation that shows how the mechanisms
 * of super classes are used to create specialized layout managers.
 */ 

public class BagFlowLayout extends AbstractFlowLayout {

	public BagFlowLayout() {
		super();
	}
	
	/**
   * Returns a NullAligner.
   */

  protected Aligner getTopElementAligner(CompositeFigure container,Figure component) {
  	//return new RelativeAligner(container,component);
  	return new NullAligner();
  }


  /**
   * Returns a NullAligner.
   */

  protected Aligner getMiddleElementAligner(CompositeFigure container,Figure component,Figure previous){
  	//return new RelativeAligner(container,component);
  	return new NullAligner();
  }


  /**
   * Returns a NullAligner.
   */

  protected Aligner getBottomElementAligner(CompositeFigure container,Figure component,Figure previous){
  	//return new RelativeAligner(container,component);
  	return new NullAligner();
  }

		/**
	 * Force rearrangement. Moves the component 'in'.
	 */
	public void invalidateLayout(Figure component) {
		//System.err.println("bag flow layout invalidated with component");
		moveIn((CompositeFigure)component.parent(),component);
		invalidateLayout();	
	}

  public Dimension preferredLayoutSize(CompositeFigure container) {
  	Rectangle r = preferredLayoutRectangle(container);
    return new Dimension(r.width,r.height);
  }
  
  public Rectangle preferredLayoutRectangle(CompositeFigure container) {
  	//System.err.println("------------------------------");
    //Vector v = getArrengement(container);
    Rectangle cBox = container.displayBox();
    //System.err.println("v: "+v);
    Vector components = getStationaryFigures();
    if (components.isEmpty()) {
    	//System.err.println(">>>> pref size empty");
    	//System.err.println("------------------------------");
    	return minimumLayoutRectangle(container);
    }
    Insets insets = container.getInsets(); 
    Point origin = container.displayBox().getLocation();
    origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = new Rectangle(origin);
    //System.err.println("r: "+r);
    for (int i = 0;i<components.size();i++) {
    	//Point pos = ((Aligner)v.elementAt(i)).align((Figure)components.elementAt(i));
    	Point pos = ((Figure)components.elementAt(i)).displayBox().getLocation();
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getPreferredSize());
      //System.err.println(">>>> box: "+box);
      //if (i==0)
      //	r = box;
      //else
   			r.add(//box);
   				new Rectangle(origin.x,origin.y,pos.x+box.width-origin.x,pos.y+box.height-origin.y)
   			);
      //System.err.println(">>>> r: "+r);
    }
    r.setSize(
    	r.width+insets.left+insets.right,
    	r.height+insets.top+insets.bottom
    );
    r.translate(-insets.left,-insets.top);
    //System.err.println(">>>> >>>> pref size "+r);
    //System.err.println("------------------------------");
    return r;
    //r.width = r.width > cBox.width ? r.width : cBox.width;
    //r.height = r.height > cBox.height ? r.height : cBox.height;
    //System.err.println(">>>> pref size "+r);
    //return new Dimension(r.width,r.height);
  }

  public Dimension minimumLayoutSize(CompositeFigure container) {
  	Rectangle r = minimumLayoutRectangle(container);
    return new Dimension(r.width,r.height);
  }
    
  /** 
   * For bag layout, the minimum display box is the box that contains
   * all components at their preferred sizes with the insets of the
   * container.
   */ 
  public Rectangle minimumLayoutRectangle(CompositeFigure container) {
    //Vector v = getArrengement(container);
    //System.err.println("v: "+v);
    Vector components = getStationaryFigures();
    if (components.isEmpty()) {
    	//System.err.println(">>>> min size empty");
    	return new Rectangle (container.displayBox().getLocation(),emptySize());
    	//return new Dimension(container.displayBox().width,container.displayBox().height);
    }
    Insets insets = container.getInsets(); 
    //Point origin = container.displayBox().getLocation();
    //origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = null; // new Rectangle(origin);
      //System.err.println("r: "+r);
    for (int i = 0;i<components.size();i++) {
    	//Point pos = ((Aligner)v.elementAt(i)).align((Figure)components.elementAt(i));
      Point pos = ((Figure)components.elementAt(i)).displayBox().getLocation();
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getPreferredSize()); //!!
      //System.err.println(">>>> box: "+box);
      if (i==0)
      	r = box;
      else
   			r.add(box);
   			//	new Rectangle(origin.x,origin.y,pos.x+box.width-origin.x,pos.y+box.height-origin.y)
   			//);
      //System.err.println(">>>> r: "+r);
    }
    r.setSize(r.width+insets.left+insets.right,r.height+insets.top+insets.bottom);
    r.translate(-insets.left,-insets.top);
    //System.err.println(">>>> >>>> minimum size "+r);
    return r;
  }

	/**
	 * If the container is empty, positions the minimum size rectangle
	 * flush to the edge or corner furthest away from the manipulator.
	 */
	public void overlayMinimumRectangle(CompositeFigure container,Rectangle minimum) {
			if (container.figureCount()> 0)
				return; // positioned absolutely already
			Handle manip = container.getManipulator();
			if (manip == null) {
				
			} else if (manip instanceof SouthHandle) {
				
			} else if (manip instanceof SouthWestHandle) {
				minimum.translate(container.displayBox().width - minimum.width,0);
			} else if (manip instanceof WestHandle) {
				minimum.translate(container.displayBox().width - minimum.width,0);
			} else if (manip instanceof NorthWestHandle) {
				minimum.translate(
					container.displayBox().width - minimum.width,
					container.displayBox().height - minimum.height
				);
			} else if (manip instanceof NorthHandle) {
				minimum.translate(0,container.displayBox().height - minimum.height);
			} else if (manip instanceof NorthEastHandle) {
				minimum.translate(0,container.displayBox().height - minimum.height);
			} else if (manip instanceof EastHandle) {
			
			} else if (manip instanceof SouthEastHandle) {
			
			} 		
	}


    
    public void layoutContainer(CompositeFigure container) {
      //System.err.println("<<-:-*-:->> lay out "+container);
			//Vector v = getArrengement(container);
      Vector holdees = getStationaryFigures();
      for (int i=0;i<holdees.size();i++) {
     		Figure fig = (Figure)holdees.elementAt(i);
     		//System.err.println("====>fig: "+fig);
        //Point newPos = ((Aligner)v.elementAt(i)).align(fig);
        Point newPos = fig.displayBox().getLocation();
        //System.err.println("====>newPos:"+newPos);
        Point oldPos = fig.displayBox().getLocation();
        //System.err.println("====>oldPos:"+oldPos);
        Rectangle newBox = new Rectangle(newPos,fig.getPreferredSize());
        if (!newBox.equals(fig.displayBox())) {
        	fig.invalidate();
        	if (!oldPos.equals(newPos))
        		fig.basicMoveBy(newPos.x-oldPos.x,newPos.y-oldPos.y);
        	fig.basicDisplayBox(
        		newPos,
        		new Point(newPos.x+newBox.width,newPos.y+newBox.height)
        	);
        	fig.markDirty(); // should markDirty() mark subfigures dirty too?
        	//fig.invalidate();
        }
    		//grow(container,fig.displayBox()); // !!!!! is this really needed
    	}
    	layoutSovereigns(container);
    }
}