package CH.ifa.draw.contrib.xcomp;

import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import java.awt.*;
import java.util.*;
import java.io.*;
import com.objectspace.jgl.algorithms.Sorting;
import com.objectspace.jgl.*;
import com.objectspace.jgl.adapters.VectorArray;
import com.objectspace.jgl.predicates.BinaryNot;

/**
 * The basic implementation for list-like layout strategies.
 * Like BagLayout this layout manager positions the contained
 * figures 'as they were'. That is, no positioning is performed.
 * This is an example implementation that shows how the mechanisms
 * of super classes are used to create specialized layout managers.
 */ 

public class CenteredVListLayout extends VListLayout {

	public CenteredVListLayout() {
		super();
	}
	
	public CenteredVListLayout(int hgap) {
		super(hgap);
	}
	
  public Rectangle preferredLayoutRectangle(CompositeFigure container) {
  	//System.err.println("------------------------------");
    Vector components = getStationaryFigures();    
    Rectangle cBox = container.displayBox();
    //System.err.println("v: "+v);
    if (components.isEmpty()) {
    	//System.err.println(">>>> pref size empty");
    	//System.err.println("------------------------------");
    	return minimumLayoutRectangle(container);
    }
    Insets insets = container.getInsets(); 
    Point origin = cBox.getLocation();
    origin.translate(cBox.width/2,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = null;      //System.err.println("r: "+r);
    int lastY = origin.y; 
    for (int i = 0;i<components.size();i++) {
    	if (i>0) 
    		lastY += hgap;
    	Point pos = new Point(origin.x,lastY);
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getPreferredSize());
     	box.translate(-box.width/2,0);
     	//    System.err.println(">>>> box: "+box);
      if (i==0)
      	r = box;
      else
   			r.add(box
   				//new Rectangle(origin.x,origin.y,pos.x+box.width-origin.x,pos.y+box.height-origin.y)
   			);
   	  lastY = r.y + r.height;
      //System.err.println(">>>> r: "+r);
    }
    r.setSize(
    	r.width+insets.left+insets.right,
    	r.height+insets.top+insets.bottom
    );
    r.translate(-(cBox.width/2-r.width/2)-insets.left,-insets.top);
    //System.err.println(">>>> >>>> pref size "+r);
    //System.err.println("------------------------------");
    return r;
    //r.width = r.width > cBox.width ? r.width : cBox.width;
    //r.height = r.height > cBox.height ? r.height : cBox.height;
    //System.err.println(">>>> pref size "+r);
    //return new Dimension(r.width,r.height);
  }

  public Rectangle minimumLayoutRectangle(CompositeFigure container) {
    Vector components = getStationaryFigures();      
    Rectangle cBox = container.displayBox();
    //System.err.println("v: "+v);
    if (components.isEmpty()) {
     	//System.err.println(">>>> min size empty");
    	return new Rectangle (container.displayBox().getLocation(),emptySize());
    	//return new Dimension(container.displayBox().width,container.displayBox().height);
    }
    Insets insets = container.getInsets(); 
    Point origin = cBox.getLocation();
    origin.translate(cBox.width/2,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = null;      //System.err.println("r: "+r);
    int lastY = origin.y; 
    for (int i = 0;i<components.size();i++) {
    	if (i>0) 
    		lastY += hgap;
   		Point pos = new Point(origin.x,lastY);
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getMinimumSize());
     	box.translate(-box.width/2,0);
     	//    System.err.println(">>>> box: "+box);
      if (i==0)
      	r = box;
      else
   			r.add(box
   				//new Rectangle(origin.x,origin.y,pos.x+box.width-origin.x,pos.y+box.height-origin.y)
   			);
   	  lastY = r.y + r.height;
      //System.err.println(">>>> r: "+r);
    }
    r.setSize(
    	r.width+insets.left+insets.right,
    	r.height+insets.top+insets.bottom
    );
    r.translate(-(cBox.width/2-r.width/2)-insets.left,-insets.top);
    //System.err.println(">>>> >>>> pref size "+r);
    //System.err.println("------------------------------");
    return r;
/*    Insets insets = container.getInsets(); 
    //Point origin = container.displayBox().getLocation();
    //origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = null; // new Rectangle(origin);
      //System.err.println("r: "+r);
    Point origin = container.displayBox().getLocation();
    origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    int lastX = origin.x;
    int lastY = origin.y; 
    for (int i = 0;i<components.size();i++) {
    	if (i>0) 
    		lastY += insets.top;
    	Point pos = new Point(lastX,lastY);
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getMinimumSize());
      System.err.println(">>>> box: "+box);
      if (i==0)
      	r = box;
      else
   			r.add(box);
   			//	new Rectangle(origin.x,origin.y,pos.x+box.width-origin.x,pos.y+box.height-origin.y)
   			//);
   		lastX = r.x;
   		lastY = r.y + r.height;
      System.err.println(">>>> r: "+r);
    }
    r.setSize(r.width+insets.left+insets.right,r.height+insets.top+insets.bottom);
    r.translate(-insets.left,-insets.top);
    System.err.println(">>>> >>>> minimum size "+r);
    return r;*/
  }

  public void layoutContainer(CompositeFigure container) {
    Vector components = getStationaryFigures();    
    Rectangle cBox = container.displayBox();
    //System.err.println("v: "+v);
    Insets insets = container.getInsets(); 
    Point origin = cBox.getLocation();
    origin.translate(cBox.width/2,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = null;      //System.err.println("r: "+r);
    int lastY = origin.y; 
    for (int i = 0;i<components.size();i++) {
    	Figure fig = (Figure)components.elementAt(i);
    	if (i>0) 
    		lastY += hgap;
      Point oldPos = fig.displayBox().getLocation();
   		Point pos = new Point(origin.x,lastY);
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getPreferredSize());
     	box.translate(-box.width/2,0);
     	pos.x -= box.width/2;
     	//    System.err.println(">>>> box: "+box);
/*    System.err.println("<<-:-*-:->> lay out "+container);
		//Vector v = getArrengement(container);
    Vector holdees = getStationaryFigures();
    Insets insets = container.getInsets();
    Point origin = container.displayBox().getLocation();
    origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    int lastX = origin.x;
    int lastY = origin.y; 
    for (int i=0;i<holdees.size();i++) {
     		Figure fig = (Figure)holdees.elementAt(i);
     		System.err.println("====>fig: "+fig);
    		if (i>0) 
    			lastY += insets.top;
        Point newPos = new Point(lastX,lastY);
        System.err.println("====>newPos:"+newPos);
        Point oldPos = fig.displayBox().getLocation();
        System.err.println("====>oldPos:"+oldPos);
        Rectangle newBox = new Rectangle(newPos,fig.getPreferredSize());*/
        if (!box.equals(fig.displayBox())) {
        	fig.invalidate();
        	if (!oldPos.equals(pos))
        		fig.basicMoveBy(pos.x-oldPos.x,pos.y-oldPos.y);
        	fig.basicDisplayBox(
        		pos,
        		new Point(pos.x+box.width,pos.y+box.height)
        	);
        	fig.markDirty(); // should markDirty() mark subfigures dirty too?
        	//fig.invalidate();
        }
        lastY += fig.displayBox().height;
    		//grow(container,fig.displayBox()); // !!!!! is this really needed
    	}
    }
  
	/**
	 * Positions the minimum size rectangle
	 * at the center of the container.
	 */
	/*public void overlayMinimumRectangle(CompositeFigure container,Rectangle minimum) {
		Rectangle cBox = container.displayBox();
		minimum.translate(
			cBox.x + cBox.width/2 - minimum.x+minimum.width/2,
			cBox.x + cBox.height/2 - minimum.x+minimum.height/2
		);
	}*/

}