package CH.ifa.draw.contrib.xcomp;

/*
 * @(#)FigurePanel.java
 */

import CH.ifa.draw.standard.*;
import CH.ifa.draw.framework.*;
import java.awt.*;
import java.util.*;
import com.objectspace.jgl.Pair;
import com.objectspace.jgl.HashMap;
import java.io.*;
import CH.ifa.draw.util.*;
import java.awt.event.*;


	
public abstract class FigurePanel extends ManagedComposite {

    

  /**
   * The display box.
   */
  protected Rectangle fDisplayBox = null;
  
    protected boolean fHandleChangeEvents = true;
    
    protected Vector fFrozenBy;
    
    public FigurePanel() {
      this(new Point(0,0),new Point(0,0));
    }

    public FigurePanel(Point origin, Point corner) {
      super();
      if (origin != null && corner != null)
        fDisplayBox = new Rectangle(origin.x,origin.y,corner.x,corner.y);
      fFrozenBy = new Vector();
    }
    
    public void freezeWith(Figure child) {
    	//System.err.println("- "+this+" frozen with "+child);
    	fFrozenBy.addElement(child);
    }
    
    public void unFreezeWith(Figure child) {
    	//System.err.println("+ "+this+" unfrozen with "+child);
    	fFrozenBy.removeElement(child);    
    }

		public boolean isFrozen() {
			return fFrozenBy.size() > 0;
		}
  /**
   * Freezes this panel.
   */
  public void beginDrag(Figure child,MouseEvent e) {
  	if (e != null && (e instanceof java.awt.event.MouseEvent) )
  		freezeWith(child);
  }

  /**
   * Unfreezes this panel.
   */
  public void endDrag(Figure child,MouseEvent e) {
  	if (e != null && (e instanceof java.awt.event.MouseEvent) )
  		unFreezeWith(child);
  }

	/**
   * Highlights this.
   */
   public void setHighlighted(boolean isHighlighted) {
    	if (isHighlighted != fIsHighlighted) {
        //System.err.println("setHighlight = "+isHighlighted+" "+this);
    		fIsHighlighted = isHighlighted;
        invalidate();
    		//changed();
    	}
    }

    /**
     * Draw this in a special way. draw a red rectangle slightly smaller than
     * the display box.
     */
    protected void highlight(Graphics g) {
      Rectangle r = new Rectangle(displayBox());
		  r.grow(-2,-2);
		  g.setColor(Color.red);
		  g.drawRect(r.x,r.y,r.width-1,r.height-1);
    }


    /**
     * Moves all the given figures by x and y. Doesn't announce
     * any changes. Moves first the root figure and then the subfigures.
     */
    public void basicMoveBy(int x, int y) {
    	fDisplayBox.translate(x,y);
      FigureEnumeration k = figures();
      while (k.hasMoreElements())
        k.nextFigure().moveBy(x,y);
    }

    public void basicDisplayBox(Point origin, Point corner) {
        fDisplayBox = new Rectangle(origin);
        fDisplayBox.add(corner);
     	//System.err.println(" basicDisplayBox set");
    }

    public Vector handles() {
        Vector handles = new Vector();
        BoxHandleKit.addHandles(this, handles);
        return handles;
    }

    public Rectangle displayBox() {
        return new Rectangle(
            fDisplayBox.x,
            fDisplayBox.y,
            fDisplayBox.width,
            fDisplayBox.height);
    }


		private void overlayMinimumRectangle(Rectangle minimum) {
			if (figureCount()> 0)
				return; // positioned absolutely already
			Handle manip = getManipulator();
			if (manip == null) {
				
			} else if (manip instanceof SouthHandle) {
				
			} else if (manip instanceof SouthWestHandle) {
				minimum.translate(displayBox().width - minimum.width,0);
			} else if (manip instanceof WestHandle) {
				minimum.translate(displayBox().width - minimum.width,0);
			} else if (manip instanceof NorthWestHandle) {
				minimum.translate(
					displayBox().width - minimum.width,
					displayBox().height - minimum.height
				);
			} else if (manip instanceof NorthHandle) {
				minimum.translate(0,displayBox().height - minimum.height);
			} else if (manip instanceof NorthEastHandle) {
				minimum.translate(0,displayBox().height - minimum.height);
			} else if (manip instanceof EastHandle) {
			
			} else if (manip instanceof SouthEastHandle) {
			
			} 
		}

    /**
     * Constrain the box defined by its origin and (left-down) corner
     * by the minimum size dimensions. Returns a pair of points giving
     * the <em>absolute</em> origin and corner of the constrained box.
     * Takes into account the possible change of the origin to facilitate
     * resizing from any corner of the display box.
     */
    protected Pair constrainDisplayBox(Rectangle minimum,Point origin, Point corner) {
      
    	Rectangle oldBox = displayBox();
      //System.err.println("** ?? ** ?? constrainDisplayBox");
      //System.err.println("minimum:  "+minimum);
      getLayout().overlayMinimumRectangle(this,minimum);
      //System.err.println("minimum': "+minimum);
      //System.err.println("origin: "+origin);
      //System.err.println("corner: "+corner);
      //System.err.println("oldBox: "+oldBox);
      Point deltaOrigin,deltaCorner;
      Point minimumCorner = 
      	new Point(minimum.x+minimum.width,minimum.y+minimum.height);
      Point minimumOrigin = minimum.getLocation();
      Handle manip = getManipulator();
      if (manip == null || (manip instanceof SouthEastHandle) ||
          (manip instanceof SouthHandle) ||
          (manip instanceof EastHandle)) {
        // new corner, old origin stays where it was
        // manip == null implies that this is a change due to a change
        // in some subfigure
        deltaOrigin = oldBox.getLocation();
        deltaCorner =
          new Point(
            Math.max(corner.x,minimumCorner.x),
            Math.max(corner.y,minimumCorner.y)
          );
      } else if (manip instanceof NorthWestHandle ||
                 manip instanceof WestHandle ||
                 manip instanceof NorthHandle) {
        //new origin, corner stays where it was
        deltaCorner =
        	//new Point(corner);
          new Point(
            Math.max(corner.x,minimumCorner.x),
            Math.max(corner.y,minimumCorner.y)
          );
        deltaOrigin =
         new Point(
            Math.min(origin.x,minimumOrigin.x),
            Math.min(origin.y,minimumOrigin.y)
          );
       } else if (manip instanceof SouthWestHandle) {
        deltaCorner =
          // the y of the corner and the x of the origin may change
          new Point(
            corner.x,
            Math.max(corner.y,minimumCorner.y)
          );
        deltaOrigin =
         new Point(
            Math.min(origin.x,minimumOrigin.x),
            origin.y
          );

       }  else if (manip instanceof NorthEastHandle) {
       // the x of the corner and the y of the origin may change
        deltaCorner =
          new Point(
            Math.max(corner.x,minimumCorner.x),
            corner.y
          );
        deltaOrigin =
         new Point(
            origin.x,
            Math.min(origin.y,minimumOrigin.y)
          );

       } else {
        //System.err.println("FrameComposite#constrainDisplayBox(): unknown manipulator");
        deltaOrigin = oldBox.getLocation();
        deltaCorner =
          new Point(
            oldBox.getLocation().x+oldBox.width,
            oldBox.getLocation().y+oldBox.height
          );
      }
      //System.err.println("deltaOrigin: "+deltaOrigin);
      //System.err.println("deltaCorner: "+deltaCorner);
      Pair result = new Pair(deltaOrigin,deltaCorner);
      return result;
    }

	/**
	 * Returns the minimum size of this composite.
	 */
	public Dimension getMinimumSize() {
		if (!isValid() || fMinSize == null)
			fMinSize = getLayout().minimumLayoutSize(this);
		//Rectangle dBox = displayBox();
		//dim.width = dim.width > dBox.width ? dim.width : dBox.width;
    //dim.height = dim.height > dBox.height ? dim.height : dBox.height;
		//System.err.println("== == "+dim);
		return fMinSize;
	}
		
	/**
	 * Returns the preferred size of this composite. Because the size
	 * of the frame is adjustable by handles, the method extends the
	 * dimensions returned by the layout manager to the width/height of the
	 * current display box if they exceed the dimensions. The effect is that
	 * changes in the sizes of the contained figures can only make the
	 * preferred size larger, not smaller. Redefine this method in sublcasses
	 * if different behavior is wanted.
	 */
	public Dimension getPreferredSize() {
		if (!isValid() || fPrefSize == null) {
			Dimension dim;
			if (figureCount()>0)
				dim = getLayout().preferredLayoutSize(this);
			else
				dim = getMinimumSize();
			if (getManipulator() != null || isFrozen()) { 
				Rectangle dBox = displayBox();
				dim.width = dim.width > dBox.width ? dim.width : dBox.width;
    		dim.height = dim.height > dBox.height ? dim.height : dBox.height;
			}
			fPrefSize = dim;
		}
		//System.err.println("^^ ^^ "+dim);
		return fPrefSize;
	}
    


	/**
	 * Returns the minimum size of this composite.
	 */
	/*public Dimension getMinimumSize() {
		Dimension dim = getLayout().minimumLayoutSize(this);
		//Rectangle dBox = displayBox();
		//dim.width = dim.width > dBox.width ? dim.width : dBox.width;
    //dim.height = dim.height > dBox.height ? dim.height : dBox.height;
		//System.err.println("== == == == >"+dim);
		return dim;
	}*/
		
	/**
	 * Returns the preferred size of this composite. Because the size
	 * of the frame is adjustable by handles, the method extends the
	 * dimensions returned by the layout manager to the width/height of the
	 * current display box if they exceed the dimensions. The effect is that
	 * changes in the sizes of the contained figures can only make the
	 * preferred size larger, not smaller. Redefine this method in sublcasses
	 * if different behavior is wanted.
	 */
	/*public Dimension getPreferredSize() {
		Dimension dim = getLayout().preferredLayoutSize(this);
		//Rectangle dBox = displayBox();
		//dim.width = dim.width > dBox.width ? dim.width : dBox.width;
    //dim.height = dim.height > dBox.height ? dim.height : dBox.height;
		//System.err.println("^^ ^^ ^^ ^^ >"+dim);
		return dim;
	}*/
    /**
     * Informs that a figure changed the area of its display box.
     * Suppress propagation of change if fHandleChangeEvents is
     * set to <pre>false</pre>.
     *
     * @see FigureChangeEvent
     * @see Figure#changed
     */
    public void changed() {
    	invalidate();
    	if (fHandleChangeEvents) { 
    		// make sure this is not being laid out!
        // would cause a forever validation loop otherwise	
        	if (fListener != null) {
        		//System.err.println("announcing change of "+this.toString());
            fListener.figureChanged(new FigureChangeEvent(this));
        	}
    	}
    }


    /**
     * Changes the display box of a composite figure. Clients usually
     * call this method. It changes the display box
     * and announces the corresponding change. Marks the layout
     * invalid (dirty).
     * @param origin the new origin
     * @param corner the new corner
     * @see displayBox
     */
    public void displayBox(Point origin, Point corner) {
      Pair points;
      //if (figureCount() == 0)  
      	//constrainDisplayBox(Convert.toPoint(getPreferredSize()),origin,corner);
      	points = constrainDisplayBox(
      		getLayout().minimumLayoutRectangle(this),origin,corner
      	);
      //else
      //	points = constrainDisplayBox(
      //		getLayout().preferredLayoutRectangle(this),origin,corner
      //	);

      super.displayBox((Point)points.first,(Point)points.second);
   }


		/**
		 * Returns insets of (3,3,3,3). 
		 */
	public Insets getInsets() {
		return new Insets(3,3,3,3);
	}


    /**
     * Default is to accept all figures. Override this in subclasses
     * to selectively accept subfigures.
     */
    public boolean accept(Figure figure) {
      return true;
    }

    /**
     * Add the dragged figure to this.
     */
     public Figure drop(Figure dragTarget) {
        if (accept(dragTarget))
          ParentChangeEvent.changeParent(dragTarget,this);
        return dragTarget;
        //return add(dragTarget);
     }

     /**
      * Traverse the hieararchy of contained figures to find the subfigure
      * at x,y that accepts the figure to be dropped on it. The protocol does not
      * assume
      * that whoever calls findDropTarget tests first whether the called figure
      * contains the point and is different from dragTarget.
      */
     public Figure findDropTarget(int x, int y, Figure dragTarget) {
        FigureEnumeration k = figuresReverse();
        if (dragTarget != this) {
          while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            Figure found = figure.findDropTarget(x, y, dragTarget);
            if (found != null)
              return found;
          }
          if (containsPoint(x,y) && accept(dragTarget))
            return this;
        }
        return null;
     }



    /**
     * Returns the Figures connector for the specified location.
     * Delegate to the root figure.
     */
    /*public Connector connectorAt(int x, int y) {
        return fRootFigure.connectorAt(x,y);
    }*/

    /**
     * Returns the named attribute or null if a
     * a figure doesn't have an attribute.
     * Delegate to the root figure.
     */
    /*public Object getAttribute(String name) {
        return fRootFigure.getAttribute(name);
    }*/

    /**
     * Sets the named attribute to the new value. Delegate to the root figure.
     */
    /*public void setAttribute(String name, Object value) {
    	fRootFigure.setAttribute(name,value);
    	changed();	
    }*/


    /**
     * Draws all the contained figures. Sets the current clip
     * to the displaybox of this - warning! assumes the clip shape
     * to be a Rectangle. Descends to children and if a child
     * is about to move adjusts the clip to the displaybox of the
     * child to let it be drawn properly (i.e. outside the display box
     * of the panel).
     */
    public void draw(Graphics g) {
    	//System.err.println("draw: "+this);
    	Shape clip = g.getClip();
    	//System.err.println("clip "+clip);
    	Rectangle r = displayBox();
      //System.err.println("r "+r);
      if (clip != null)
    		g.setClip(((Rectangle)clip).intersection(r));
    	else
    		g.setClip(r);
    	//System.err.println("newclip: "+g.getClip());
    	//fRootFigure.draw(g);
    	drawBackground(g);
      FigureEnumeration k = figures();
      while (k.hasMoreElements()) {
        Figure child = k.nextFigure();
        //System.err.println("child "+child);
        if ((child.inTransit() && !this.inTransit())
            || (child instanceof Sovereign)) {
          //System.err.println("drawing child "+child);
          r = child.displayBox();
          r.grow(4,4); // grow a little to make sure that both dimensions
                       // of a connection are non-zero
          //System.err.println("r "+r);
          Shape clip2 = g.getClip();
          //System.err.println("clip2 "+clip);
          g.setClip(r);
          //System.err.println("g.getClip() "+g.getClip());
          child.draw(g);
          g.setClip(clip2);
        } else
          child.draw(g);
      }
 			drawCorners(g);
      if (fIsHighlighted) highlight(g);
      g.setClip(clip);
    }
    
    protected void drawCorners(Graphics g) {
    	Color c = g.getColor();
    	g.setColor(Color.black);
      Rectangle r = displayBox();
      g.drawOval(r.x,r.y,3,3);
      g.drawOval(r.x+r.width-4,r.y,3,3);
      g.drawOval(r.x+r.width-4,r.y+r.height-4,3,3);
      g.drawOval(r.x,r.y+r.height-4,3,3); 
      g.setColor(c); 
    }	
    
    /**
     * Does nothing.
     */	
    public void drawBackground(Graphics g) {
    }
     /**
      * Find the subfigure that can be selected at the given point
      * or return this, if no subfigure is found.
      */
     public Figure findSelectable(int x, int y) {
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            Figure selectable = null;
            //if (figure.containsPoint(x, y)) {
            selectable = figure.findSelectable(x,y);
            if (selectable != null && ((selectable instanceof Sovereign) || containsPoint(x,y)))
              return selectable;
            //}
        }
        //if (containsPoint(x,y))
        //  return this;
        return null; // FigurePanels are not normally selectable
     }

     public Figure findCtrlSelectable(int x, int y) {
      	//System.err.println("findSelectable"+this);
        FigureEnumeration k = figuresReverse();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            Figure ctrlSelectable = null;
            //if (figure.containsPoint(x, y)) {
            ctrlSelectable = figure.findCtrlSelectable(x,y);
            if (ctrlSelectable != null && ((ctrlSelectable instanceof Sovereign) || containsPoint(x,y)))
              return ctrlSelectable;
            //}
        }
        if (containsPoint(x,y))
        	return this; // FigurePanels are ctrl-selectable
        return null; 
     }

     /**
      * Find the subfigure that can be connected at the given point
      * or return this, if no subfigure is found. If this is not connectable,
      * return null.
      */
    public Figure findConnectable(int x, int y,Figure connection) {
      if (containsPoint(x,y)) {
          FigureEnumeration k = figuresReverse();
          while (k.hasMoreElements()) {
            Figure figure = (k.nextFigure()).findConnectable(x,y,connection);
            if (figure != null)
              return figure;
          }
          if (canConnect()) // && !includes(connection))
            return this;
      }
      return null;
    }


    /**
     * Propagates the removeFromDrawing request up to the container.
     * @see FigureChangeListener
     */
    public void figureRequestRemove(FigureChangeEvent e) {
        /*
        System.err.println(this+".listener(): "+listener());
        if (listener() != null)
            listener().figureRequestRemove(new FigureChangeEvent(this));
        */
        Figure figure = e.getFigure();
        if (fFigures.contains(figure)) {
            fFigures.removeElement(figure);
           	getLayout().removeLayoutComponent(figure);
            figure.removeFromContainer(this);   // will invalidate figure
            figure.release();
            markDirty();
        } else
            System.out.println(this+": Attempt to remove non-existing figure "+figure);
    }

    /**
     * Handle a subfigure change. Ignore subfigures in transit (being dragged),
     * connection figures, subfigures brought to front, and event if itself in
     * transit. Otherwise accommodate the changed subfigure.
     */
    public void figureChanged(FigureChangeEvent e) {
    	if (fHandleChangeEvents) {
    		Figure f = (Figure)e.getSource();
    		if (!f.inTransit() // ignore moving subfigure
    	    	//&& !((f instanceof Sovereign) && (f instanceof Connection))
          	&& !this.inTransit()
          	&& !(e instanceof BroughtToFrontEvent)
    	   	) {
    			markDirty();
    			//changed(); // markDirty() announces the change...
    	  }
    		else
    			//System.err.println(e+" ignored by "+this);
    			;
    	}
    	else
    		//System.err.println(e+" suppressed by "+this);
    		;

    }

  /**
   * By default, does nothing. Containers (composites) that maintain
   * a lay out of subfigures should implement this.
   */
  public void doLayout() {	
  	//System.err.println("<-*-> do lay out "+this);
  	fHandleChangeEvents = false;
  	getLayout().layoutContainer(this);
  	fHandleChangeEvents = true;
  }
  


   /**
    * De-serialize a panel composite. Set itself as the parent of subfigures.
    * 
    */
   private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException {

        s.defaultReadObject();

        FigureEnumeration k = figures();
        while (k.hasMoreElements()) {
            Figure figure = k.nextFigure();
            figure.addToContainer(this);
            figure.setParent(this);
        }
    }

    /**
     * Reads the figures from StorableInput.
     */
    public void read(StorableInput dr) throws IOException {
        super.read(dr);
        int size = dr.readInt();
        fFigures = new Vector(size);
        for (int i=0; i<size; i++) {
        		Figure figure = (Figure)dr.readStorable();
            figure.addToContainer(this);
            figure.setParent(this);
        }
    }


} 