

package CH.ifa.draw.contrib.xcomp;

import java.awt.*;
import java.awt.event.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import java.util.*;
import com.objectspace.jgl.Pair;
import CH.ifa.draw.samples.statechart.*;   
import java.io.*; 

/**
 * A test implementation of a holder for connection labels. A Labels is a
 * CompartmentPanel with an invisible frame. It has a single LabelList-holder
 * as the compartment. A Labels keeps its display box at exactly the preferred
 * size of its compartment.
 */
public class LabelPanel extends FigurePanel implements Sovereign {

  protected Figure  fObservedFigure = null;
  protected Locator fLocator = null;

  public LabelPanel() {
    super();
  }
  public LabelPanel(Point origin, Point corner) {
    super(origin,corner);
  }

  protected DrawLayoutManager getDefaultLayoutManager() {
   	VListLayout l = new VListLayout(1);
   	l.setEmptySize(new Dimension(20,20));
   	return l;
  } 

  public Locator getLocator() {
    return fLocator;
  }
  
  public Figure getObservedFigure() {
    return fObservedFigure;
  }

  public boolean accept(Figure figure) {
  	if (figure instanceof TextFigure)
  		return true;
  	return false;	
  }
  
  	/*protected boolean acceptChange(Point origin, Point corner) {
      return false;
    }*/
    
    /**
     * This has no handles.
     */
    public Vector handles() {
        Vector handles = new Vector();
        //handles.addElement(new NullHandle(this, RelativeLocator.northWest()));
        //handles.addElement(new NullHandle(this, RelativeLocator.northEast()));
        //handles.addElement(new NullHandle(this, RelativeLocator.southEast()));
        //handles.addElement(new NullHandle(this, RelativeLocator.southWest()));
        return handles;
    }

    /**
     * Labels can not be explicitly connected.
     */
    public boolean canConnect() {
        return false;
    }


    /**
     * Updates the location relative to the connected figure.
     * The panel is centered around the located point.
     */
    protected void updateLocation() {
      //System.err.println("updateLocation() of "+this);
        if (fLocator != null) {
            Point p = fLocator.locate(fObservedFigure);
            //System.err.println("p: "+p);
            p.x -= displayBox().x + displayBox().width/2;
            p.y -= displayBox().y + displayBox().height/2;
            //System.err.println("p: "+p);

            if (p.x != 0 || p.y != 0) {
            		//System.err.println("moving "+this);
                boolean markedForMove = inTransit();
                //DragTransactionManager.getInstance().isAboutToMove(fig);
                if (!markedForMove)
                  beginDrag(null);
                
                invalidate();
                basicMoveBy(p.x, p.y);
                invalidate();
                if (!markedForMove)
                  endDrag(null);
            }
            //System.err.println("Labels "+this+" at "+displayBox().getLocation());
        }
    }

    public void release() {
        super.release();
        if (fObservedFigure != null)
            fObservedFigure.removeFigureChangeListener(this);
    }

    public void figureRemoved(FigureChangeEvent e) {
      if (e.getFigure() == fObservedFigure) {
        //System.err.println("my conn removed "+this);
        ((CompositeFigure)parent()).removeChild(this);
      }
    }
    
    /**
     * Propagates the removeFromDrawing request up to the container.
     * @see FigureChangeListener
     */
    public void figureRequestRemove(FigureChangeEvent e) {
        if (e.getFigure() != fObservedFigure)  // ignore the connection
          super.figureRequestRemove(e);        // connection removal is handled
    }                                          // by figureRemoved()

    /**
     * Disconnects the text figure.
     */
    public void disconnect() {
        //System.err.println("disconnecting "+this);
        fObservedFigure.removeFigureChangeListener(this);
        //if (fObservedFigure instanceof LabelledConnection)
        //  ((LabelledConnection)fObservedFigure).removeConnected(this);
        fObservedFigure = null;
        fLocator = null;
    }

    /**
     * Updates the location whenever the figure changes itself.
     */
    public void changed() {
      //System.err.println("changed()");
        super.changed();
        updateLocation();
    }

    public boolean isConnected() {
      return fObservedFigure != null;
    }

    /**
     * Connect with figure using the connectedTextLocator() of figure.
     * Deprecated?
     */
    public void connect(Figure figure) {
        if (fObservedFigure != null) {
            fObservedFigure.removeFigureChangeListener(this);
            fLocator = null;
            //if (fObservedFigure instanceof LabelledConnection)
            //  ((LabelledConnection)fObservedFigure).removeConnected(this);
        }

        fObservedFigure = figure;
        fLocator = figure.connectedTextLocator(this);
        fObservedFigure.addFigureChangeListener(this);
        //if (fObservedFigure instanceof LabelledConnection)
        //  ((LabelledConnection)fObservedFigure).addConnected(this);
        updateLocation();
    }

    /*public void figureChanged(FigureChangeEvent e) {
    	Figure f = (Figure)e.getSource();
    	System.err.println(">> "+this);
    	if (!f.inTransit() // ignore moving subfigure
    	    && !(f instanceof ConnectionFigure)
          && !this.inTransit()
          && !(e instanceof BroughtToFrontEvent)
    	   ) {
    		accommodate(f);
      } else if (f==fObservedFigure && e instanceof BroughtToFrontEvent) {
          //System.err.println("observed brought to front");
          parent().bringToFront(this);
      } else if (f==fObservedFigure)
          updateLocation();
      else
      	System.err.println(e.toString()+" ignored by "+this);
    }*/

     /**
     * Handle a subfigure change. Ignore subfigures in transit (being dragged),
     * connection figures, subfigures brought to front, and event if itself in
     * transit. Otherwise accommodate the changed subfigure.
     */
    public void figureChanged(FigureChangeEvent e) {
    	if (fHandleChangeEvents) {
    		Figure f = (Figure)e.getSource();
    		if (!f.inTransit() // ignore moving subfigure
    	    	&& !(f instanceof Sovereign)
          	&& !this.inTransit()
          	&& !(e instanceof BroughtToFrontEvent)
    	   	) {
    			markDirty();
    			//changed(); // markDirty() announces the change...
    	  }	else if (f==fObservedFigure && e instanceof BroughtToFrontEvent) {
          //System.err.println("observed brought to front");
          parent().bringToFront(this);
      	} else if (f==fObservedFigure)
          updateLocation();
    		else
    			//System.err.println(e+" ignored by "+this);
    			;
    	}
    	else
    		//System.err.println(e+" suppressed by "+this);
    		;

    }

   /**
     * Connect with figure using locator as the locator that computes
     * the position of this relative to the figure.
     */
    public void connect(Figure figure,Locator locator) {
        if (fObservedFigure != null) {
            fObservedFigure.removeFigureChangeListener(this);
        }

        fObservedFigure = figure;
        fLocator = locator;
        fObservedFigure.addFigureChangeListener(this);
        updateLocation();
    }

  /**
   * Returns the string contents of the label figures held by this.
   */
  public String getLabelString() {
  	StringBuffer buf = new StringBuffer();
    Vector labels = ((VListLayout)getLayout()).getComponents();
    for (int i = 0;i<labels.size();i++) {
      buf.append(((TextFigure)labels.elementAt(0)).getText());
      if (i<labels.size()-1)
        buf.append(" ");
    }
    return buf.toString();
  }

  /**
   * Returns true if this has text labels.
   */
  public boolean hasLabels() {
    return ((VListLayout)getLayout()).getComponents().size() > 0;
  }

  /**
   * Returns a Vector of the label Figures held by this in the layout order.
   */
  public Vector getLabels() {
    return ((VListLayout)getLayout()).getComponents();
  }

  /**
   * Moves the label Figures of other to this by dropping them onto the
   * LabelList compartment ot this.
   */
  public void merge(LabelPanel other) {
   Enumeration label = other.getLabels().elements();
   while (label.hasMoreElements())
    //System.err.println(
    	drop((Figure)label.nextElement());
    //);
  }

    /**
     * Return true only if the point is inside the displayBox and this
     * is not empty. This is to ensure that empty label holders can not
     * be selected by tools.
     */
    public Figure findSelectable(int x, int y) {
        if (hasLabels())
          return super.findSelectable(x,y);
        return null;
    }

	/**
	 * Does nothing.
	 */
	public void drawCorners(Graphics g) {
		//if (parent() != null && (parent() instanceof Drawing))
		//	super.drawCorners(g);
	}
	
	/**
   * If highlighted, draws an orange oval inside the display box of this.
   */	
 	public void drawBackground(Graphics g) {
 		if (fIsHighlighted) {
    	Rectangle r = displayBox();
    /*r.grow(-1,-1);
    RenderingTool.drawDashedLine(g,r.x,r.y,r.x+r.width-1,r.y,2);
    RenderingTool.drawDashedLine(g,r.x+r.width-1,r.y,r.x+r.width-1,r.y+r.height-1,2);
    RenderingTool.drawDashedLine(g,r.x+r.width-1,r.y+r.height-1,r.x,r.y+r.height-1,2);
    RenderingTool.drawDashedLine(g,r.x,r.y+r.height-1,r.x,r.y,2);
    r.grow(-1,-1);
    Color c = g.getColor();
    g.setColor(Color.orange);
    g.fillRect(r.x,r.y,r.width,r.height);
    g.setColor(c);  */
    	r.grow(-1,-1);
    	Color c = g.getColor();
    	g.setColor(Color.orange);
    //int ext = Math.max(r.width,r.height);
    //g.fillOval(r.x-(ext-r.width)/2,r.y-(ext-r.height)/2,ext,ext);
    	g.fillOval(r.x,r.y,r.width,r.height);
    	g.setColor(c);
 		}
  }

 /**
   * Does nothing.
   */
  public void highlight(Graphics g) {
  }

   /**
    * De-serialize a panel composite. Set itself as the parent of subfigures.
    * 
    */
   private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException {

        s.defaultReadObject();
		
				if (fObservedFigure != null)
        	setParent(fObservedFigure.parent());
        if (parent() == null)
        	setParent((Figure)StandardDrawingView.getTheDrawing());
    }

}