package CH.ifa.draw.contrib.xcomp;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.util.Vector;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;


/**
 * Tool to select and manipulate figures.
 * A selection tool is in one of three states, e.g., background
 * selection, figure selection, handle manipulation. The different
 * states are handled by different child tools.
 * <hr>
 * <b>Design Patterns</b><P>
 * <img src="images/red-ball-small.gif" width=6 height=6 alt=" o ">
 * <b><a href=../pattlets/sld032.htm>State</a></b><br>
 * SelectionTool is the StateContext and child is the State.
 * The SelectionTool delegates state specific
 * behavior to its current child tool.
 * <hr>
 */

public class SpecialSelectionTool extends SelectionTool {

    public SpecialSelectionTool(DrawingView view) {
        super(view);
    }

    /**
     * Handles mouse down events and starts the corresponding tracker.
     */
    public void mouseDown(MouseEvent e, int x, int y)
    {
        // on Windows NT: AWT generates additional mouse down events
        // when the left button is down && right button is clicked.
        // To avoid dead locks we ignore such events
        if (fChild != null)
            return;

        view().freezeView();
				if (e.isControlDown()) {
            Figure figure = drawing().findCtrlSelectable(e.getX(), e.getY());
            if (figure != null) {
                //figure.parent().bringToFront(figure);
                //figure.broughtToFront();
                fChild = createDragTracker(view(), figure);
            }
				} else {
        	Handle handle = view().findHandle(e.getX(), e.getY());
        	if (handle != null) {
            	fChild = createHandleTracker(view(), handle);
        	}
        	else {
          	  Figure figure = drawing().findSelectable(e.getX(), e.getY());
            	if (figure != null) {
              	  //figure.parent().bringToFront(figure);
                	//figure.broughtToFront();
                	fChild = createDragTracker(view(), figure);
            	}
            	else {
              	  if (!e.isShiftDown()) {
                	    view().clearSelection();
                	}
                	fChild = createAreaTracker(view());
            	}
        	}
        }
        if (fChild != null)
        	fChild.mouseDown(e, x, y);
    }
}