package CH.ifa.draw.contrib.xcomp;

import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import java.awt.*;
import java.util.*;
import java.io.*;
import com.objectspace.jgl.algorithms.Sorting;
import com.objectspace.jgl.*;
import com.objectspace.jgl.adapters.VectorArray;
import com.objectspace.jgl.predicates.BinaryNot;

/**
 * The basic implementation for list-like layout strategies.
 * Like BagLayout this layout manager positions the contained
 * figures 'as they were'. That is, no positioning is performed.
 * This is an example implementation that shows how the mechanisms
 * of super classes are used to create specialized layout managers.
 */ 

public class VListCenteringLayout extends VListLayout {


	/**
   * Returns a NullAligner.
   */
  protected Aligner getTopElementAligner(CompositeFigure container,Figure component) {
  	//return new RelativeAligner(container,component);
  	return new TopCenteringAligner(container);
  }


  /**
   * Returns a NullAligner.
   */
  protected Aligner getMiddleElementAligner(CompositeFigure container,Figure component,Figure previous){
  	//return new RelativeAligner(container,component);
  	return new MiddleCenteringAligner(container,previous);
  }


  /**
   * Returns a NullAligner.
   */
  protected Aligner getBottomElementAligner(CompositeFigure container,Figure component,Figure previous){
  	//return new RelativeAligner(container,component);
  	return new MiddleCenteringAligner(container,previous);
  }
  
  /*protected void grow(CompositeFigure container,Rectangle childBox) {
		childBox.translate(0,-container.getInsets().top);
		super.grow(container,childBox);
  }*/

}


/**
 * A TopLeftAligner computes a position for a figure so that it
 * is aligned horizontally with the left edge of the display box of its holder and
 * positioned below the top edge of the display box of its holder.
 */
class TopCenteringAligner implements Aligner {
  protected CompositeFigure fContainer;

  public TopCenteringAligner(CompositeFigure container) {
    fContainer = container;
  }

  public Point align(Figure child) {
    return new Point(
      fContainer.displayBox().x + fContainer.displayBox().width / 2
      - child.displayBox().width / 2,
      fContainer.displayBox().y + fContainer.getInsets().top
    );
  }
}


/**
 * A MiddleLeftAligner computes a position for a figure so that it
 * is aligned horizontally with the left edge of the display box of its holder and
 * positioned below the previous figure.
 */
class MiddleCenteringAligner extends TopCenteringAligner {
  protected Figure fPrevious;

  public MiddleCenteringAligner(CompositeFigure container,Figure previous) {
    super(container);
    fPrevious = previous;
  }

  public Point align(Figure child) {
    return new Point(
      fContainer.displayBox().x + fContainer.displayBox().width / 2
      - child.displayBox().width / 2,
      fPrevious.displayBox().y + fPrevious.displayBox().height + fContainer.getInsets().top
    );
  }
}