package CH.ifa.draw.contrib.xcomp;

import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import java.awt.*;
import java.util.*;
import java.io.*;
import com.objectspace.jgl.algorithms.Sorting;
import com.objectspace.jgl.*;
import com.objectspace.jgl.adapters.VectorArray;
import com.objectspace.jgl.predicates.BinaryNot;

/**
 * The basic implementation for list-like layout strategies.
 * Like BagLayout this layout manager positions the contained
 * figures 'as they were'. That is, no positioning is performed.
 * This is an example implementation that shows how the mechanisms
 * of super classes are used to create specialized layout managers.
 */ 

public class VListLayout extends BagFlowLayout2 {

	int hgap; // ??
	
	public VListLayout() {
		this(2);
	}
	
	public VListLayout(int hgap) {
		super();
		this.hgap = hgap;
	}

	public int getHgap() {
		return hgap;
	}
	
	public void setHgap(int gap) {
		hgap = gap;
	}
	

  /**
   * Returns the ordering predicate. Figures are ordered in ascending
   * order by the y-coordinate of the origin of their display box.
   */
  protected BinaryPredicate getOrderingPredicate() {
    return new BinaryPredicate () {
      public boolean execute(Object first,Object second) {
        return
          ((Figure)first).displayBox().getLocation().y >
          ((Figure)second).displayBox().getLocation().y;
      }
    };
  }
  
	/**
	 * Force rearrangement.
	 */
	public void invalidateLayout(Figure component) {
		//System.err.println("vlist layout invalidated with component");
		//moveIn((CompositeFigure)component.parent(),component);
		invalidateLayout();	
	}

  protected void attachAligners() {
  	Sorting.sort(
  		new VectorArray(fComponents),
  		new BinaryNot(getOrderingPredicate())
  	);
		super.attachAligners();
  }

  public Rectangle preferredLayoutRectangle(CompositeFigure container) {
  	//System.err.println("------------------------------");
    Vector components = getStationaryFigures();    
    //System.err.println("v: "+v);
    if (components.isEmpty()) {
    	//System.err.println(">>>> pref size empty");
    	//System.err.println("------------------------------");
    	return minimumLayoutRectangle(container);
    }
    Insets insets = container.getInsets(); 
    Point origin = container.displayBox().getLocation();
    origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = new Rectangle(origin);      //System.err.println("r: "+r);
    int lastX = origin.x;
    int lastY = origin.y; 
    for (int i = 0;i<components.size();i++) {
    	if (i>0) 
    		lastY += hgap;
    	Point pos = new Point(lastX,lastY);
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getPreferredSize());
      //System.err.println(">>>> box: "+box);
      //if (i==0)
      //	r = box;
      //else
   			r.add(//box);
   				new Rectangle(origin.x,origin.y,pos.x+box.width-origin.x,pos.y+box.height-origin.y)
   			);
   		lastX = r.x;
   		lastY = r.y + r.height;
      //System.err.println(">>>> r: "+r);
    }
    r.setSize(
    	r.width+insets.left+insets.right,
    	r.height+insets.top+insets.bottom
    );
    r.translate(-insets.left,-insets.top);
    //System.err.println(">>>> >>>> pref size "+r);
    //System.err.println("------------------------------");
    return r;
    //r.width = r.width > cBox.width ? r.width : cBox.width;
    //r.height = r.height > cBox.height ? r.height : cBox.height;
    //System.err.println(">>>> pref size "+r);
    //return new Dimension(r.width,r.height);
  }

  public Rectangle minimumLayoutRectangle(CompositeFigure container) {
      Vector components = getStationaryFigures();    Rectangle cBox = container.displayBox();
    //System.err.println("v: "+v);
    if (components.isEmpty()) {
     	//System.err.println(">>>> min size empty");
    	return new Rectangle (container.displayBox().getLocation(),emptySize());
    	//return new Dimension(container.displayBox().width,container.displayBox().height);
    }
    Insets insets = container.getInsets(); 
    //Point origin = container.displayBox().getLocation();
    //origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    Rectangle r = null; // new Rectangle(origin);
      //System.err.println("r: "+r);
    Point origin = container.displayBox().getLocation();
    origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    int lastX = origin.x;
    int lastY = origin.y; 
    for (int i = 0;i<components.size();i++) {
    	if (i>0) 
    		lastY += hgap;
    	Point pos = new Point(lastX,lastY);
      //Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getMinimumSize());
      Rectangle box = new Rectangle(pos,((Figure)components.elementAt(i)).getPreferredSize());
      //System.err.println(">>>> box: "+box);
      if (i==0)
      	r = box;
      else
   			r.add(box);
   			//	new Rectangle(origin.x,origin.y,pos.x+box.width-origin.x,pos.y+box.height-origin.y)
   			//);
   		lastX = r.x;
   		lastY = r.y + r.height;
      //System.err.println(">>>> r: "+r);
    }
    r.setSize(r.width+insets.left+insets.right,r.height+insets.top+insets.bottom);
    r.translate(-insets.left,-insets.top);
    //System.err.println(">>>> >>>> minimum size "+r);
    return r;
  }

  public void layoutContainer(CompositeFigure container) {
    //System.err.println("<<-:-*-:->> lay out "+container);
		//Vector v = getArrengement(container);
    Vector holdees = getStationaryFigures();
    Insets insets = container.getInsets();
    Point origin = container.displayBox().getLocation();
    origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    int lastX = origin.x;
    int lastY = origin.y; 
    for (int i=0;i<holdees.size();i++) {
     		Figure fig = (Figure)holdees.elementAt(i);
     		//System.err.println("====>fig: "+fig);
    		if (i>0) 
    			lastY += hgap;
        Point newPos = new Point(lastX,lastY);
        //System.err.println("====>newPos:"+newPos);
        Point oldPos = fig.displayBox().getLocation();
        //System.err.println("====>oldPos:"+oldPos);
        Rectangle newBox = new Rectangle(newPos,fig.getPreferredSize());
        if (!newBox.equals(fig.displayBox())) {
        	fig.invalidate();
        	if (!oldPos.equals(newPos))
        		fig.basicMoveBy(newPos.x-oldPos.x,newPos.y-oldPos.y);
        	fig.basicDisplayBox(
        		newPos,
        		new Point(newPos.x+newBox.width,newPos.y+newBox.height)
        	);
        	fig.markDirty(); // should markDirty() mark subfigures dirty too?
        	//fig.invalidate();
        }
        lastX = fig.displayBox().x;
        lastY += fig.displayBox().height;
    		//grow(container,fig.displayBox()); // !!!!! is this really needed
    	}
    }

	/**
   * Returns a NullAligner.
   */
  protected Aligner getTopElementAligner(CompositeFigure container,Figure component) {
  	//return new RelativeAligner(container,component);
  	return new TopLeftAligner(container);
  }


  /**
   * Returns a NullAligner.
   */
  protected Aligner getMiddleElementAligner(CompositeFigure container,Figure component,Figure previous){
  	//return new RelativeAligner(container,component);
  	return new MiddleLeftAligner(container,previous);
  }


  /**
   * Returns a NullAligner.
   */
  protected Aligner getBottomElementAligner(CompositeFigure container,Figure component,Figure previous){
  	//return new RelativeAligner(container,component);
  	return new MiddleLeftAligner(container,previous);
  }
  
  /**
   * Find the position of figure among the subfigures according to the current
   * layout strategy. Return an index to the vector of subfigures. The default
   * is to return the last+1 position.
   */
  protected int findPos(Figure figure) {
    Vector figures = getStationaryFigures();
    int pos = 0;
    while (pos < figures.size() &&
    	     (
    	     	figure == figures.elementAt(pos) ||
           	figure.displayBox().getLocation().y >
           	((Figure)figures.elementAt(pos)).displayBox().getLocation().y
           )
           //figure.center().y >
           //((Figure)figures.elementAt(pos)).center().y
          )
      pos++;
    return pos;
  }
}


/**
 * A TopLeftAligner computes a position for a figure so that it
 * is aligned horizontally with the left edge of the display box of its holder and
 * positioned below the top edge of the display box of its holder.
 */
class TopLeftAligner implements Aligner {
  protected CompositeFigure fContainer;

  public TopLeftAligner(CompositeFigure container) {
    fContainer = container;
  }

  public Point align(Figure child) {
    return new Point(
      fContainer.displayBox().x + fContainer.getInsets().left,
      fContainer.displayBox().y + fContainer.getInsets().top
    );
  }
}


/**
 * A MiddleLeftAligner computes a position for a figure so that it
 * is aligned horizontally withthe left edge of the display box of its holder and
 * positioned below the previous figure.
 */
class MiddleLeftAligner extends TopLeftAligner {
  protected Figure fPrevious;

  public MiddleLeftAligner(CompositeFigure container,Figure previous) {
    super(container);
    fPrevious = previous;
  }

  public Point align(Figure child) {
    return new Point(
      fContainer.displayBox().x + fContainer.getInsets().left,
      fPrevious.displayBox().y + fPrevious.displayBox().height + fContainer.getInsets().top
    );

  }
}