package CH.ifa.draw.contrib.xcomp;

import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import java.awt.*;
import java.util.*;
import java.io.*;
import com.objectspace.jgl.algorithms.Sorting;
import com.objectspace.jgl.*;
import com.objectspace.jgl.adapters.VectorArray;
import com.objectspace.jgl.predicates.BinaryNot;

/**
 * The basic implementation for list-like layout strategies.
 * Like BagLayout this layout manager positions the contained
 * figures 'as they were'. That is, no positioning is performed.
 * This is an example implementation that shows how the mechanisms
 * of super classes are used to create specialized layout managers.
 */ 

public class VTiledLayout extends VListLayout {

	public VTiledLayout() {
		super();
	}
	
	public VTiledLayout(int hgap) {
		super(hgap);
	}
		
	protected int findMaxWidth(Vector dimensions,CompositeFigure container) {
		int w = 
			container.displayBox().width - container.getInsets().left - container.getInsets().right;
		for (int i=0;i<dimensions.size();i++) 
			w = Math.max(w,((Dimension)dimensions.elementAt(i)).width);
		return w;
	}
		 
	protected Vector preferredSizes(Vector figures) {
		Vector sizes = new Vector();
		for (int i=0;i<figures.size();i++) 
			sizes.addElement(((Figure)figures.elementAt(i)).getPreferredSize());
		return sizes;
	}

	protected Vector minimumSizes(Vector figures) {
		Vector sizes = new Vector();
		for (int i=0;i<figures.size();i++) 
			sizes.addElement(((Figure)figures.elementAt(i)).getMinimumSize());
		return sizes;
	}

  public void layoutContainer(CompositeFigure container) {
    //System.err.println("<<-:-*-:->> lay out "+container);
		//Vector v = getArrengement(container);
    Vector holdees = getStationaryFigures();
    Vector sizes = preferredSizes(holdees);
    int maxWidth = findMaxWidth(sizes,container);
    Insets insets = container.getInsets();
    Point origin = container.displayBox().getLocation();
    origin.translate(insets.left,insets.top);
    //System.err.println("<<<< pos: "+pos+" origin: "+origin);
    int lastX = origin.x;
    int lastY = origin.y; 
    for (int i=0;i<holdees.size();i++) {
     		Figure fig = (Figure)holdees.elementAt(i);
     		//System.err.println("====>fig: "+fig);
    		if (i>0) 
    			lastY += hgap;
        Point newPos = new Point(lastX,lastY);
        //System.err.println("====>newPos:"+newPos);
        Point oldPos = fig.displayBox().getLocation();
        //System.err.println("====>oldPos:"+oldPos);
        Rectangle newBox = new Rectangle(newPos,(Dimension)sizes.elementAt(i));
        if (i == holdees.size()-1) 
           newBox.height += 
        		container.displayBox().getLocation().y +
        		container.displayBox().height - container.getInsets().bottom -
        		(newBox.y + newBox.height);
        newBox.width = Math.max(maxWidth,newBox.width);
        if (!newBox.equals(fig.displayBox())) {
        	fig.invalidate();
        	if (!oldPos.equals(newPos))
        		fig.basicMoveBy(newPos.x-oldPos.x,newPos.y-oldPos.y);
        	fig.basicDisplayBox(
        		newPos,
        		new Point(newPos.x+newBox.width,newPos.y+newBox.height)
        	);
        	fig.markDirty(); // should markDirty() mark subfigures dirty too?
        	//fig.invalidate();
        }
        lastX = fig.displayBox().x;
        lastY += fig.displayBox().height;
    		//grow(container,fig.displayBox()); // !!!!! is this really needed
    	}
    }

	/*public void layoutContainer(CompositeFigure container) {
      System.err.println("<<<<<-:-**-:->>>>> lay out "+container);
			Vector v = getArrengement(container);
      Vector holdees = getStationaryFigures();
      //Vector sizes = minimumSizes(holdees);
      Vector sizes = preferredSizes(holdees);

      int maxWidth = findMaxWidth(sizes,container);
      for (int i=0;i<v.size();i++) {
     		Figure fig = (Figure)holdees.elementAt(i);
     		System.err.println("====>fig: "+fig);
        Point newPos = ((Aligner)v.elementAt(i)).align(fig);
        System.err.println("====>newPos:"+newPos);
        Point oldPos = fig.displayBox().getLocation();
        System.err.println("====>oldPos:"+oldPos);
        Rectangle newBox = new Rectangle(newPos,(Dimension)sizes.elementAt(i));
        if (i == v.size()-1) 
           newBox.height += 
        		container.displayBox().getLocation().y +
        		container.displayBox().height - container.getInsets().bottom -
        		(newBox.y + newBox.height);
        newBox.width = Math.max(maxWidth,newBox.width);
        if (!newBox.equals(fig.displayBox())) {
        	fig.invalidate();
        	if (!oldPos.equals(newPos))
        		fig.basicMoveBy(newPos.x-oldPos.x,newPos.y-oldPos.y);
        	fig.basicDisplayBox(
        		newPos,
        		new Point(newPos.x+newBox.width,newPos.y+newBox.height)
        	);
        	fig.markDirty(); // should markDirty() mark subfigures dirty too?
        	//fig.invalidate();
        }
    		//grow(container,fig.displayBox());
    	}
   }*/
}