

package CH.ifa.draw.framework;

import java.awt.event.*;
/**
 * A draggable object may need to notify its dependants when it is
 * being dragged. The interface provides methods for signalling
 * that an object is about to be dragged (i.e. selected for drag),
 * it is being dragged (mouseDrag), and released (mouseUp).
 * This gives the possibility to distribure the responsibility of
 * managing a dragging transaction between tools and dragged figures.
 */
public interface Draggable {

  /**
   * Tell this that it is about to be dragged.
   */
  void beginDrag(MouseEvent e);

  /**
   * Tell this that it has been dragged.
   */
  void dragged(MouseEvent e);

  /**
   * The drag transaction is over.
   */
  void endDrag(MouseEvent e);

  /**
   * Answer true if this is the target of a drag transaction at the
   * moment (being dragged).
   */
  boolean inTransit();
} 