package CH.ifa.draw.framework;


import CH.ifa.draw.standard.CompositeFigure;
import java.awt.Dimension;
import java.awt.Rectangle;


/**
 * Interface for layout managers. Layout managers encapsulate a strategy
 * for laying out containers (CompositeFigures).
 */
public interface DrawLayoutManager extends java.io.Serializable {
	
	/**
	 * Lays out the specified container.
	 */
	public void layoutContainer(CompositeFigure container);
	
	/**
	 * Invalidates the layout. Makes the layout manager to discard all
	 * cached information.
	 */
	public void invalidateLayout();
	
	/**
	 * Invalidates the layout. Makes the layout manager to discard all
	 * cached information. Indicates the component that has caused the
	 * layout to become invalid.
	 */
	public void invalidateLayout(Figure component);

	/**
	 * Calculates the minimum size dimensions for the specified container.
	 */ 
	public Dimension minimumLayoutSize(CompositeFigure container);

	/**
	 * Calculates the minimum size display box for the specified container
	 * at its current location.
	 */ 
	public Rectangle minimumLayoutRectangle(CompositeFigure container);
	
	/**
	 * Translates the rectangle defining the minimum size of the container
	 * according to the layout strategy. This way layout managers can specify
	 * how containers behave when resized by dragging their handles.
	 */
	public void overlayMinimumRectangle(CompositeFigure container,Rectangle minimum);
	
	/**
	 * Calculates the preferred size dimensions for the specified container.
	 */ 
	public Dimension preferredLayoutSize(CompositeFigure container);
	
	/**
	 * Calculates the preferred size display box for the specified container
	 * at its current location.
	 */ 
	public Rectangle preferredLayoutRectangle(CompositeFigure container);
	
	/**
	 * Adds a component (figure) to the layout.
	 */
	public void addLayoutComponent(Figure component);
	
	/**
	 * Removes a component (figure) from the layout.
	 */
	public void removeLayoutComponent(Figure component);
	
	/**
	 * Replaces a component in the layout.
	 */
	public void replaceLayoutComponent(Figure component, Figure replacement);
	
}