/**
 * @(#)ParentChangeEvent.java
 */

package CH.ifa.draw.framework;

import CH.ifa.draw.standard.*;

/**
 * Indicate changing the parent of a figure. Provides a static method
 * for actually doing the change.
 */

public class ParentChangeEvent
	extends FigureChangeEvent {
	
	public ParentChangeEvent(Figure source) {
		super(source);
	}
	/**
   * Changes the parent of a Figure. The method does an orphan-add transaction
   * (removeChild()-add()) rather than remove-add because removing the Figure also releases
   * it (destroys). The method also takes into account the special case
   * when a new Figure is inserted for the first time in the drawing and
   * has a null parent. The method checks that the change won't break the
   * hierarchy in which case the figure is hung directly under the drawing.
   */
	public static void changeParent(Figure figure,Figure newParent) {
 		//System.err.println("newParent == "+newParent);
 		//System.err.println("<< before: figure.parent()=="+figure.parent());
    if (figure.parent() != null)  // a new figure?
 			((CompositeFigure)figure.parent()).removeChild(figure);
    //else System.err.println("new figure");
    if (newParent != null && newParent != figure && !figure.includes(newParent)) // check consistency
 			  figure.parent(newParent);
 		else
 			  figure.parent((Figure)StandardDrawingView.getTheDrawing());  // hang it somewhere!
    //System.err.println("before add");
    ((CompositeFigure)figure.parent()).add(figure);
 		//System.err.println(">> after: figure.parent()=="+figure.parent());
    //System.err.println("parent of "+figure+" set to "+figure.parent());
	}
}