/*
 * @(#)RootFigure
 * 
 */


package CH.ifa.draw.framework;

import java.awt.Point;
import java.awt.Dimension;

/**
 * A RootFigure acts as the physical frame figure 
 * of an owning FrameComposite. 
 * The only figure that holds a reference
 * to the root figure is the owning FrameComposite.
 * The idea is that with a small set of changes, any Figure
 * could be used as a frame.<p>
 * A FrameComposite delegates requests for handles, connectors, and
 * for setting the display box to its frame figure.
 * A RootRectangleFigure targets the handles and the connector
 * to its FrameComposite. It also calls the willChange() and changed()
 * methods of its FrameComposite before and (resp.) after setting
 * its display box. A RootFigure
 * can set a minimum size for the frame.<p>
 * Due to the absence of multiple inheritance in Java, concrete root figures
 * need to duplicate the code for managing the owning frame.
 */
public interface RootFigure
	extends Figure {

   	public void displayBox(Point origin, Point corner);
   	public Dimension getMinimumSize();

}