package CH.ifa.draw.samples.statechart;

import relap.LanguageModel.*;
import java.io.*;
import com.objectspace.jgl.*;

public class StateChart extends RelationalGrammarImplementation {

	HashSet fFixedStartAttrs;

	public String terminalDeclarations () {
		return "rrect text arrow initial final statePanel namePanel "+
			     "itPanel pseudoPanel labelPanel";
	}

	public String nonTerminalDeclarations () {

		return "StateChart Initial Final State StateSymbol Trans "+
			     "NameCompartment StateCompartment ITCompartment Label";

	}
	public String startSymbolDeclaration () {

		return "StateChart";

	}

	public String relationDeclarations () {

		return "inside enters exits attached ";

	}

	public String attributeDeclarations () {

		return "root ";

	}

	public HashSet fixedStartAttributes() {
		if (fFixedStartAttrs == null) {
			fFixedStartAttrs = new HashSet();
			fFixedStartAttrs.add("root");
		}
		return fFixedStartAttrs;
	} 

	public GrammarProduction StateChart_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("State machine");
		p.rightHandSide("pseudoPanel Initial State+ Final*");
		p.constraints("inside(2:root,1) inside(3:root,1) inside(4:root,1)");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.StateChartRep");
		return p;	 
	}
	
	public GrammarProduction Initial_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("initial (pseudo) state");
		p.rightHandSide("initial");
		//p.constraints("exits(2:root,1)");
		p.predicates("1 {arrow} exits(_,1)");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.InitialRep");
		return p;	 
	}

	public GrammarProduction Final_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("final (pseudo) state");
		p.rightHandSide("final Trans+");
		p.constraints("enters(2:root,1)");
		p.predicates("1 {rrect,initial} exits(2:root,_) ~{arrow}exits(_,1)");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.FinalRep");
		return p;	 
	}


	public GrammarProduction State_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("state with zero or more incoming transitions");
		p.rightHandSide("StateSymbol Trans*");
		p.constraints("enters(2:root,1:root)");
		p.predicates("1 {rrect,initial} exits(2:root,_)");
		p.assignments("0:root = 1:root");
		p.semantics("CH.ifa.draw.samples.statechart.StateRep");
		return p;	 
	}
	

	public GrammarProduction StateSymbol_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("state symbol structure");
		p.rightHandSide("rrect NameCompartment? ITCompartment? StateCompartment*");
		p.constraints("inside(2:root,1) inside(3:root,1) inside(4:root,1)");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.StateSymbolRep");
		return p;	 
	}


public GrammarProduction NameCompartment_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("name compartment");
		p.rightHandSide("namePanel text+");
		p.constraints("inside(2,1)");
		//p.notExistsConstraints("only 1 {statePanel} parent(1,_)");
		p.order("2 above ");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.NameCompartmentRep");
		return p;	 
	}

public GrammarProduction ITCompartment_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("internal transition compartment");
		p.rightHandSide("itPanel text+");
		p.constraints("inside(2,1)");
		//p.notExistsConstraints("only 1 {statePanel} parent(1,_)");
		p.order("2 above ");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.ITCompartmentRep");
		return p;	 
	}

public GrammarProduction StateCompartment_empty(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("empty state compartment");
		p.rightHandSide("statePanel");
		p.disambiguate("~{rrect,initial,text,final}inside(_,1)");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.StateCompartmentRep");
		return p;	 
	}

public GrammarProduction StateCompartment_collapsedTexts(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("state compartment with collapsed text compartments");
		p.rightHandSide("statePanel text+");
		p.constraints("inside(2,1)");
		p.disambiguate("~{rrect,initial,final}inside(_,1)");
		p.order("2 above ");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.StateCompartmentRep");
		return p;	 
	}

public GrammarProduction StateCompartment_nestedDiagram(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("state compartment with nested statechart diagram");
		p.rightHandSide("statePanel text* Initial? State+ Final*");
		p.constraints("inside(2,1) inside(3:root,1) inside(4:root,1) inside(5:root,1)");
		//p.notExistsConstraints("~{text}inside(_,1)");
		p.order("2 above ");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.CompositeStateCompartmentRep");
		return p;	 
	}


	public GrammarProduction Trans_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("transition");
		p.rightHandSide("arrow Label?");
		p.constraints("attached(2:root,1)");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.TransitionRep");
		return p;	 
	}

	public GrammarProduction Label_(GrammarProduction p) 
	throws InvalidGrammarException {

		p.description("multi-line transition label");
		p.rightHandSide("labelPanel text+");
		p.constraints("inside(2,1)");
		p.order("2 above");
		p.assignments("0:root = 1");
		p.semantics("CH.ifa.draw.samples.statechart.LabelRep");
		return p;	 
	}

}