package CH.ifa.draw.samples.statechart;

import relap.Analyzer.*;
import relap.Util.*;
import relap.LanguageModel.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import java.io.*;
import java.util.Date;
import java.util.Enumeration;
import java.awt.event.*;
import java.awt.*;
import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.*;
import com.objectspace.jgl.predicates.*;
import relap.Editor.*;
import CH.ifa.draw.samples.statechart.semantics.StateChartAnalyzer;

public class StateChartAnalysisView
 
	extends RelapAnalysisView implements ActionListener {
	
  Button dismissButton;
  Button clearButton;
  Button analyzeButton;
  Button contextButton;
  Button simulateButton;
  Button generateButton;
  Checkbox loggingToggle;
  Panel actionPanel;
  Panel resultPanel;
  Panel dlgCntrlPanel;
  Panel fill;
  String analysisContext;
  PromptUser contextPrompt;
  			
	public StateChartAnalysisView(Frame editorFrame,
																 Analyzer analyzer,String title) {
		super(editorFrame,analyzer,title);
		analysisContext = null;
		GridBagConstraints c = new GridBagConstraints();
		GridBagLayout l = new GridBagLayout();
		setLayout(l);
		analyzeButton  = new Button("Analyze");
		analyzeButton.addActionListener(this);
		contextButton  = new Button("Context...");
		contextButton.addActionListener(this);
		loggingToggle = new Checkbox("Write log",false);
		analyzer.setLogging(false);
		loggingToggle.addItemListener(
			new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (e.getSource() == loggingToggle)
						getAnalyzer().setLogging(loggingToggle.getState());
				}
			}
		);
		simulateButton  = new Button("Simulate");
		simulateButton.addActionListener(this);
		simulateButton.setEnabled(false);
		generateButton  = new Button("Generate");
		generateButton.addActionListener(this);
		generateButton.setEnabled(false);
		dismissButton = new Button("Close");
		dismissButton.addActionListener(this);
		clearButton = new Button("Clear");
		clearButton.addActionListener(this);
		clearButton.setEnabled(false);
		actionPanel = new Panel(new FlowLayout(FlowLayout.LEFT));
		actionPanel.add(analyzeButton);
		actionPanel.add(contextButton);
		actionPanel.add(loggingToggle);
		actionPanel.add(simulateButton);
		actionPanel.add(generateButton);
		resultPanel = new Panel();
		dlgCntrlPanel = new Panel(new FlowLayout(FlowLayout.RIGHT));
		dlgCntrlPanel.add(clearButton);
		dlgCntrlPanel.add(dismissButton);
		l.setConstraints(actionPanel,c);
		add(actionPanel);
		c.weightx = 1.0;
		fill = new FillerPanel(dismissButton.getSize());
		l.setConstraints(fill,c);
		add(fill);
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weightx = 0.0;
		l.setConstraints(dlgCntrlPanel,c);
		add(dlgCntrlPanel);
		c.weightx = 0.9;
		c.weighty = 1.0;	
		l.setConstraints(resultPanel,c);
		add(resultPanel);
		pack();
	}
		
	public void actionPerformed(ActionEvent event) {
    Object source = event.getSource();
    if (source == dismissButton) {
    		reset();
        setVisible(false);
    } else if (source == analyzeButton) {
    	if (analysisContext == null)
    		getContext();
    	if (analysisContext != null) {
    		((StateChartAnalyzer)fAnalyzer).setAnalysisContextClass(analysisContext);
    		reset();
      	fAnalyzer.analyzeDrawing();     
    	}
    } else if (source == contextButton) {
    	getContext();
    } else if (source == clearButton) {
    	reset();
    } else if (source == simulateButton) {
    	//showSemantics();
    	System.err.println("PING!");
    } else if (source == generateButton) {
    	//showSemantics();
    	System.err.println("PONG!");
    }
	}
	
	public void update(RelapEvent e) {
		System.err.println("update: "+e);
		if (e.getSource() == fAnalyzer) {
			if (e instanceof AnalysisExceptionEvent) {
				showException(((AnalysisExceptionEvent)e).getException());
				clearButton.setEnabled(true);
			} else if (e instanceof StartEvent) {
				//System.err.println(">> got StartEvent");
				reset();
			} else if (e instanceof StopEvent) {
				showResult(((StopEvent)e).getResult());
				clearButton.setEnabled(true);
			}
		}
	}

	private void setResultContents(Panel result) {
		//if (resultPanel instanceof ParseSuccessPanel)
		//	((ParseSuccessPanel)resultPanel).cleanUp();
		// else 
		if (resultPanel instanceof ParseFailurePanel)
			((ParseFailurePanel)resultPanel).cleanUp();
		remove(resultPanel);
		resultPanel = result;
		GridBagLayout l = (GridBagLayout)getLayout();
		GridBagConstraints c = new GridBagConstraints();
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weightx = 0.9;
		c.weighty = 1.0;	
		l.setConstraints(resultPanel,c);
		add(resultPanel);
	}
	
	public void reset() {
		System.err.println("reset()");
		simulateButton.setEnabled(false);
		generateButton.setEnabled(false);
		clearButton.setEnabled(false);
		//fAnalyzer.clearSelection();
		setResultContents(new Panel());
		pack();
		repaint();
	}	
	
	public void showResult(ProcessingResult result) {
		System.err.println("showResult()");
		if (result.success()) {
			/*
			fAnalyzer.setSemantics(
				((InActiveState)((ParseResult)result).parseTrees().start().get()).getSemantics()
			);
			setResultContents(new ParseSuccessPanel(fAnalyzer,(ParseResult)result));
			pack();
			semanticsButton.setEnabled(true);
			*/
			if (result.warnings().isEmpty())
				showSemantics();
			else {
				setResultContents(
					new ParseFailurePanel((Frame)parent,fAnalyzer,result,"Warnings:")
				);
				pack();
			}
			generateButton.setEnabled(true); // just to make sure
			simulateButton.setEnabled(true);
		}
		else {
			//fAnalyzer.setSemantics(null);
			//System.err.println(result);
			setResultContents(
				new ParseFailurePanel((Frame)parent,fAnalyzer,result)
			);
			pack();
			generateButton.setEnabled(false); // just to make sure
			simulateButton.setEnabled(false);
		}
	}
	
	public void showSemantics() {
		if (((StateChartAnalyzer)fAnalyzer).getStateMachine() != null) {
			Panel p = new Panel();
			TextArea text = new TextArea(8,60);
			p.add(text);
			text.setText( 
				((StateChartAnalyzer)fAnalyzer).getStateMachine().toString()
			);
			setResultContents(p);
			pack();
		}
		else 
			System.err.println("Semantics of drawing not analyzed");
	}
	
	protected void showException(Exception e) {
		setResultContents(new ExceptionPanel(	e.toString()));
		pack();
	}
	
	protected void getContext() {
		if (contextPrompt == null)
			contextPrompt = 
				new PromptUser(
					parent,
					"Set the context for analyzing the statechart",
					"Context class:",
					"CH.ifa.draw.samples.statechart.test.TestContext"
				);
		Point origin = getLocation();
		origin.translate(5,getSize().height+5);
		contextPrompt.setLocation(origin);
		contextPrompt.setVisible(true);
		analysisContext = contextPrompt.getInput();
		//System.err.println(">>>>>>>@@ "+analysisContext);
	}
}