

package CH.ifa.draw.samples.statechart;

import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.contrib.xcomp.*;
import java.util.*;
import java.awt.*;
import java.io.*;

public class TransitionSegment 
	implements LabelledSegment, Serializable {
		
  static final int CenterLeft = 0;
  static final int CenterRight = 1;

  int fSegment;
  LabelledConnection fOwner;
  Vector fLabels;

  public int first() {
    return CenterLeft;
  }

  public int last() {
    return CenterRight;
  }

  public TransitionSegment(LabelledConnection owner, int segment) {
    fOwner = owner;
    fSegment = segment;
    fLabels = new Vector(2);
    createLabelSlots();
  }

  public void setSegment(int segment) {
    Enumeration label = fLabels.elements();
    while (label.hasMoreElements()) {
      LabelPanel l = (LabelPanel)label.nextElement();
      if (l != null)
        ((SlotLocator)l.getLocator()).setSegment(segment);
    }
    fSegment = segment;
  }

  public LabelPanel getLabelAt(int index) {
    //0 <= index <= 1
    return (LabelPanel)fLabels.elementAt(index);
  }

  public void setLabelAt(LabelPanel label,int index) {
    //0 <= index <= 1
    fLabels.setElementAt(label,index);
    ParentChangeEvent.changeParent(label,fOwner.parent());
    label.connect(
      fOwner,
      mapToSlot(index,label)
    );
  }

  public void createLabelAt(int index) {
    //0 <= index <= 1
    LabelPanel label = new LabelPanel();
    label.displayBox(new Point(fOwner.displayBox().getLocation()),new Point(10,10));
    setLabelAt(label,index);
  }

  protected SlotLocator mapToSlot(int index, LabelPanel label) {
    switch (index) {
      case CenterLeft :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getCenterSlotLocator(fOwner,label,true)
          );
        //break;
      }
      case CenterRight :
      {
        return new SlotLocator(
            fOwner,label,fSegment,
            SlotLocator.getCenterSlotLocator(fOwner,label,false)
          );
        //break;
      }
      default :
        throw new ArrayIndexOutOfBoundsException("Invalid slot index "+index);
    }
  }

  protected void createLabelSlots() {
    for (int i = first(); i <= last(); i++) {
      fLabels.addElement(null);
      createLabelAt(i);
    }
  }

  public void merge(LabelledSegment other) {
    other.setSegment(fSegment);   // still needed ??
    for (int i=first();i<=last();i++) {
    	LabelPanel l = other.getLabelAt(i);
      getLabelAt(i).merge(l);
      l.disconnect();
      ((CompositeFigure)l.parent()).remove(l);
    }
  }
}