package CH.ifa.draw.samples.statechart.semantics;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.*;
import relap.Util.*;
import java.util.*;

public class CompositeState  
	extends State {

	protected SList fSubStates;
	protected boolean fIsConcurrent;
	protected SList fInitials;
	
	public CompositeState(boolean isConcurrent) {
		super();
		fSubStates = new SList();
		fIsConcurrent = isConcurrent;
		fInitials = null;
	}

	public CompositeState() {
		super();
		fSubStates = new SList();
	}

	public void setIsConcurrent(boolean flag) {
		fIsConcurrent = flag;
	}
	
	public boolean isConcurrent() {
		return fIsConcurrent;
	}
	
	public boolean isRegion() {
		if (getContainer() != null && getContainer().isConcurrent())
			return true;
		return false;
	}

	public void setSubStates(SList vertices) {
		fSubStates.clear();
		Copying.copy(vertices,fSubStates);
	}
	
	public void addSubState(State child) {
		fSubStates.pushBack(child);
	}
	
	public SList getSubStates() {
		return new SList(fSubStates);
	}
	
	public Initial getInitialVertex() {
		if (fInitials == null)
			fInitials = getInitials();
		return (Initial)fInitials.front();
	}	
	
	private SList getInitials() {
		SList initials = new SList();
		Enumeration sub = fSubStates.elements();
		while (sub.hasMoreElements()) {
			State s = ((State)sub.nextElement());
			if (s instanceof Initial)
				initials.pushBack(s);
		}
		return initials;
	}
	
	/*	
	private void checkInitialCount(Vector accuErrors) {
		initials = Utils.asHashSet(getInitials());
			if (initials.size() > 1) {
			Flaw f = new Flaw("Composite state can have only one initial vertex",initials,Severity.createError());	
			if (!accuErrors.contains(f))
				accuErrors.addElement(f);
		}
	}
	*/
	
	public void check(Vector accuErrors) {
		super.check(accuErrors);
		//checkInitialCount(accuErrors);
		Enumeration sub = fSubStates.elements();
		while (sub.hasMoreElements()) 
			((State)sub.nextElement()).check(accuErrors);
	}

	public String	 describeStructure() {
		StringBuffer buf = new StringBuffer(inset()+"composite state with substates:\n");
		Enumeration state = fSubStates.elements();
		while (state.hasMoreElements()) {
			buf.append(state.nextElement());
			if (state.hasMoreElements())
				buf.append("\n");
		}
		return buf.toString();
	}
}