package CH.ifa.draw.samples.statechart.semantics;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.*;
import relap.Util.*;
import java.util.*;

public abstract class State 
	extends StateVertex {
	
	protected SList fInternals;
	protected String fEntry;
	protected String fExit;
	protected String fDoActivity;
	protected StateMachine fMachine;
	
	public State() {
		fInternals = new SList();
		fEntry = "";
		fExit = "";
		fDoActivity = "";
	}

	public void setStateMachine(StateMachine machine) {
		fMachine = machine;
	}
	
	public StateMachine getStateMachine() {
		if (fMachine == null && getContainer() != null)
			return getContainer().getStateMachine();
		return fMachine;
	}
	
	public SList getInternals() {
		return new SList(fInternals);
	}
	
	public void setInternals(SList internals) {
		fInternals.clear();
		Copying.copy(internals,fInternals);
		for (ForwardIterator i=fInternals.start();!i.atEnd();i.advance())
			((InternalTransition)i.get()).setOwner(this);
	}
	
	public void addInternals(SList internals) {
		for (ForwardIterator i=internals.start();!i.atEnd();i.advance())
			addInternal((InternalTransition)i.get());
	}
	
	public void addInternal(InternalTransition it) {
		fInternals.pushBack(it);
		it.setOwner(this);
	}
	
	public String getEntry() {
		return fEntry;
	}
	
	public String setEntry(String entry) {
		return fEntry = entry;
	}
	
	public String getExit() {
		return fExit;
	}
	
	public String setExit(String exit) {
		return fExit = exit;
	}
	
	public String getDoActivity() {
		return fDoActivity;
	}
	
	public String setDoActivity(String doActivity) {
		return fDoActivity = doActivity;
	}
	
	protected HashSet checkName() {
		HashSet conflictSet = new HashSet();
		if (getName().equals(""))
			return conflictSet;
		Enumeration states = getStateMachine().getStates().elements();
		while (states.hasMoreElements()) {
			State state = (State)states.nextElement();
			if (state != this && state.getName().equals(getName()))
				conflictSet.add(state);
		}
		if (!conflictSet.isEmpty())
			conflictSet.add(this);
		return conflictSet;
	}
				
	protected void screenNameFromInternals(Vector accuErrors) {
		Vector invalidIT = new Vector();
		Object name = null;
		for (ForwardIterator trn=getInternals().start();!trn.atEnd();trn.advance()) {
			InternalTransition it = (InternalTransition)trn.get();
			it.check(invalidIT);		
			if (invalidIT.size()>0) {
				if (it.getEffect().equals("")) {// possible state name
					if (getName().equals(""))
						setName(it.getTrigger());
					else if (!getName().equals(it.getTrigger())) {
						HashSet objects = new HashSet();
						objects.add(it);
						objects.add(this);
						addError(
							accuErrors,
							"State name \""+it.getTrigger()+"\" does not match the already declared name",
							objects
						);
					}
					name = it;
					break;
				}
			}	 
		}
		if (name != null) {
			SList internals = getInternals();
			internals.remove(name);
			setInternals(internals);
		}
	}
	
	protected void screenActionsFromInternals(Vector accuErrors) {
		SList internals = getInternals();
		SList actions = new SList();
		for (ForwardIterator trn=internals.start();!trn.atEnd();trn.advance()) {
			InternalTransition it = (InternalTransition)trn.get();
			if (it.getTrigger().equals("entry")) {
				actions.pushBack(it);
				setEntry(it.getEffect());
			} else if (it.getTrigger().equals("exit")) {
				actions.pushBack(it);
				setExit(it.getEffect());
			} else if (it.getTrigger().equals("do")) {
				actions.pushBack(it);
				setDoActivity(it.getEffect());
			} 
		}
		for (ForwardIterator p = actions.start();!p.atEnd();p.advance()) 
			internals.remove(p.get());
		setInternals(internals);
	}


	protected void checkActions(Vector accuErrors) {
		if (!getEntry().equals(""))
			if (!getStateMachine().getContext().getActionNames().contains(getEntry()))
				addError(accuErrors,"Unknown entry action '"+getEntry()+"'",this);
		if (!getExit().equals(""))
			if (!getStateMachine().getContext().getActionNames().contains(getExit()))
				addError(accuErrors,"Unknown exit action '"+getExit()+"'",this);
		if (!getDoActivity().equals(""))
			if (!getStateMachine().getContext().getActionNames().contains(getDoActivity()))
				addError(accuErrors,"Unknown entry action '"+getDoActivity()+"'",this);
	}

	public void check(Vector accuErrors) {
		screenNameFromInternals(accuErrors);
		screenActionsFromInternals(accuErrors);
		HashSet nameConflicts = checkName();
		//System.err.println("State::check -> "+nameConflicts);
		if (!nameConflicts.isEmpty())
			addWarning(accuErrors,"States should have distinct names",nameConflicts);
		for (ForwardIterator trn=getInternals().start();!trn.atEnd();trn.advance()) 
			((InternalTransition)trn.get()).check(accuErrors);
		checkActions(accuErrors);
		
	}

	public abstract String	 describeStructure();

	protected String describeInternalTransitions() {
		ForwardIterator t = getInternals().start();
		StringBuffer actions = new StringBuffer();
		if (!getEntry().equals(""))
			actions.append(inset()+"entry: "+getEntry()+"\n");
		if (!getExit().equals(""))
			actions.append(inset()+"exit: "+getExit()+"\n");
		if (!getDoActivity().equals(""))
			actions.append(inset()+"do: "+getDoActivity()+"\n");
		if (t.atEnd())
			return actions.toString()+inset()+"<no internal transitions in this state>";
		StringBuffer buf = new StringBuffer(inset()+"internal transitions:\n");
		while (!t.atEnd()) {
			buf.append(inset()+t.get().toString());
			t.advance();
			if (!t.atEnd())
				buf.append("\n");
		}
		return actions.toString()+buf.toString();
	}	

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(inset()+"-------------------------------------------\n");
		buf.append(inset()+"State "+getName()+" (id: "+getId()+")\n");
		buf.append(describeInternalTransitions()+"\n");
		buf.append(describeExitingTransitions()+"\n");
		buf.append(inset()+"State structure:\n");
		buf.append(describeStructure());
		return buf.toString();
	}			
	
}