package CH.ifa.draw.samples.statechart.semantics;

import relap.Editor.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.samples.statechart.*;
import java.io.*;
import java.util.*;
import java.awt.*;
import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.*;
import com.objectspace.jgl.predicates.*;
import relap.Analyzer.*;
import relap.LanguageModel.*;


public class StateChartAnalyzer extends AbstractAnalyzer {

	protected StateMachine fMachine;
	protected SemanticMap fMap;
	protected Context fAnalysisContext;
	protected String fContextClass;
	
	public StateChartAnalyzer(DrawingEditor editor) {
		super(editor);
		fMachine = null;
		fMap = null;
		fAnalysisContext = null;
		fContextClass = "";
	}
	
	public String getAnalysisContextClass() {
		return fContextClass;
	}
	
	public void setAnalysisContextClass(String className) {
		fContextClass = className;
	}

	public StateMachine getStateMachine() {
		return fMachine;
	}
	
	protected String getGrammar() {
		return "CH.ifa.draw.samples.statechart.StateChart";
	}
	
	protected ObjectToVisitorMap getVisitorMap() {
		return new StateChartInputGeneratorMap();
	}
	
	protected ObjectToSymbolMap getSymbolMap() {
		return new StateChartEditorMap();
	}
	
	public SemanticProcessor getSemanticProcessor() {
		return new CH.ifa.draw.samples.statechart.semantics.StateChartBuilder(fAnalysisContext);
	} 
	
	protected void preParseActions() throws AnalysisException {
		fMachine = null;
		fMap = null;
		try {
			fAnalysisContext = (Context)Class.forName(fContextClass).newInstance();
		}
		catch (ClassNotFoundException e) {
			throw(new AnalysisException("\n  "+e.toString()));
		} 
		catch (InstantiationException e) {
			throw(new AnalysisException("\n  "+e.toString()));
		} 
		catch (IllegalAccessException e) {
			throw(new AnalysisException("\n  "+e.toString()));
		}
	}
	
	protected Vector getErrors(Vector flaws) {
		Vector errors = new Vector();
		Enumeration f = flaws.elements();
		while (f.hasMoreElements()) {
			Flaw flaw = (Flaw)f.nextElement();
			if (Severity.ERROR(flaw.getSeverity()))
				errors.addElement(
					new StateChartError(flaw,fMap)
				);
		}
		return errors;
	}
	
	protected Vector getWarnings(Vector flaws) {
		Vector warnings = new Vector();
		Enumeration f = flaws.elements();
		while (f.hasMoreElements()) {
			Flaw flaw = (Flaw)f.nextElement();
			if (Severity.WARNING(flaw.getSeverity()))
				warnings.addElement(
					new StateChartError(flaw,fMap)
				);
		}
		return warnings;
	}

	protected ProcessingResult postParseActions(ParseResult result) {
		if (result.success()) {
			SemanticProcessor pr = getSemanticProcessor();
			fMachine =
				(StateMachine)
					pr.process(
						((InActiveState)result.parseTrees().start().get()).getSemantics()
					);
			fMap = ((StateChartBuilder)pr).getSemanticMap();
			Vector flaws = fMachine.checkSemantics();
			Vector errors = getErrors(flaws);
			Vector warnings = getWarnings(flaws);
			if (errors.isEmpty()) 
				return new Pass(fMachine,warnings);
			else 
				return new Fail(fMachine,errors,warnings);
		}
		else
			return null;	
	}

	protected void initRelations(IndexedMDSet input) {
		input.addRelation(
			new ImplicitExpanderRelation("inside","inside of","containing")
		);	
		input.addRelation(
			new ImplicitExpanderRelation("enters","entering","entered by")
		);	
		input.addRelation(
			new ImplicitExpanderRelation("exits","exiting","exited by")
		);	
		input.addRelation(
			new ImplicitExpanderRelation("attached","attached to","labelled by")
		);	
	}	
	
	protected Object getStartObject(IndexedMDSet input,ObjectToSymbolMap map) {
		return getDrawing();
	}

	public BinaryPredicate above() {
		return new BinaryPredicate() {
			public boolean execute(Object f1,Object f2) {
				return 
					((Figure)f1).displayBox().getLocation().y < ((Figure)f2).displayBox().getLocation().y;
			} 
		};
	}

	public BinaryPredicate left() {
		return new BinaryPredicate() {
			public boolean execute(Object f1,Object f2) {
				return ((Figure)f1).displayBox().getLocation().x < ((Figure)f2).displayBox().getLocation().x;
			} 
		};
	}
	
}