package CH.ifa.draw.samples.statechart.semantics;

import com.objectspace.jgl.*;
import relap.Editor.*;
import relap.Analyzer.SyntaxTreeNode;
import relap.Analyzer.MiddleNode;
import relap.Analyzer.LeafNode;
import relap.Analyzer.EmptyNode;
import relap.Util.*;
import java.util.*;
import CH.ifa.draw.samples.statechart.*;
import CH.ifa.draw.figures.TextFigure;

public class StateChartBuilder
	implements StateChartVisitor,SemanticProcessor {

	SList fTransitions;
	SList fStates;
	StateMachine fMachine;
	SemanticMap fStructureMap;
	Context fProcessingContext;
	
	public StateChartBuilder(Context context) {
		fTransitions = new SList();
		fStates = new SList();
		fStructureMap = new SemanticMap();
		fMachine = null;
		fProcessingContext = context;
	}
	
	public Context getProcessingContext() {
		return fProcessingContext;
	}
	
	public void setProcessingContext(Context context) {
		fProcessingContext = context;
	}

	public SemanticMap getSemanticMap() {
		return fStructureMap;
	}
	
	protected void addState(State state) {
		fStates.pushBack(state);
		state.setId(fStates.size());
	}
	
	public void visitTransitionRep(TransitionRep node) {
		Transition t = new Transition();
		if (node.getLabel() != null) {
			node.getLabel().accept(this);
			String label = 
				(String)fStructureMap.getSemantic(node.getLabel().valueOfAttribute("root"))
					.nextElement();
			if (label.indexOf('/') > -1) {
				t.setTrigger(
					label.substring(0,label.indexOf('/'))
		 		);
		 		t.setEffect(
					label.substring(label.indexOf('/')+1)
		 		);
			}
			else {
				t.setTrigger(label);
				t.setEffect("");	
			}
			fStructureMap.remove(node.getLabel().valueOfAttribute("root"),label);
			if (!t.getTrigger().equals(""))
				fStructureMap.add(node.getLabel().valueOfAttribute("root"),t.getTrigger());
			if (!t.getEffect().equals(""))
				fStructureMap.add(node.getLabel().valueOfAttribute("root"),t.getEffect());
		} 
		else { 
			t.setTrigger("");
			t.setEffect("");	
		}		 	
		fTransitions.pushBack(t);
		fStructureMap.add(
			node.valueOfAttribute("root"),t
		);
		//System.err.println("t: "+t);
	}
	
	
	protected void linkStates() {
		for (ForwardIterator trans = fTransitions.start();!trans.atEnd();
		     trans.advance()) {
		 	Transition t = (Transition)trans.get();
		 	CH.ifa.draw.samples.statechart.Transition tRep = 
		 		(CH.ifa.draw.samples.statechart.Transition)fStructureMap.getGraphic(t).nextElement();
		 	if (!fStructureMap.getSemantic(tRep.startFigure()).hasMoreElements())
		 		t.setStart(tRep.startFigure());
		 	else
		 		t.setStart(
		 			fStructureMap.getSemantic(tRep.startFigure()).nextElement()
		 		);   
		 	if (!fStructureMap.getSemantic(tRep.endFigure()).hasMoreElements())
		 		t.setEnd(tRep.endFigure());
		 	else
		 		t.setEnd(
		 			fStructureMap.getSemantic(tRep.endFigure()).nextElement()
		 		); 
		 	if (t.getStart() instanceof DirectedGraphNode) 
		 		((DirectedGraphNode)t.getStart()).addOutEdge(t); 
		 	if (t.getEnd() instanceof DirectedGraphNode) 
		 		((DirectedGraphNode)t.getEnd()).addInEdge(t); 
		}
	}
	
	
	protected void setInternals(State state,Object key) {
		Enumeration t = fStructureMap.getSemantic(key);
		SList interns = new SList();
		while (t.hasMoreElements())
			interns.pushBack(t.nextElement());
		state.addInternals(interns);
		t = interns.elements();
		while (t.hasMoreElements())
			fStructureMap.remove(key,t.nextElement());	
	}
	
	protected CompositeState buildCompositeState(boolean concurrent,SList compartments) {
		CompositeState state = new CompositeState();
		state.setIsConcurrent(concurrent);
		Enumeration c = compartments.elements();
		while (c.hasMoreElements()) {
			StateContainer cRep = (StateContainer)c.nextElement();
				//fStructureMap.remove(cRep.valueOfAttribute("root")); // internals
			CompositeState container = state;
			if (concurrent) {
				container = new CompositeState();
				container.setIsConcurrent(false);
				container.setContainer(state);	
				addState(container);
				state.addSubState(container);
			}
			Enumeration substate = cRep.getContainedStates().elements();
			while (substate.hasMoreElements()) {
				MiddleNode subRep = (MiddleNode)substate.nextElement();
				State sub = 
					(State)fStructureMap.getSemantic(subRep.valueOfAttribute("root")).nextElement();
				sub.setContainer(container);
				container.addSubState(sub);
			}
			setInternals(container,((MiddleNode)cRep).valueOfAttribute("root")); // needs to be done BEFORE...
			if (concurrent)
				fStructureMap.add(((MiddleNode)cRep).valueOfAttribute("root"),container); // ...this
		}
		return state;	
	}
	
	
	protected SimpleState buildSimpleState(SList compartments) {
		SimpleState state = new SimpleState();
			// handle internals
		if (!compartments.isEmpty())
			setInternals(
				state,
				((StateCompartmentRep)compartments.front()).valueOfAttribute("root")
			);
		return state;
	}
	
	public void visitStateRep(StateRep node) {
		
		node.getStateSymbol().accept(this);
		SList arcs = node.getTransitions();
		if (arcs != null) {
			Enumeration arc = arcs.elements();
			while (arc.hasMoreElements())
				((TransitionRep)arc.nextElement()).accept(this);
		}
				
		State created = null;
		SList compartments = node.getStateSymbol().getStateCompartments();
		if (compartments == null)
			compartments = new SList();
		boolean concurrent = false;
		if (compartments.size()>1)
			concurrent = true;
		boolean composite = 	
			concurrent || 
			(compartments.size() ==1 && compartments.front() instanceof CompositeStateCompartmentRep);
		if (composite) 
			created = buildCompositeState(concurrent,compartments);
		else 
			created = buildSimpleState(compartments);
		addState(created);
		if (node.getStateSymbol().getNameCompartment() != null)
			created.setName(
				(String)
					fStructureMap.getSemantic(
						node.getStateSymbol().getNameCompartment().valueOfAttribute("root")
					).nextElement()
			);
		if (node.getStateSymbol().getITCompartment() != null)
			setInternals(
				created,
				node.getStateSymbol().getITCompartment().valueOfAttribute("root")
			);	
		fStructureMap.add(
			node.valueOfAttribute("root"),created
		);
		//System.err.println("s: "+created);
	}
	
	public void visitInitialRep(InitialRep node) {
		
		//node.getTransition().accept(this);
		Initial init = new Initial();
		addState(init);
		init.setName(String.valueOf(fStates.size()));
		fStructureMap.add(
			node.valueOfAttribute("root"),
			init
		);
		//System.err.println(init);
	}
	
	public void visitFinalRep(FinalRep node) {
		SList arcs = node.getTransitions();
		for (ForwardIterator arc = arcs.start();!arc.atEnd();arc.advance())
			((TransitionRep)arc.get()).accept(this);
		Final fin = new Final();
		addState(fin);
		fin.setName(String.valueOf(fStates.size()));
		fStructureMap.add(
			node.valueOfAttribute("root"),
			fin
		);
		//System.err.println(fin);
	}
	
	public void visitStateChartRep(StateChartRep node) {
		
		CompositeState top = new CompositeState();
		node.getInitial().accept(this);
		SList states = node.getStates();
		if (states != null)
			for (ForwardIterator state = states.start();!state.atEnd();state.advance()) 
				((StateRep)state.get()).accept(this);
		SList finals = node.getFinals();
		if (finals != null)
			for (ForwardIterator fin = finals.start();!fin.atEnd();fin.advance()) 
				((FinalRep)fin.get()).accept(this);
		SList allStates = node.getContainedStates();
		for (ForwardIterator state = allStates.start();!state.atEnd();state.advance()) {
			State sub = 
					(State)fStructureMap.getSemantic(((MiddleNode)state.get()).valueOfAttribute("root")).nextElement();
			sub.setContainer(top);
			top.addSubState(sub);
		}

		top.setIsConcurrent(false);
		addState(top);
		fStructureMap.add(node.getDiagramContainer().getGraphic(),top);
		fMachine.setTopState(top);
		//System.err.println(top);
	}
	
	/*public void visitStateMachineRep(StateMachineRep node) {
		node.getInitial().accept(this);
		Enumeration item = node.getStateList().elements();
		while (item.hasMoreElements())
			((StateChartNode)item.nextElement()).accept(this);
		item = node.getFinalList().elements();
		while (item.hasMoreElements())
			((StateChartNode)item.nextElement()).accept(this);		
		StateRep top = new StateRep();
		fStructureMap.add(
			top.valueOfAttribute("root"),top
		);
		node.setTopState(top);
	}*/

	
	public void visitStateSymbolRep(StateSymbolRep node) {
		if (node.getNameCompartment() != null)
			node.getNameCompartment().accept(this);
		if (node.getITCompartment() != null)
			node.getITCompartment().accept(this);
		SList compartments;
		if ((compartments=node.getStateCompartments()) != null) {
			Enumeration t = compartments.elements();
			while (t.hasMoreElements())
				((StateChartNode)t.nextElement()).accept(this);		
		}
	}
	
	public void visitNameCompartmentRep(NameCompartmentRep node) {
		String name = new String(""); // need an unique object
		if (node.getName()!=null) {
			StringBuffer buf = new StringBuffer("");
			Enumeration part = node.getName().elements();
			while (part.hasMoreElements()) {
				buf.append(
					((TextFigure)((LeafNode)part.nextElement()).getGraphic()).getText()
				);
				if (part.hasMoreElements())
					buf.append(" ");
			}
			name = buf.toString().trim();
		}
		fStructureMap.add(
			node.getPanel().getGraphic(),
			name
		);	
		//System.err.println("nc: "+name);	
	}
	
	protected void processInternals(SList internals,MiddleNode node) {
		if (internals!=null) {
			Enumeration it = internals.elements();
			while (it.hasMoreElements()) {
				TextFigure text = (TextFigure)((LeafNode)it.nextElement()).getGraphic();
				String rep = text.getText();
				InternalTransition t = new InternalTransition();
				if (rep.indexOf('/') > -1) {
					t.setTrigger(
						rep.substring(0,rep.indexOf('/'))
		 			);
		 			t.setEffect(
						rep.substring(rep.indexOf('/')+1)
		 			);
				}
				else {
					t.setTrigger(rep);
					t.setEffect("");	
				}
				if (!t.getTrigger().equals(""))
					fStructureMap.add(text,t.getTrigger());
				if (!t.getEffect().equals(""))
					fStructureMap.add(text,t.getEffect());
				fStructureMap.add(node.valueOfAttribute("root"),t);
				//System.err.println("it: "+t);
			}
		}	
	}

	public void visitITCompartmentRep(ITCompartmentRep node) {
		//SList internals;
		processInternals(node.getInternals(),node);
	}

	public void visitStateCompartmentRep(StateCompartmentRep node) {
		//fStructureMap.add(
		//	node.getPanel().getGraphic(),
		//	node
		//);	
		processInternals(node.getInternalTransitions(),node);
	}
	
	public void visitCompositeStateCompartmentRep(CompositeStateCompartmentRep node) {
		//fStructureMap.add(
		//	node.getPanel().getGraphic(),
		//	node
		//);	
		processInternals(node.getInternals(),node);
		if (node.getInitial() != null)
			node.getInitial().accept(this);
		Enumeration s = node.getStates().elements();
		while (s.hasMoreElements()) 
			((StateRep)s.nextElement()).accept(this);		
		SList finals = node.getFinals();
		if (finals != null)
			for (ForwardIterator fin = finals.start();!fin.atEnd();fin.advance())
				((FinalRep)fin.get()).accept(this);
	}
	
	public void visitLabelRep(LabelRep node) {
		StringBuffer buf = new StringBuffer();
		Enumeration t = node.getLabelTexts().elements();
		while (t.hasMoreElements()) {
			LeafNode it = (LeafNode)t.nextElement();
			buf.append(((TextFigure)it.getGraphic()).getText());
			if (t.hasMoreElements())
				buf.append(" "); 
		}		
		fStructureMap.add(node.valueOfAttribute("root"),buf.toString());		
	}

	
	public Object process(SyntaxTreeNode syntaxTree) {
		System.err.println(syntaxTree);
		StateChartRep stch = (StateChartRep)syntaxTree;
		fMachine = new StateMachine(fProcessingContext);
		stch.accept(this);
		linkStates();
		fMachine.setTransitions(fTransitions);
		fMachine.setStates(fStates);
		fMachine.getTopState().setStateMachine(fMachine);
		//System.err.println(stch.getDescription());
		//System.err.println(fStructureMap);
		return fMachine;
		//return new TreePrinter().process(syntaxTree);
	}
}