package CH.ifa.draw.samples.statechart.semantics;

import com.objectspace.jgl.SList;
import com.objectspace.jgl.ForwardIterator;
import com.objectspace.jgl.HashMap;
import com.objectspace.jgl.HashSet;
import CH.ifa.draw.framework.*;
import relap.Analyzer.ProcessingResult;
import java.util.Vector;

public class StateMachine {

	protected SList fTransitions;
	protected SList fStates;
	protected State fTopState;
	protected Context fContext;
	protected Vector fReservedActionLabels;
	
	public StateMachine() {
		fTransitions = null;
		fStates = null;
		fTopState = null;
		fContext = null;
		fReservedActionLabels = createActionLabels();
	}
	
	public StateMachine(Context context) {
		fTransitions = null;
		fStates = null;
		fTopState = null;
		fContext = context;
		fReservedActionLabels = createActionLabels();
	}

	protected Vector createActionLabels() {
		Vector labels = new Vector();
		labels.addElement("entry");
		labels.addElement("exit");
		labels.addElement("do");
		labels.addElement("include");
		return labels;
	}

	public Vector getReservedActionLabels() {
		return (Vector)fReservedActionLabels.clone();
	}
	 
	public void setContext(Context context) {
		fContext = context;
	}
	
	public Context getContext() {
		return fContext;
	}

	public void setTopState(State top) {
		fTopState = top;
	}
	
	public State getTopState() {
		return fTopState;
	}
	
	public SList getTransitions() {
		return new SList(fTransitions);
	}
	
	public void setTransitions(SList tList) {
		fTransitions = new SList(tList);
	}
	

	public SList getStates() {
		return new SList(fStates);
	}
	
	public void setStates(SList sList) {
		fStates = new SList(sList);
	}
	
	/*public StateRep getLeastCommonAncestor(Transition transition) {
		return 
			(StateRep)fStructureIndex.get(
				((Figure)transition.getArrow().getGraphic()).parent()
			);
	}*/

	protected void checkContext(Vector accuErrors) {
		Vector events = fContext.getEventNames();  	
		for (int i=0;i<events.size();i++)
			if (fReservedActionLabels.contains(events.elementAt(i)))
				accuErrors.addElement(
					new Flaw(
						"Error in context - invalid event name: "+events.elementAt(i),
						new HashSet(),
						Severity.createError()
					)
				);	
	}
			

	public Vector checkSemantics() {
		Vector errors = new Vector();
		checkContext(errors);
		getTopState().check(errors);
		for (ForwardIterator trn=getTransitions().start();!trn.atEnd();trn.advance()) 
			((Transition)trn.get()).check(errors);
		return errors;
	}
		
	public String toString() {
		StringBuffer buf = new StringBuffer("State machine: \n");
		//buf.append(getInitial().getDescription()+"\n");
		buf.append("******* Top State ********\n"+fTopState);
		//buf.append(getFinal().getDescription());
		return buf.toString();
	}
}