package CH.ifa.draw.samples.statechart.semantics;

import relap.Analyzer.MiddleNode;
import com.objectspace.jgl.*;
import relap.Util.*;
import java.util.Enumeration;

abstract public class StateVertex extends StateChartObject
	implements DirectedGraphNode {

	protected DirectedGraphNodeImpl fDiGN;
	protected int fDepth;
	protected CompositeState fContainer;
	protected String fName;
	protected SList fAliases;
	protected long fId;
	
	public StateVertex() {
		fDiGN = new DirectedGraphNodeImpl();
		fDepth = -1;
		fContainer = null;
		fName = null;
		fAliases = null;
	}

	public void setContainer(CompositeState container) {
		fContainer = container;
	}

	public CompositeState getContainer() {
		return fContainer;
	}

	public CompositeState getRegion() {
		if (fContainer == null)
			return null;
		if (fContainer.isRegion())
			return fContainer;
		else 
			return fContainer.getRegion();
	}
	
	public int getNestingDepth() {
		if (fDepth < 0) {
			if (fContainer == null)
				fDepth = 0;
			else
				fDepth = fContainer.getNestingDepth() + 1;
		}
		return fDepth;
	}
	
	public void setName(String name) {
		fName = name;
	}
	
	public String getName() {
		if (fName == null)
			fName = "";
		return	fName;
	}

	public void addAlias(String alias) {
		if (fAliases == null)
			fAliases = new SList();
		fAliases.pushBack(alias);
	}
	
	public SList getAliases() {
		if (fAliases == null)
			return new SList();
		return new SList(fAliases);
	}
	
	public void setId(long id) {
		fId = id;
	}
	
	public long getId() {
		return fId;
	}

	public ForwardIterator getIncoming() {
		return getInEdges();
	}
	
	public ForwardIterator getOutgoing() {
		return getOutEdges();
	}
	
 	public Edge addEdge(Edge edge) {
 		return fDiGN.addEdge(edge);
 	}
	
 	public Edge changeEdge(Edge oldEdge,Edge newEdge) {
 		return fDiGN.changeEdge(oldEdge, newEdge);
 	}

	public Edge removeEdge(Edge edge) {
 		return fDiGN.removeEdge(edge);
 	}

	public ForwardIterator getEdges() {
 		return fDiGN.getEdges();
 	}

 	public Edge addOutEdge(Edge edge) {
 		return fDiGN.addOutEdge(edge);
 	}

 	public Edge addInEdge(Edge edge) {
 		return fDiGN.addInEdge(edge);
 	}

	public ForwardIterator getOutEdges() {
 		return fDiGN.getOutEdges();
 	}

	public ForwardIterator getInEdges() {
 		return fDiGN.getInEdges();
 	}

	public String inset() {
		String filler = "";
		for (int i = getNestingDepth();i>0;i--)
			filler += "  ";
		return filler;
	}
	
	protected String describeExitingTransitions() {
		ForwardIterator t = getOutEdges();
		if (t.atEnd())
			return inset()+"<no transitions from this state>";
		StringBuffer buf = new StringBuffer(inset()+"transitions:\n");
		while (!t.atEnd()) {
			buf.append(inset()+t.get());
			t.advance();
			if (!t.atEnd())
				buf.append("\n");
		}
		return buf.toString();
	}	
		
}