

package CH.ifa.draw.standard;

import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.io.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.util.*;
import CH.ifa.draw.figures.*;

public class ConnectedText extends TextFigure {

  public ConnectedText() {
    super();
  }

    public void connect(Figure figure) {
        if (fObservedFigure != null) {
            fObservedFigure.removeFigureChangeListener(this);
            if (fObservedFigure instanceof LabelledConnection)
              ((LabelledConnection)fObservedFigure).removeConnected(this);
        }

        fObservedFigure = figure;
        fLocator = new OffsetLocator(figure.connectedTextLocator(this));
        fObservedFigure.addFigureChangeListener(this);
        if (fObservedFigure instanceof LabelledConnection)
          ((LabelledConnection)fObservedFigure).addConnected(this);
        updateLocation();
    }

    /**
     * Disconnects the text figure.
     */
    public void disconnect() {
        System.err.println("disconnecting "+this);
        fObservedFigure.removeFigureChangeListener(this);
        if (fObservedFigure instanceof LabelledConnection)
          ((LabelledConnection)fObservedFigure).removeConnected(this);
        fObservedFigure = null;
        fLocator = null;
    }

    /**
     * Disconnects this.
     */
    public void dragged(MouseEvent e) {
      super.dragged(e);
      if (isConnected() && e!=null && !e.isControlDown())
        disconnect();
    }

    /**
     * Disconnects this.
     */
    public void endDrag(MouseEvent e) {
      if (e != null)
        System.err.println("ending drag "+this+" "+e.isControlDown());
      if (isConnected() && e != null && !e.isControlDown() && fHasMoved)
        disconnect();
      else if (isConnected() && e != null && e.isControlDown() && fHasMoved) {
        Locator l = fObservedFigure.connectedTextLocator(this);
        if (l != null) {
          fLocator = new OffsetLocator(l);
          updateLocation();
        } else
          disconnect();
      }
      super.endDrag(e);
    }

  public void draw(Graphics g) {
    	Shape clip = g.getClip();
    	//System.err.println("clip "+clip);
    	Rectangle r = displayBox();
      //System.err.println("r "+r);
      r.grow(20,20);
    	g.setClip(r);
    	//System.err.println("newclip: "+g.getClip());
    	super.draw(g);
      if (inTransit() && isConnected())
        drawCrossHairs(g);
      g.setClip(clip);

  }
  
    private void drawCrossHairs(Graphics g) {
      Rectangle r = displayBox();
      g.drawLine(r.x,r.y+r.height/2,r.x+19,r.y+r.height/2);
    }
}