

package CH.ifa.draw.standard;

import java.awt.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.Geom;

/**
 * A SegmentLocator defines some basic functionality for locators that are
 * associated to a segment of a polyline rather than just the display box
 * of the polyline.
 */
abstract public class SegmentLocator extends AbstractLocator {
    protected int fSegment;
    protected Figure fAttachment;

    /**
     * return the segment of pLine that is closest to the figure attachment.
     */
    public static int findSegment(PolyLineFigure pLine,Figure attachment) {
        Point p1, p2;
        int dist = 50;
        Rectangle aBox = attachment.displayBox();
        Point h1 = new Point(aBox.x-dist,aBox.y+aBox.height/2);
        Point h2 = new Point(aBox.x+aBox.width+dist,aBox.y+aBox.height/2);
        Point v1 = new Point(aBox.x+aBox.width/2,aBox.y-dist);
        Point v2 = new Point(aBox.x+aBox.width/2,aBox.y+aBox.height+dist);
        ClosestSegment finder =
          new ClosestSegment(new Point(aBox.x+aBox.width,aBox.y+aBox.height));
        for (int i = 0; i < pLine.pointCount()-1; i++) {
            p1 = pLine.pointAt(i);
            p2 = pLine.pointAt(i+1);
            finder.testSegment(h1,h2,p1,p2,i);
            finder.testSegment(v1,v2,p1,p2,i);
        }
        return finder.getClosest();

    }

    public SegmentLocator(PolyLineFigure pLine,Figure attachment) {
      Rectangle dBox = attachment.displayBox();
      fAttachment = attachment;
      fSegment = pLine.findSegment(dBox.x + dBox.width/2,dBox.y + dBox.height/2);
    }

    public SegmentLocator(PolyLineFigure pLine,Figure attachment,int segment) {
      Rectangle dBox = attachment.displayBox();
      fAttachment = attachment;
      fSegment = segment;
    }

    /**
     * Make sure that fSegment is in range [0,pointCount()-2].
     */
    protected void updateSegment(PolyLineFigure pLine) {
      fSegment = (fSegment == -1) ? 0 : fSegment;
      fSegment = (fSegment >= pLine.pointCount()-1) ? pLine.pointCount() - 2 : fSegment;
    }

    /**
     * Find a location for the located figure relative to the figure owner.
     */
    public Point locate(Figure owner) {
      updateSegment((PolyLineFigure)owner);
      Point p1 = ((PolyLineFigure)owner).pointAt(fSegment);
      Point p2 = ((PolyLineFigure)owner).pointAt(fSegment+1);
      return findLocation(p1,p2);
    }

    abstract protected Point findLocation(Point p1,Point p2);

    /**
     * Returns the located figure.
     */
    public Figure getLocatedFigure() {
      return fAttachment;
    }
}

    class ClosestSegment {

      int fSegment;
      long fDistance;
      Point fRefPoint;

      ClosestSegment(Point refPoint) {
        fSegment = -1;
        fDistance = Long.MAX_VALUE;
        fRefPoint = refPoint;
      }

      int getClosest() {
        return fSegment;
      }

      void testSegment(Point l1,Point l2,Point s1,Point s2,int currSegment) {
        Point isec =
          Geom.intersect(l1.x,l1.y,l2.x,l2.y,s1.x,s1.y,s2.x,s2.y);
        if (isec != null) {
          long d = Geom.length(fRefPoint.x,fRefPoint.y,isec.x,isec.y);
          if (d < fDistance) {
            fDistance = d;
            fSegment = currSegment;
          }
        }
      }

    }

