package relap.Analyzer;

import relap.LanguageModel.*;
import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Applying;
import java.io.*;
import java.util.Enumeration;

public class ARGRecognizer {

	protected RelationalGrammarImplementation _grammar;
	protected Deque _agenda;
	protected ParseTable _parseTable;
	protected PredictionTable _predictionTable;
	protected IndexedMDSet _input;
	protected Array _roots; 
	protected Array _errors;
	protected int _comparisons;
	protected Object startObject;
	protected Object currentKey;
	protected String parseAction;
	
	public ARGRecognizer() {
		_grammar = null;
		_agenda = null;
		_parseTable = null;
		_predictionTable = null;
		_input = null;
		_roots = null;
		_errors = null;
	}

	public RelationalGrammarImplementation getGrammar() {
		return _grammar;
	}
	
	public IndexedMDSet getInput() {
		return _input;
	}
	
	
	protected AgendaItem closeParse(ActiveState a_state,InActiveState i_state)
	throws QueryException {
		/*
		AgendaItem new_item = null;
			BranchingState new_state =
				new BranchingState(a_state.productionVariant().production().lhs());
			new_state.productionVariant(a_state.productionVariant());
			DescendantList descendants = new DescendantList(a_state.subTrees());
			descendants.addDescendant(a_state.dot(),i_state);
			new_state.inversePermuteDescendants(descendants);
			HashSet cover = new HashSet(a_state.cover());
			Copying.copy(i_state.cover(),cover);
			new_state.cover(cover);
			new_state.computeAttributeValues(a_state,i_state);
			new_state.previous(a_state);
			new_state.creator(i_state.creator());
			new_state.originator(a_state.originator());
			new_state.addRelationTuples(a_state.relationTuples());
			new_state.addToMask(a_state.inversePermutedMask());
			if (evaluatePredicates(a_state,new_state)) {
				//a_state.next(new_state);
				new_item = new AgendaItem(new_state);
				new_item.keys(new HashSet(new_state.indices()));
				handleSemantics(new_state);
			}
			return new_item;	
		*/	
		SList l = new SList();
		l.pushBack(a_state);
		return closeParse(a_state,l,i_state);
		
	}
	
	protected AgendaItem closeParse(ActiveState a_state,SList prevStates,InActiveState i_state)
	throws QueryException {
		AgendaItem new_item = null;
			BranchingState new_state =
				new BranchingState(a_state.productionVariant().production().lhs());
			new_state.productionVariant(a_state.productionVariant());
			DescendantList descendants = new DescendantList(a_state.subTrees());
			descendants.addDescendant(a_state.dot(),i_state);
			new_state.inversePermuteDescendants(descendants);
			HashSet cover = new HashSet(a_state.cover());
			Copying.copy(i_state.cover(),cover);
			new_state.cover(cover);
			new_state.computeAttributeValues(a_state,i_state);
			Enumeration pState = prevStates.elements();
			while (pState.hasMoreElements())
				new_state.previous((ParseState)pState.nextElement());
			new_state.creator(i_state.creator());
			new_state.originator(a_state.originator());
			new_state.addRelationTuples(a_state.relationTuples());
			new_state.addToMask(a_state.inversePermutedMask());
			if (evaluatePredicates(a_state,new_state)) {
				//a_state.next(new_state);
				new_item = new AgendaItem(new_state);
				new_item.keys(new HashSet(new_state.indices()));
				handleSemantics(new_state);
			}
			return new_item;				
	}

	protected AgendaItem consumePendingInput(ActiveState a_state,InActiveState i_state,
																						ActiveState new_state)
	throws QueryException {
		AgendaItem new_item = null;
		SList pendingObjects = new SList(a_state.getPendingInput());
				//System.err.println("a_state: "+a_state);
				//System.err.println("pendingObjects: "+pendingObjects);
		HashSet resultSet = new HashSet();
		resultSet.add(pendingObjects.popFront());
		new_state.setPendingInput(pendingObjects); 
		new_state.dot(a_state.dot());
		DescendantList sub_trees = new DescendantList(a_state.subTrees());
		sub_trees.addDescendant(new_state.dot(),i_state);
		new_state.subTrees(sub_trees);
		new_state.previous(a_state);
		new_state.originator(a_state.originator());
		new_state.creator(a_state.creator());
		new_state.addRelationTuples(a_state.relationTuples());
		new_state.addToMask(a_state.mask());
		new_state.queryConstraint(a_state.queryConstraint());
		if (evaluatePredicates(a_state,new_state)) {
					//System.err.println("new_state: "+new_state);
			new_item = new AgendaItem(new_state);
			new_item.keys(resultSet);
		}
		return new_item;
	}
	
	protected AgendaItem queryForInput(ActiveState a_state,InActiveState i_state,
																		  ActiveState new_state)
	throws QueryException {
		AgendaItem new_item = null;
		/*
		new_state.dot(a_state.dot() + 1);
		DescendantList sub_trees = new DescendantList(a_state.subTrees());
		sub_trees.addDescendant(a_state.dot(),i_state);
		new_state.subTrees(sub_trees);
		new_state.previous(a_state);
		new_state.originator(a_state.originator());
		new_state.creator(a_state.creator());
		new_state.addRelationTuples(a_state.relationTuples());
		new_state.addToMask(a_state.mask());
		*/
		initNewState(a_state,i_state,new_state);
		if (evaluatePredicates(a_state,new_state)) {
			HashSet result_set = queryInput(new_state);
			try {
				result_set = processQueryResult(a_state,i_state,new_state,result_set); 
				if (!result_set.isEmpty()) {
						//a_state.next(new_state);
					new_item = new AgendaItem(new_state);
					new_item.keys(result_set);
				} else if (new_state.productionVariant().isOptional(new_state.dot())) {
					new_item = skipOptionalSymbol(new_state);	
				}
			} catch (QueryResultException e) {
				new_item = null;
			}
		}
		return new_item;
	}
	
	protected void initNewState(ActiveState a_state,InActiveState i_state,
															 ActiveState new_state) {
		/*
		new_state.dot(a_state.dot() + 1);
		DescendantList sub_trees = new DescendantList(a_state.subTrees());
		sub_trees.addDescendant(a_state.dot(),i_state);
		new_state.subTrees(sub_trees);
		new_state.previous(a_state);
		new_state.originator(a_state.originator());
		new_state.creator(a_state.creator());
		new_state.addRelationTuples(a_state.relationTuples());
		new_state.addToMask(a_state.mask());
		*/
		SList l = new SList();
		l.pushBack(a_state);
		initNewState(a_state,l,i_state,new_state);
	}

	protected void initNewState(ActiveState a_state,SList prevStates,InActiveState i_state,
															 ActiveState new_state) {
		new_state.dot(a_state.dot() + 1);
		DescendantList sub_trees = new DescendantList(a_state.subTrees());
		sub_trees.addDescendant(a_state.dot(),i_state);
		new_state.subTrees(sub_trees);
		Enumeration pState = prevStates.elements();
		while (pState.hasMoreElements())
			new_state.previous((ParseState)pState.nextElement());
		new_state.originator(a_state.originator());
		new_state.creator(parseAction);
		new_state.addRelationTuples(a_state.relationTuples());
		new_state.addToMask(a_state.mask());
	}

	protected AgendaItem skipOptionalSymbol(ActiveState state)
	throws QueryException {
		parseAction = "skip";
		AgendaItem new_item = null;
		//SList prevs = new SList(state.previous());
		while (new_item == null && state.productionVariant().isOptional(state.dot())) {
			//System.err.println("@@@@\n"+state);
			state = (ActiveState)_parseTable.addSkipState(currentKey,state);
			state.markRelationTuples();
			EmptyLeaf empty = new EmptyLeaf("optional element not present");
			empty.cover(new HashSet());
			empty.creator("skip");
			handleSemantics(empty);
			state.maskSymbolAt(state.dot()-1); // odd, but this is how it works
			if (state.dotAtEnd()) {
				//return closeParse(state,prevs,empty);
				return closeParse(state,empty);
			}
		
			ActiveState newState = new ActiveState(state.productionVariant());
			//initNewState(state,prevs,empty,newState);
			initNewState(state,empty,newState);
			newState.creator("skip");
			HashSet result_set = queryInput(newState);
			try {
				result_set = processQueryResult(state,empty,newState,result_set);
				if (!result_set.isEmpty()) {
				//state.unLinkFromParse() // if more than one symbol skipped?
					new_item = new AgendaItem(newState);
					new_item.keys(result_set);
				}
			} catch (QueryResultException e) {
				new_item = null;
			}	
			state = newState;
		}
		return new_item;
	}	

	public AgendaItem advance(ActiveState a_state,InActiveState i_state)
	throws QueryException {
		AgendaItem new_item = null;
		//System.err.println("\nAdvancing "+a_state.toString()+"\nwith "+i_state.toString());
		if (a_state.dotAtEnd()) {
			new_item = closeParse(a_state,i_state);
		}
		else {
			ActiveState new_state = new ActiveState(a_state.productionVariant());
			if (a_state.productionVariant().isRepetitive(a_state.dot()) &&
			    !a_state.getPendingInput().isEmpty()) {
			  new_item = consumePendingInput(a_state,i_state,new_state);  
			} else{
				new_item = queryForInput(a_state,i_state,new_state);
			}
		}
		//if (new_item!=null) System.err.println(new_item.toString());
		return new_item;
	}
	
	public ParseTable getParseTable() {
		return _parseTable;
	}
	
	// hook; override in subclasses for creating syntax tree nodes
	protected void handleSemantics(BranchingState newState) {
	}
	
	// hook; override in subclasses for creating syntax tree nodes
	protected void handleSemantics(LeafState newState) {
	}
	
	// hook; override in subclasses for creating syntax tree nodes
	protected void handleSemantics(EmptyLeaf empty) {
	}
	
	protected HashSet queryInput(ActiveState newState)
		throws QueryException {
		HashSet result;

		try {
			setExpanderConstraint(newState);
			result = launchQuery(newState);
		}
		catch (InvalidQueryArgumentsException e) {
			throw(new QueryException(e.getMessage()));
		}
		//System.err.println("queryInput: "+result);
		return result;
	}
	
	// hook - do nothing to the result set
	protected HashSet processQueryResult(ActiveState previousState,
	                                     InActiveState advancedWith,
	                                     ActiveState newState,
	                                     HashSet queryResult)
		throws QueryException,QueryResultException {
		
		return queryResult;
	}                 
		

	String agendaToString() {
		StringBuffer buf = new StringBuffer("\n-- Agenda --");
		for (DequeIterator item=_agenda.begin();!item.atEnd();item.advance())
			buf.append("\n"+item.get().toString());
		buf.append("\n-- end Agenda --");
		return buf.toString();
	}
	
	protected Array buildQueryArguments(ActiveState state,Object input_object,
	                                    Constraint constraint) {
	    Array arguments = new Array(2);                          
	    SList refVal = new SList();                        
		refVal.pushBack(input_object);
		ObjectReference bound_argument = constraint.argumentBefore(state.dot());
		if (bound_argument==constraint.firstArg()) {
			arguments.put(0,state.dereference(bound_argument));
			arguments.put(1,refVal);
		}
		else {
			arguments.put(0,refVal);
			arguments.put(1,state.dereference(bound_argument));		
		}

		return arguments;
	}

	protected Array buildQueryArguments(ActiveState state,InActiveState i_state,
	                                    Constraint constraint) {
	    Array arguments = new Array(2);  
	    SList refVal = new SList();                        
		ObjectReference bound_argument = constraint.argumentBefore(state.dot());
		if (bound_argument==constraint.firstArg()) {
			arguments.put(0,state.dereference(bound_argument));
			refVal.pushBack(i_state.referencedValue(constraint.secondArg()));
			arguments.put(1,refVal);
		}
		else {
			refVal.pushBack(i_state.referencedValue(constraint.firstArg()));
			arguments.put(0,refVal);
			arguments.put(1,state.dereference(bound_argument));		
		}

		return arguments;
	}

	protected Array buildQueryArguments(ActiveState state,
	                                    Constraint constraint) {
	    Array arguments = new Array(2); 
	    SList refVal = new SList();
	    refVal.pushBack(null);                         
		ObjectReference bound_argument = constraint.argumentBefore(state.dot());
		if (bound_argument==constraint.firstArg()) {
			arguments.put(0,state.dereference(bound_argument));
			arguments.put(1,refVal);
		}
		else {
			arguments.put(0,refVal);
			arguments.put(1,state.dereference(bound_argument));		
		}

		return arguments;
	}


	// hook: do nothing
	protected void createCoverError(ActiveState state,Object input_object) {
	}

	// hook: do nothing
	protected void createCoverError(ActiveState state,InActiveState i_state) {
	}

	// hook: do nothing
	protected void createExtraRelationsError(ActiveState a_state,
										     InActiveState i_state,
										     ActiveState new_state,
										     HashSet objects) {
	}
	
	// hook: do nothing
	protected void createMissingRelationError(ActiveState a_state,
										      InActiveState i_state,
										      ActiveState new_state)
	throws QueryException { 
	}

	// hook: do nothing
	protected void createScanningError(ActiveState state,Object input_object) {
	}

	// hook: do nothing
	protected void createUnsatisfiedConstraintsError(ActiveState state,
	               									 Object input_object)
	throws QueryException {
	}

	// hook: do nothing
	protected void createUnsatisfiedConstraintsError(ActiveState state,
	               									 InActiveState i_state)
	throws QueryException {
	}


	protected Deque complete(InActiveState i_state,Object input_object,
	                         Array a_states)
	throws AnalysisException {
			parseAction = "complete";
	    Deque new_items = new Deque();
	    HashSet active_states = new HashSet();
	    HashSet init_states = new HashSet();
	    Copying.copy(_parseTable.activeStatesAt(input_object),active_states);
	   	Copying.copy(a_states,init_states);
		for (HashSetIterator state=active_states.union(init_states).begin();
			 !state.atEnd();state.advance()) {
			 
//System.err.println("\nCompleting "+state.get().toString()+"\nwith "+i_state.toString());
	
			if (i_state.category().equals(((ActiveState)state.get()).symbolAtDot())) {
				if (i_state.intersectionOfCovers((ActiveState)state.get()).isEmpty()) {
					if (satisfiesCurrentConstraints(i_state,(ActiveState)state.get())) {
				// --> satisfiesCurrentConstraints(i_state,state) !!! 
						newItem((ActiveState)state.get(),i_state,new_items,"complete");
					}
					else {
						createUnsatisfiedConstraintsError((ActiveState)state.get(),i_state);
					}
				}
				else {
					createCoverError((ActiveState)state.get(),i_state);
				}
			}
		}		
		return new_items;
	}
	
	// a hook
	protected Array completionInitStates() {
		return _predictionTable.initStates();
	}

	// a hook
	protected HashSet filterCandidates(HashSet candidates,ActiveState new_state) {
		return candidates;
	}
	
	protected HashSet findRoots() {
		HashSet roots = new HashSet();
		for (HashSetIterator object=_input.objects().begin();
		     !object.atEnd();object.advance())
			for (SListIterator state=_parseTable.inActiveStatesAt(object.get()).begin();
				 !state.atEnd();state.advance())
				 if (((InActiveState)state.get()).category().equals(_grammar.startSymbol()) &&
				     ((InActiveState)state.get()).cover().size() == _input.objects().size())
				 	roots.add(state.get());
		return roots;
	}	
	
	// hook
	protected boolean evaluatePredicates(ActiveState prevState,InActiveState newState) 
	throws QueryException {
		return true;
	}

	// hook
	protected boolean evaluatePredicates(ActiveState prevState,ActiveState newState) 
	throws QueryException {
		return true;
	}

	protected Deque handleActiveState(ActiveState state,Object input_object)
	throws AnalysisException {
		Deque new_items = new Deque();

		if (_grammar.isTerminal(state.symbolAtDot())) {
			AgendaItem item = scan(state,input_object);
			if (item != null)
				new_items.pushBack(item);
		}
		else {
			Copying.copy(
				predict(state,input_object),
				new_items
			);
			Copying.copy(
				inverse_complete(state,input_object),
				new_items
			);
		}
		
		return new_items;
	}

	public void initializeAgenda(Array states,Object key) {
		for (ArrayIterator state=states.begin();!state.atEnd();state.advance())
			_agenda.pushBack(new AgendaItem((ParseState)state.get()).addKey(key));
	}
	
	// a hook
	protected void initParse(IndexedMDSet input,
						 	 RelationalGrammarImplementation grammar)
	throws InvalidGrammarException { 
		if (_grammar != grammar) {
			_grammar = grammar;
			_predictionTable = new PredictionTable();
			_predictionTable.buildWith(_grammar);
		}
		_input = input;
		_parseTable = new ParseTable(_input.objects());
		_agenda = new Deque();
		_errors = new Array();
		currentKey = null;
		parseAction = "init";
	}
	
	// a hook
	protected Array initStates() {
		return _predictionTable.initStates();
	}
	
	protected Array inverse_complete(ActiveState state,Object input_object)
	throws AnalysisException {
		parseAction = "inverse-complete";
		Array new_items = new Array();
		for (SListIterator i_state=_parseTable.inActiveStatesAt(input_object).begin();
			 !i_state.atEnd();i_state.advance()) {
			if (((InActiveState)i_state.get()).category().equals(state.symbolAtDot())) 
				if (((InActiveState)i_state.get()).intersectionOfCovers(state).isEmpty()) { 
					if (satisfiesCurrentConstraints((InActiveState)i_state.get(),state))
				// --> satisfiesCurrentConstraints(i_state,state) !!! 
						newItem(state,(InActiveState)i_state.get(),new_items,
						        "inverse-complete");
					else
						createUnsatisfiedConstraintsError(state,(InActiveState)i_state.get());
				} 
				else
					createCoverError(state,(InActiveState)i_state.get());
		}
		return new_items;
	}


	protected void setExpanderConstraint(ActiveState state) {
		// mask masked symbols out of constraint selection!!
		state.queryConstraint(state.productionVariant().constraintAt(state.dot()));
	}

	protected HashSet launchQuery(ActiveState state)
	throws InvalidQueryArgumentsException {
		if (state.queryConstraint() == null)		
			return new HashSet();
		return 
			_input.getRelation(state.queryConstraint().relation()).query(
				buildQueryArguments(state,state.queryConstraint())
			);
	}

	
	protected void newItem(ActiveState a_state, InActiveState i_state,
	                       Sequence new_items,String creator)
	throws AnalysisException {
		AgendaItem item = advance(a_state,i_state);
		if (item != null) {	
			new_items.pushBack(item);
			item.state().creator(creator);
		}
	}
	
	public ParseResult parse(IndexedMDSet input,
						 	 RelationalGrammarImplementation grammar,
	                     	 Object first_object,PrintWriter log,
	                     	 int max_iterations)
	throws InvalidGrammarException,AnalysisException {
		int counter = 0;
		_comparisons = 0;
		int max_error_count = 10;
		int error_count = 0;
		startObject = first_object;
		ParseResult result;
		initParse(input,grammar);

		log.println(_predictionTable.toString());

		initializeAgenda(initStates(),first_object);

		log.println(agendaToString());
		
		result = null;
		counter = processInput(max_iterations);
		result = getResult();
		log.println(_parseTable.toString());
		log.println("** "+(new Integer(counter)).toString()+" iterations **");
		log.println("** "+(new Integer(_comparisons)).toString()+" comparisons **");

		return result;
	}	
	
	protected HashSet getPredictionAttributes(ActiveState state) {
		return state.toBeBoundAttrs();
	}
	
	protected Array predict(ActiveState a_state,Object input_object) {
		parseAction = "predict";
		Array new_items = new Array();
		HashSet predictionAttributes = getPredictionAttributes(a_state);
		for (HashSetIterator attr=predictionAttributes.begin();
			 !attr.atEnd();attr.advance()) 
			for (ArrayIterator state=
					_predictionTable.statesFor(a_state.symbolAtDot(),(String)attr.get()).begin();
				 !state.atEnd();state.advance()) {
				((ActiveState)state.get()).creator("predict");
				((ActiveState)state.get()).up(a_state);
				((ActiveState)state.get()).originator(a_state);
				new_items.pushBack(
					new AgendaItem((ActiveState)state.get())
				);
				((AgendaItem)new_items.back()).addKey(input_object);
				// System.err.println("predict: "+attr.get()+" "+state.get());
			}
		if (new_items.isEmpty())
			System.err.println("Prediction failed at :\n"+a_state.toString());
		return new_items;
	}
	
	protected void handlePotentialAmbiguity(ParseState newState,SList oldStates) {
		//hook; do nothing
	}
	
	protected int processInput(int max_iterations)
	throws AnalysisException {
		int counter = 0;
		AgendaItem item;
		ParseState new_state;
		while (!_agenda.isEmpty() && counter <= max_iterations) {
			//System.err.println(_parseTable);
			counter++;
			item = (AgendaItem)_agenda.popFront();
			for (HashSetIterator key=item.keys().begin();
			     !key.atEnd();key.advance()) {
				//System.err.println("item: "+item);
			    _comparisons +=_parseTable.statesAt(key.get()).size(); 
			    SList slot = _parseTable.statesAt(key.get());
			  currentKey = key.get();
				if (!slot.contains(item.state())) {				
//				if (!_parseTable.statesAt(key.get()).contains(item.state())) {				
//				if (!_parseTable.statesAtIncludes(key.get(),item.state())) {				
					new_state = _parseTable.add(key.get(),item.state());
					new_state.markRelationTuples();
					if (new_state.isActive())
						putOnAgenda ( 
							handleActiveState((ActiveState)new_state,key.get())
					    );
					else
						putOnAgenda( 
							complete((InActiveState)new_state,key.get(),
									 completionInitStates())
					    );											
				}
				else {
					handlePotentialAmbiguity(item.state(),slot);
					reLink(item.state(),slot);
				}
			}
			item.state().unLinkFromParse();
		}
		return counter;
	}
	
	protected void putOnAgenda(Container items) {
		Copying.copy(items,_agenda); 
	}
	
	protected ParseResult getResult() {
		HashSet parse_trees = findRoots();

		if (parse_trees.isEmpty()) {			
			return new SingleFailure(new HashSet());
		}
		else {
			return new Success(parse_trees);
		}
	}

	protected boolean recover(SingleFailure fault) throws QueryException {
		return false;
	}

	private void reLink(ParseState newState,SList slot) {
		SListIterator duplicate = slot.find(newState);
		if (!newState.previous().isEmpty()) {
			for (SListIterator p=newState.previous().begin();
				!p.atEnd();p.advance()) {
				if (duplicate.get()!=p.get())
					((ParseState)duplicate.get()).previous((ActiveState)p.get());
			}
		}
		else if (!newState.up().isEmpty()) {
			for (SListIterator u=newState.up().begin();
				!u.atEnd();u.advance()) {
				if (duplicate.get()!=u.get()) 
					((ParseState)duplicate.get()).up((ActiveState)u.get());
			}
		}	
	}
	
	protected boolean satisfiesCurrentConstraints(Object input_object,
	                                              ActiveState state)
	throws QueryException {
		//System.err.println("\n*"+state.toString());
		//System.err.println("\n*"+state.productionVariant().constraintsToString());
		SList constraints = state.currentConstraints();
			//state.productionVariant().constraintsAt(state.dot());   
		if (constraints.isEmpty())
			return true;
		else {
			boolean result = true;
			try {
				for (SListIterator c=constraints.begin();!c.atEnd();c.advance())
					result = 
		            	_input.getRelation(((Constraint)c.get()).relation()).
		            	predicate(buildQueryArguments(state,input_object,(Constraint)c.get()))
						&& 
						result;
		  	}
		  	catch (InvalidQueryArgumentsException e) {
		  		throw(new QueryException(e.getMessage()));
		  	}
			return result;
		 }                             
	                                                                
	}

	protected boolean satisfiesCurrentConstraints(InActiveState i_state,
	                                              ActiveState state)
	throws QueryException {
		//System.err.println("\n*"+state.toString());
		//System.err.println("\n*"+state.productionVariant().constraintsToString());
		SList constraints = state.currentConstraints();
			//state.productionVariant().constraintsAt(state.dot());   
		if (constraints.isEmpty())
			return true;
		else {
			boolean result = true;
			try {
				for (SListIterator c=constraints.begin();!c.atEnd();c.advance())
					result = 
		            	_input.getRelation(((Constraint)c.get()).relation()).
		            	predicate(buildQueryArguments(state,i_state,(Constraint)c.get()))
						&& 
						result;
		  	}
		  	catch (InvalidQueryArgumentsException e) {
		  		throw(new QueryException(e.getMessage()));
		  	}
			return result;
		 }                             
	                                                                
	}

	protected AgendaItem scan(ActiveState state,Object input_object)
	throws AnalysisException {
		AgendaItem new_item = null;
		parseAction = "scan";
		if (_input.getTerminalType(input_object).equals(state.symbolAtDot())) {
			if (state.cover().get(input_object)==null) {
				LeafState leaf = new LeafState(state.symbolAtDot());
				HashSet cover = new HashSet();
				cover.add(input_object);
				leaf.cover(cover);
				leaf.creator("scan");
				if (satisfiesCurrentConstraints(input_object,state)) {
					handleSemantics(leaf);
					new_item = advance(state,leaf);
					//state.cover().add(input_object); 
					// A hack to indicate a lexical match inspite of an error
					// in 'next' input beyond this point. Note that this has
					// no effect if advancement was succesfull.
				}
				else {
					createUnsatisfiedConstraintsError(state,input_object);
				}
			}
			else {	
				createCoverError(state,input_object);
			}
		}
		else {
			createScanningError(state,input_object);
		}
		//if (new_item != null)
		//	new_item.state().creator("scan");
			
		return new_item;
	}

}