package relap.Analyzer;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Applying;
import com.objectspace.jgl.algorithms.Comparing;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.algorithms.Finding;
import com.objectspace.jgl.predicates.BindSecondPredicate;
import com.objectspace.jgl.predicates.BindFirstPredicate;
import relap.LanguageModel.OrderingVariant;
import relap.LanguageModel.AttributeAssignment;
import relap.LanguageModel.ObjectReference;
import relap.LanguageModel.InDirectReference;
import relap.LanguageModel.Constraint;
import relap.LanguageModel.ExistenceConstraint;
import relap.LanguageModel.Verifiable;
import relap.LanguageModel.InvalidQueryArgumentsException;
import java.util.Enumeration;

public class ActiveState extends ParseState {

	private int _dot;
	private DescendantList _sub_trees;
	private HashSet _cover;
	private SList _completions;
	private Constraint _queryConstraint;
	private SList pendingInput;
	
	class AddToCover implements UnaryFunction {
		public Object execute(Object arg) {
			Copying.copy(((InActiveState)arg).cover(),_cover);
			return arg;
		}
	}

	public ActiveState() {
		_productionVariant = null;
		_dot = 1;
		_sub_trees = null;
		_completions = new SList();
		_queryConstraint = null;
		pendingInput = new SList();
	}

	public ActiveState(OrderingVariant variant) {
		_productionVariant = variant;
		_dot = 1;
		_sub_trees = new DescendantList(variant);
		_completions = new SList();
		_queryConstraint = null;
		pendingInput = new SList();
	}


	public int completionCount() {
		return _completions.size();
	}
	
	public void addToCompletions(InActiveState i_state) {
		_completions.pushBack(i_state);
	}
	
	public SList completions() { return _completions; }

	public void completions(SList list) { _completions = list; }

	public HashSet cover() {
    	if (_cover == null) {
       		_cover = new HashSet();
        	Applying.forEach(_sub_trees.flatRep(),new AddToCover());
        }
		return _cover;
	}
  
	public void cover(HashSet set) {
		_cover = set;
	}
	

	public SList currentConstraints() {
		return constraintsAt(_dot);
	}
	
	public SList constraintsAt(int position) {
		SList constraints = new SList();
		for (SListIterator c = _productionVariant.constraintsAt(position).begin();!c.atEnd();c.advance())
			if (Finding.every(mask(),new BindSecondPredicate(new IsNotMasked(),(Constraint)c.get())))
				constraints.add(c.get());
		return constraints;
	}


	public Sequence dereference(ObjectReference reference) {
		//System.err.println(this);
		//System.err.println(_sub_trees);
		return _sub_trees.dereference(reference);
	}
	
	public String describe() {
		StringBuffer buf  = new StringBuffer();
		buf.append("[ ");
		
		buf.append(_productionVariant.production().lhs());
		buf.append(" -> ");
		Array rhs = _productionVariant.production().rhs();
		for (ArrayIterator it = rhs.begin();!it.atEnd();it.advance()) {
			if (it.index()+1 == _dot)
				buf.append(" @ ");
			buf.append(it.get().toString());
			if (_productionVariant.isRepetitive(it.index()+1) && _productionVariant.isOptional(it.index()+1))
				buf.append("* ");
			else if (_productionVariant.isRepetitive(it.index()+1) && !_productionVariant.isOptional(it.index()+1))
				buf.append("+ ");
			else if (!_productionVariant.isRepetitive(it.index()+1) && _productionVariant.isOptional(it.index()+1))
				buf.append("? ");
			else 
				buf.append(" ");
		}
		buf.append("]");
		return buf.toString();
	}
	
	
	
	public boolean bindsAttribute(int position) {
		return
			Finding.some(
				_productionVariant.attributeAssignments(),
				new BindFirstPredicate(
					new BinaryPredicate () {
						public boolean execute(Object pos,Object assgn) {
    						return ((AttributeAssignment)((Pair)assgn).second).value().position() 
    						       == ((Integer)pos).intValue();
						
						}
					},
					new Integer(position)
				)
			);
	}

	public int dot() { return _dot; }
 
	public void dot(int pos) {
		_dot = pos;
	}
 
	public boolean dotAtEnd() {
		return 
			_productionVariant.production().rhs().size() == _dot &&
			pendingInput.isEmpty();
	}
 
 	public SList getPendingInput() {
 		return pendingInput;
 	}
 	
 	public void setPendingInput(Sequence list) {
 		pendingInput.clear();
 		Copying.copy(list,pendingInput);
 	}
 	
	public boolean equals(Object state) {
		return
			super.equals(state) ||  
			(state instanceof ActiveState &&
			 equalState((ActiveState)state));
				/*
		  	_productionVariant==((ActiveState)state).productionVariant() &&
		  	_dot==((ActiveState)state).dot() &&
		  	pendingInput.size() == 
		  		((ActiveState)state).getPendingInput().size() &&
		  	equalSubTrees((ActiveState)state) );
		  	*/
	}

	public boolean equalState(ActiveState state) {
		return 	
			_productionVariant == state.productionVariant() &&
		  _dot == state.dot() &&
		  pendingInput.size() == 
		  	state.getPendingInput().size() && equalSubTrees(state);
	}
	
  public int hashCode() {
 		//System.err.println(this);
 		//System.err.println(_productionVariant.production().lhs().hashCode());
		return _productionVariant.production().lhs().hashCode();
	}
 
	public boolean equalSubTrees(ActiveState state) {
		return Comparing.equal(_sub_trees.rep(),state.subTrees().rep());
	}
	
	public void initCopy(ParseState new_state) {
    	super.initCopy(new_state);
    	((ActiveState)new_state).productionVariant(_productionVariant);
    	((ActiveState)new_state).dot(_dot);
    	((ActiveState)new_state).subTrees(new DescendantList(_sub_trees));
    	((ActiveState)new_state).completions(new SList(_completions));
    	((ActiveState)new_state).queryConstraint(_queryConstraint);
    	((ActiveState)new_state).setPendingInput(pendingInput);
    	if (_cover!=null) 
    		((ActiveState)new_state).cover(new HashSet(_cover));
   	}
   
	public boolean isActive() {
		return true;
	}
	
	public SList inversePermutedMask() {
		SList permutedMask = new SList();
		for (SListIterator pos = mask().begin();!pos.atEnd();pos.advance())
			permutedMask.pushBack(
				new Integer(((ObjectReference)_productionVariant.inversePermutation().at(((Integer)pos.get()).intValue())).position())
			);	
		return permutedMask;
	}
	
	public void maskSymbolAt(int position) {
		mask().pushBack(new Integer(position));
	} 
	
	public void unMaskSymbolAt(int position) {
		mask().remove(new Integer(position));
	} 
	
	public void queryConstraint(Constraint c) {
		_queryConstraint = c;
	}

	public Constraint queryConstraint() {
		return _queryConstraint;
	}

	public SList evaluateNonExistenceConstraintsAt(IndexedMDSet input,int pos)
	throws QueryException {
		SList failedPredicates = new SList();
		for (SListIterator neC = _productionVariant.neConstraintsAt(pos).begin();
		 	 !neC.atEnd();neC.advance()) {
		 	try {
		 		if (Finding.every(mask(),new BindSecondPredicate(new IsNotMasked(),(ExistenceConstraint)neC.get())))
		 			if (!((ExistenceConstraint)neC.get()).verify(this,input)) {
		 				//System.err.println("failed: "+neC.get());
		 				Enumeration fData = ((Verifiable)neC.get()).getFailureData().elements();
		 				while (fData.hasMoreElements()) {
		 					failedPredicates.pushBack(
		 						new NotExistsError(
		 							this,
		 							(Verifiable)neC.get(),
		 							(FailureData)fData.nextElement()
		 						)
		 					);	
		 					//System.err.println("failedPredicates.back(): "+failedPredicates.back());
		 				}
		 			}
		 	} 
		 	catch (InvalidQueryArgumentsException e) {
		 		throw(new QueryException(e.getMessage()));
		 	}
		}
		return failedPredicates;
	
	}

	public SList evaluatePredicatesAt(IndexedMDSet input,int pos)
	throws QueryException {
		SList failedPredicates = new SList();
		for (SListIterator c = _productionVariant.predicatesAt(pos).begin();
		 	 !c.atEnd();c.advance()) {
		 	// System.err.println("!!!!"+c.get());
		 	try {
		 		if (Finding.every(mask(),new BindSecondPredicate(new IsNotMasked(),(Verifiable)c.get())))
		 			if (!((Verifiable)c.get()).verify(this,input)) {
		 				//System.err.println(">>> Failed: "+c.get());
		 				Enumeration fData = ((Verifiable)c.get()).getFailureData().elements();
		 				while (fData.hasMoreElements())
		 					failedPredicates.pushBack(
		 						new PredicateEvaluationError(
		 							this,
		 							(Verifiable)c.get(),
		 							(FailureData)fData.nextElement()
		 						)
		 					);		
		 			}
		 	} 
		 	catch (InvalidQueryArgumentsException e) {
		 		throw(new QueryException(e.getMessage()));
		 	}
		}
		//if (!failedPredicates.isEmpty())
		//	System.err.println(">>> Failed: "+failedPredicates);
		return failedPredicates;	
	}

	public String parseChainToString() {
		StringBuffer buf = new StringBuffer();
	    buf.append("\nwhile parsing a: \n"+_productionVariant.production().toString()
    	           +"\nat position: "+(new Integer(_dot)).toString());
    	if (_previous != null)
    		if (_previous.size()>0)
        	buf.append(((ActiveState)_previous.front()).parseChainToString());
		else
        	if (_up != null) 
        		buf.append(((ActiveState)_up.front()).parseChainToString());
		return buf.toString();
	}
	
	


/*
selectCandidate
        "Select the i-state having the largest cover for completion from completionCandidates."
    | cSize result |
    cSize := 0.
    completionCandidates do: [ :array |
        (array at: 1) cover size > cSize ifTrue: [
            result := array.
            cSize := (result at: 1) cover size]
     ].
    ^result!
*/
  
	public String toString() {
		StringBuffer buf  = new StringBuffer();
		buf.append("[ "+_productionVariant.toString()+
				   " "+(new Integer(_dot)).toString()+" ]\n   (");
		for (ForwardIterator it = cover().start();!it.atEnd();it.advance())
			buf.append(" "+it.get().toString());
        
     	if (_dot == 1)
        	buf.append(")\n   creator: "+_creator.toString()+" from: "+upToString()
        			   +" "+originatorToString());
        else
        	buf.append(")\n   creator: "+_creator.toString()+" from: "+previousToString()
        			   +" "+originatorToString());
        if (!pendingInput.isEmpty())
        	buf.append("\npending "+pendingInput);
		return buf.toString();
	}

	public DescendantList subTrees() {
		return _sub_trees;
	}
  
	public void subTrees(DescendantList desc) {
		_sub_trees = desc;
	}
 
	public String symbolAtDot() {
		return _productionVariant.symbolAt(_dot);
	}
  
  public String symbolAt(int pos) {
  	return _productionVariant.symbolAt(pos);
  }

	public HashSet toBeBoundAttrs() {
        //Answer the set of attributes to be bound of the nonterminal
        //in the expander constraint at current dot position.
	    if (_dot == 1)
			return _productionVariant.firstAttrs();
		else {
			HashSet set = new HashSet();
//			//System.err.println(toString()+"\n"+_productionVariant.constraintsToString());
			set.add(((InDirectReference)_productionVariant.constraintAt(_dot).argumentAt(_dot)).attribute());
			return set;
		}
	}        
}