package relap.Analyzer;

import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.Array;
import com.objectspace.jgl.Container;
import com.objectspace.jgl.algorithms.Copying;
import relap.Util.Utils;
import relap.LanguageModel.Constraint;
import relap.LanguageModel.ExpanderRelation;
import relap.LanguageModel.Tuple;
import relap.LanguageModel.Verifiable;
import relap.Editor.ExplicitExpanderRelation;

public class ConstraintError extends ErrorDescriptor {
	protected Object _index;
	protected HashSet indices;
	protected ExpanderRelation relation;
	protected Array types;
	protected boolean isMissing;
	protected Verifiable failedConstraint;
	
	public ConstraintError(ActiveState state,
	                        String msg,Object input_object,
	                        ExpanderRelation relation,Verifiable failed) {
		super(state,msg);
		_index = input_object;
		indices = new HashSet();
		this.relation = relation;
		types = new Array();
		isMissing = false;
		failedConstraint = failed;
	}
		
	public ConstraintError(ActiveState state,
	                        String msg,Array input_objects,Array types,
	                        ExpanderRelation relation,Verifiable failed) {
		super(state,msg);
		_index = null;
		indices = new HashSet();
		if (relation instanceof ExplicitExpanderRelation) {
			Object figure = 
				((ExplicitExpanderRelation)relation)
					.getFigure(
						new Tuple(new Array(input_objects),relation)
					);
			if (figure != null) {
				isMissing = false;
				indices.add(figure);
			} else {
				isMissing = true;
				Copying.copy(input_objects,indices);
			}
		} else {
			isMissing = true;
			Copying.copy(input_objects,indices);
		}
		this.relation = relation;
		this.types = types;
		failedConstraint = failed;
	}

	public Verifiable getFailedCondition() {
		return failedConstraint;
	}

	public Object index() {
		return _index;
	}

	public HashSet indices() {
		return indices;
	}
	
	public String describe() {
		String msg;
		if (isMissing)
			msg = relation.describeMissingTuple();
	    else
	    	msg = relation.describeInvalidTuple(types);
		return 
			msg
			+"\nto recognize a "
			+((ActiveState)errorLoc).productionVariant().production().description();

	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("at "+Utils.listContents(indices,",","and")+":\n");
		buf.append(_message);
		return buf.toString();
	}
}
