package relap.Analyzer;

import relap.LanguageModel.*;
import com.objectspace.jgl.*;
import relap.Util.GraphNode;

public class DependencyGraph {

	private ProductionDependency root;
	private RelationalGrammarImplementation grammar;
	private String attribute;
	private String nonTerminal;
	private HashSet nodes;
	
	public DependencyGraph(RelationalGrammarImplementation grammar,
	                       String attribute,String nonTerminal) {
		this.grammar = grammar;      
	    this.attribute = attribute;
	    this.nonTerminal = nonTerminal;
	    root = null;
	    nodes = new HashSet();
	}

	public void buildGraph(PredictionTable table) {
		Stack toDo = new Stack();
		root = 
			new ProductionDependency(
				table.variantsFor(nonTerminal,attribute),
				nonTerminal,
				attribute
			);
		nodes.add(root);
		toDo.push(root);	
		while (!toDo.isEmpty()) {
			ProductionDependency current = (ProductionDependency)toDo.pop();
			for (int i = 0;i<current.variants().size();i++) {
				OrderingVariant v = (OrderingVariant)current.variants().at(i);
				if (grammar.isNonTerminal(v.symbolAt(1))) {
					String attrib =
						((InDirectReference) 
							((AttributeAssignment)v.attributeAssignments().get(current.attribute())).value()
						 ).attribute();
					ProductionDependency dependant =
						new ProductionDependency(
							table.variantsFor(v.symbolAt(1),attrib),
							v.symbolAt(1),
							attrib
						);
					if (nodes.get(dependant) != null)
						current.addNeighbour((ProductionDependency)nodes.get(dependant));
					else {
						nodes.add(dependant);
						current.addNeighbour(dependant);
						toDo.push(dependant);
					}								
				}
			}
		
		}                                                 
	}
	
	public ProductionDependency root() {
		return root;
	}
	
	public String toString() {
		StringBuffer buf = new StringBuffer(
			"DependecyGraph { "+nonTerminal+"."+attribute+"\n"
		);
		for (ForwardIterator node=nodes.start();!node.atEnd();node.advance()) {
			buf.append("\n"+node.get().toString());
			for (int i = 0;i<((GraphNode)node.get()).neighbours().size();i++)
				buf.append(
					"\n-->"+
					((ProductionDependency)((GraphNode)node.get()).neighbours().at(i)).nonTerminal()+
					"."+
					((ProductionDependency)((GraphNode)node.get()).neighbours().at(i)).attribute()
				);
		}
		buf.append("\n}");
		return buf.toString();
	}
}           
