package relap.Analyzer;

import relap.LanguageModel.SymbolSequence;
import com.objectspace.jgl.Array;
import com.objectspace.jgl.ArrayIterator;
import com.objectspace.jgl.ForwardIterator;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.Sequence;
import com.objectspace.jgl.Container;
import com.objectspace.jgl.algorithms.Transforming;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.functions.BindFirst;
import com.objectspace.jgl.BinaryFunction;
import relap.LanguageModel.ObjectReference;

public class DescendantList
	implements SymbolSequenceHolder {

	protected SymbolSequence repetitionFlags;
	protected Array descendants; 
	private int size;
	
	public DescendantList(SymbolSequence repetitionFlags) {
		this.repetitionFlags = repetitionFlags;
		descendants = new Array(repetitionFlags.sequenceLength());
		size = 0;
	}	 
	
	public DescendantList(DescendantList copyOf) {
		this.repetitionFlags = copyOf.getRepetitionFlags();
		//System.err.println(">> "+repetitionFlags);
		descendants = new Array(repetitionFlags.sequenceLength());
		size = copyOf.size();
		for (int i=0;i<size;i++) {
			//System.err.println(">>> at("+(i+1)+"): "+copyOf.at(i+1));
			if (repetitionFlags.isRepetitive(i+1))
				descendants.put(i,new SList((SList)copyOf.at(i+1)));
			else
				descendants.put(i,copyOf.at(i+1));
		}
	}	 
	
	public SymbolSequence getRepetitionFlags() {
		return repetitionFlags;
	}
	
	public int size() {
		return size;
	}
	
	public Object at(int pos) {
		return descendants.at(pos-1);
	}
	
	public Object put(int pos,Object what) {
		descendants.put(pos-1,what);
		if (what != null && pos > size) size = pos;
		return what;
	}
	
	public Array rep() {
		Array rep = new Array(size);
		for (int i=0;i<size;i++) rep.put(i,descendants.at(i));
		return rep;
	}
	
	public Array flatRep() {
		Array flatSequence = new Array();
		for (int i = 1;i <= size;i++)
			if (repetitionFlags.isRepetitive(i))
				Copying.copy((SList)descendants.at(i-1),flatSequence);
			else
				flatSequence.pushBack(descendants.at(i-1));
		return flatSequence;	
	}
	
	public ArrayIterator symbols() {
		return flatRep().begin();	
	}
	
	public ForwardIterator descendants() {
		SList desc = new SList();
		for (int i = 0;i<size;i++) desc.pushBack(descendants.at(i));
		return descendants.start();
	}
	
	public void addDescendant(int pos,InActiveState descendant) {
		if (repetitionFlags.isRepetitive(pos)) {
			if (descendants.at(pos-1) == null)
				descendants.put(pos-1,new SList());
			((SList)descendants.at(pos-1)).pushBack(descendant);
		} else
			descendants.put(pos-1,descendant);
		if (pos > size) size = pos;
	}
	
	public Sequence dereference(ObjectReference reference) {
		SList objects = new SList();
		if (reference.position() == -1)
			objects.pushBack(null);
		else {
			if (repetitionFlags.isRepetitive(reference.position()))
				return 
				(Sequence)Transforming.collect(
					(Container)descendants.at(reference.position()-1),
					new BindFirst(
						new BinaryFunction() {
							public Object execute(Object ref,Object iState) {
								return ((InActiveState)iState).referencedValue((ObjectReference)ref);
							}
						},
						reference
					)
				);
			else
				objects.pushBack(
					((InActiveState)descendants.at(reference.position()-1))
						.referencedValue(reference)
				);
		}
		return objects;
	}
	
	public String toString() {
		return "size "+size+" "+descendants.toString();
	}
}