package relap.Analyzer;

import relap.LanguageModel.*;
import relap.Util.Utils;
import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Applying;
import com.objectspace.jgl.algorithms.Sorting;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.algorithms.Finding;
import com.objectspace.jgl.algorithms.Transforming;
import com.objectspace.jgl.predicates.*;
import java.io.*;
import java.util.Enumeration;

public class FixedStartParser extends ARGRecognizer {

	protected Array _inits; 
	protected SList _input_ambiguities;
	protected HashMap _grammar_ambiguities;
	protected HashSet _island_objects;
	protected HashSet oldErrors;
	protected HashSet allDeadEnds;
	protected HashSet coverConflictSet;
	protected String recoveryDiagnostics;
	protected boolean detourTaken;

	public FixedStartParser() {
		super();
		_inits = null;
		_input_ambiguities = null;
		_grammar_ambiguities = null;
		_island_objects = null;
		oldErrors = new HashSet();
		allDeadEnds = new HashSet();
		coverConflictSet = new HashSet();
		recoveryDiagnostics = "";
		detourTaken = false;
	}
	
	protected void handleSemantics(BranchingState newState) {
		newState.createSemantics();
	}
	
	protected void handleSemantics(LeafState newState) {
		newState.createSemantics(this);
	}
	
	protected void handleSemantics(EmptyLeaf empty) {
	//	System.err.println("handling empty semantics");
		empty.createSemantics();
	}
	
	protected Array collectDeadEnds(ParseState state,HashSet acc_cover,ParsePath path) {
		Array dead_ends = new Array();
		//System.err.println(state.surrogate().toString());
		if (state.isActive()) {
			if (path.contains((ActiveState)state)) {
				path.push((ActiveState)state);
				createCircularityError((ActiveState)state,path);
				path.pop();
				dead_ends.add(
					new DeadEnd((ActiveState)state,
			    		        state.errorDescription().cover().union(acc_cover), 
			        		    _parseTable.inputObjectForSurrogate(state.surrogate()))
				);
				return dead_ends;
			}
			path.push((ActiveState)state);
			if (state.errorDescription() != null) {
				if (state.errorDescription().isError())
					dead_ends.add(
						new DeadEnd((ActiveState)state,
					    	        state.errorDescription().cover().union(acc_cover), 
					        	    _parseTable.inputObjectForSurrogate(state.surrogate()))
					);
				//else
				//	System.err.println(state.errorDescription());
			}
			else {
				SList next_states = _parseTable.nextStatesFor(state);
				//System.err.println(next_states);
				if (next_states.isEmpty()) {
					// if there are no next states because a predicate failed,
					// prune this path
					SList down_states = _parseTable.downStatesFor(state);
					if (down_states.isEmpty()) {
						state.errorDescription(
							new OtherError((ActiveState)state,"Grammar error?")
						);
						dead_ends.add(
							new DeadEnd((ActiveState)state,
							            state.cover().union(acc_cover),
						           	 	_parseTable.inputObjectForSurrogate(state.surrogate()))
						);
						
					} else {
						//System.err.println("\n"+state.toString());
						for (SListIterator down=down_states.begin();!down.atEnd();down.advance()) {
							Array temp = 
								collectDeadEnds((ParseState)down.get(),
							                	((ParseState)down.get()).cover().union(acc_cover),
							                	path);
							Copying.copy(temp,dead_ends);
						}
					}
				}
				else // add ambiguity check in completion and bookkeeping in ActiveState!
					for (SListIterator next=next_states.begin();!next.atEnd();next.advance())
						Copying.copy(
							collectDeadEnds((ParseState)next.get(),
						 	                ((ParseState)next.get()).cover().union(acc_cover),
						 	                path),
							dead_ends
						);
			}
			path.pop();
		}
			
		return dead_ends;
	} 

	
	protected Array completionInitStates() {
		return new Array();
	}

	protected void createCircularityError(ActiveState state,ParsePath path) {
		state.errorDescription(
			new CircularityError(state,new ParsePath(path))
		);
	}

	protected void createCompletionAmbiguityError(ActiveState a_state) {
		CompletionAmbiguity amb = (CompletionAmbiguity)_grammar_ambiguities.get(a_state);
		if (amb == null) {
			amb = 
				new CompletionAmbiguity(
					a_state,
					_parseTable.inputObjectForSurrogate(a_state.surrogate())
				);
			_grammar_ambiguities.put(a_state,amb);
		}
	}

	protected void createPathAmbiguityError(ParseState state) {
		PathAmbiguity amb = (PathAmbiguity)_grammar_ambiguities.get(state);
		if (amb == null) {
			if (state instanceof ActiveState)
				amb = 
					new PathAmbiguity(
						state,
						_parseTable.inputObjectForSurrogate(state.surrogate())
					);
			else
				amb = 
					new PathAmbiguity(
						state,
						((InActiveState)state).indices().start().get()
					);
			
			_grammar_ambiguities.put(state,amb);
		}
	}



	protected void createCoverError(ActiveState prevState,ActiveState state,HashSet input_objects) {
		Constraint constraint = state.queryConstraint();
		ObjectReference bound_argument = constraint.argumentBefore(state.dot());
		Array args = new Array(2);
		if (bound_argument==constraint.firstArg()) {
			args.put(0,state.dereference(bound_argument));
			args.put(1,Utils.asSList(input_objects));
		} else {
			args.put(0,Utils.asSList(input_objects));
			args.put(1,state.dereference(bound_argument));
		}
		SList invalidRelations = new SList();
		for (ForwardIterator errData= Utils.cartesianProduct(args).start();
		     !errData.atEnd();errData.advance()) {
		     Array arr = (Array)errData.get();
		     Array types = new Array(2);
		     types.put(0,_input.describeTerminal(_input.getTerminalType(arr.at(0))));
		     types.put(1,_input.describeTerminal(_input.getTerminalType(arr.at(1))));
		     invalidRelations.pushBack(
		     	new ConstraintError(
		     		state,
		     		"invalid relation: '"+arr.at(0)+" "+constraint.relation()+" "+arr.at(1)+"'",
		     		arr,
		     		types,
		     		(ExpanderRelation)_input.getRelation(constraint.getName()),
		     		constraint
		     	)
		     );
		}
		prevState.errorDescription(
			new CompositeError(state,invalidRelations)
		);
	}

	protected void createCoverError(ActiveState state,Object input_object) {
		HashSet objects = new HashSet();
		objects.add(input_object);
		createCoverError(state,state,objects);
	}

	protected void createCoverError(ActiveState state,InActiveState i_state) {
		state.errorDescription(
			new CoverError(state,i_state)
		);
	}

 	protected void createExtraRelationsError(ActiveState a_state,
										InActiveState i_state,
										ActiveState new_state,
										HashSet objects,
										int expectedCount) {
		Constraint constraint = new_state.queryConstraint();
		int pos = -1;
		HashSet expected = expectedTerminals(new_state);
		ObjectReference bound_argument = constraint.argumentBefore(new_state.dot());
		SList extras = new SList();
		for (ForwardIterator first = new_state.dereference(bound_argument).start();
			 !first.atEnd();first.advance()) {
			StringBuffer buf = new StringBuffer();
			if (bound_argument==constraint.firstArg())
			{
				pos = 1;
				buf.append("expecting only "+expectedCount+" objects x of type "+
				       	Utils.listContents(_input.describeTerminals(expected),", "," or ")+" such that ");
				buf.append("'"+first.get().toString()+" ");
				buf.append(constraint.relation()+" x' and found");
				buf.append(Utils.contentsToString(objects));
			}
			else {
				pos = 0;
				buf.append("expecting only "+expectedCount+" objects x of type "+
			           		Utils.listContents(_input.describeTerminals(expected),", "," or ")+" such that ");
				buf.append("'x "+constraint.relation());
				buf.append(" "+first.get().toString()+"' and found");
				buf.append(Utils.contentsToString(objects));
			}
			extras.pushBack(
				new ExtraRelations(
					a_state,i_state,buf.toString(),
			    	first.get(),
			    	_input.describeTerminal(_input.getTerminalType(first.get())),
			    	objects,
			    	(ExpanderRelation)_input.getRelation(constraint.relation()),
			    	pos,
			    	expectedCount,
			    	_input.describeTerminals(expected),
			    	constraint
				)
			);
		}
		a_state.errorDescription(new CompositeError(a_state,extras));

	}
 
	protected void createMissingRelationError(ActiveState a_state,
										  InActiveState i_state,
										  ActiveState new_state)
	throws QueryException { 
		int at = -1;
		Constraint constraint = new_state.queryConstraint();
		HashSet expected = expectedTerminals(new_state);
		ObjectReference bound_argument = constraint.argumentBefore(new_state.dot());
		SList missing = new SList();
		for (ForwardIterator first = new_state.dereference(bound_argument).start();
			 !first.atEnd();first.advance()) {
			StringBuffer buf = new StringBuffer();
			if (bound_argument==constraint.firstArg()) {
				at = 1;
				buf.append("expecting an object x of type "+
			           	Utils.listContents(_input.describeTerminals(expected),", "," or ")+" such that ");
				buf.append("'"+first.get().toString()+" ");
				buf.append(constraint.relation()+" x'\n");
			}
			else {
				at = 0;
				buf.append("expecting an object x of type "+
			           Utils.listContents(_input.describeTerminals(expected),", "," or ")+" such that ");
				buf.append("'x "+constraint.relation());
				buf.append(" "+first.get().toString()+"'");
			}
			missing.pushBack(
				new MissingRelationTuple(
					a_state,
					i_state,
					buf.toString(),
			        first.get(),
			        _input.describeTerminal(_input.getTerminalType(first.get())),
			        _input.describeTerminals(expected),
			        at,
			        _input.getRelation(constraint.relation()),
			        constraint
			   	)
			);
		}
		//System.err.println(missing);
		a_state.errorDescription(new CompositeError(a_state,missing));
	}

	protected void createScanningError(ActiveState state,Object input_object) {
		String expected = _input.describeTerminal(state.symbolAtDot());
		state.errorDescription(
			new ScanningError(
				state,input_object,expected,new LexicalMatch(expected)
			)
		);
		//System.err.println(state.errorDescription());
	}

	protected void createUnsatisfiedConstraintsError(ActiveState state,
	               									 Object input_object)
	throws QueryException {
		SList constraints = state.currentConstraints();
			//state.productionVariant().constraintsAt(state.dot());   
		if (!constraints.isEmpty()) {
			try {
				for (SListIterator c=constraints.begin();!c.atEnd();c.advance()) {
					Array arguments = 
						buildQueryArguments(state,input_object,(Constraint)c.get());
					ExpanderRelation relation = 
						(ExpanderRelation)_input.getRelation(((Constraint)c.get()).relation());
					if (!relation.predicate(arguments)) {
						SList errors = new SList();
						for (ForwardIterator arg = relation.getFailedArgs().start();
						     !arg.atEnd();arg.advance()) {
						    Array types = new Array(2);
						    Array objects = (Array)arg.get();
						    types.put(0,_input.describeTerminal(_input.getTerminalType(objects.at(0))));
						    types.put(1,_input.describeTerminal(_input.getTerminalType(objects.at(1))));
		            		errors.pushBack(
		            			new ConstraintError(
		            				state,
		            	        	"expecting relation '"+
		            	        	 objects.at(0)+
		            	        	 " "+relation+
		            	        	 " "+objects.at(1)+"'",
		            	         	objects,
		            	        	 types,
		            	         	relation,
		            	         	(Constraint)c.get()
		            	        )
		            	    );
		            	}
		            	state.errorDescription(new CompositeError(state,errors));
		            }
		        }
		  	}
		  	catch (InvalidQueryArgumentsException e) {
		  		throw(new QueryException(e.getMessage()));
		  	}
		}                             
	}


	protected void createUnsatisfiedConstraintsError(ActiveState state,
	               									 InActiveState i_state)
	throws QueryException {
		SList constraints = state.currentConstraints();
			//state.productionVariant().constraintsAt(state.dot());   
		if (!constraints.isEmpty()) {
			try {
				for (SListIterator c=constraints.begin();!c.atEnd();c.advance()) {
					Array arguments = 
						buildQueryArguments(state,i_state,(Constraint)c.get());
					ExpanderRelation relation = 
						(ExpanderRelation)_input.getRelation(((Constraint)c.get()).relation());
					if (!relation.predicate(arguments)) {
						SList errors = new SList();
						for (ForwardIterator arg = relation.getFailedArgs().start();
						     !arg.atEnd();arg.advance()) {
						    Array types = new Array(2);
						    Array objects = (Array)arg.get();
						    types.put(0,_input.describeTerminal(_input.getTerminalType(objects.at(0))));
						    types.put(1,_input.describeTerminal(_input.getTerminalType(objects.at(1))));
		            		errors.pushBack(
		            			new ConstraintError(
		            				state,
		            	        	"expecting relation '"+
		            	        	 objects.at(0)+
		            	        	 " "+relation+
		            	        	 " "+objects.at(1)+"'",
		            	         	objects,
		            	        	 types,
		            	         	relation,
		            	         	(Constraint)c.get()
		            	        )
		            	    );
		            	}
		            	state.errorDescription(new CompositeError(state,errors));
		            }
		        }
		  	}
		  	catch (InvalidQueryArgumentsException e) {
		  		throw(new QueryException(e.getMessage()));
		  	}
		}                             
	}

	public boolean evaluatePredicates(ActiveState prevState,InActiveState state)
	throws QueryException {
		SList failedPredicates = state.evaluatePredicates(_input);
		SList failedConstraints = state.evaluateNonExistenceConstraints(_input);
		if (failedPredicates.isEmpty() && failedConstraints.isEmpty())
			return true;
		if (!failedPredicates.isEmpty() && failedConstraints.isEmpty())
			prevState.errorDescription(
				new FailedPredicates(
					prevState,failedPredicates
				)
			);
		else  // !failedConstraints.isEmpty()
			prevState.errorDescription(
				new FailedNotXConstraints(
					prevState,failedConstraints
				)
			);
			
		//System.err.println("prevState.errorDescription() = "+prevState.errorDescription());
		return false;
	}

	protected boolean tryEvaluatePredicates(ActiveState prevState,InActiveState state)
	throws QueryException {
		if (prevState.dot() == 1) 
			return true;
		SList failedPredicates = 
			prevState.evaluatePredicatesAt(_input,prevState.dot()-1);
		SList failedConstraints = 
			prevState.evaluateNonExistenceConstraintsAt(_input,prevState.dot()-1);
		if (failedPredicates.isEmpty() && failedConstraints.isEmpty())
			return true;
		if (!failedPredicates.isEmpty() && failedConstraints.isEmpty())
			prevState.errorDescription(
				new FailedPredicates(
					prevState,failedPredicates
				)
			);
		else  // !failedConstraints.isEmpty()
			prevState.errorDescription(
				new FailedNotXConstraints(
					prevState,failedConstraints
				)
			);
			
		//System.err.println("prevState.errorDescription() = "+prevState.errorDescription());
		return false;
	}

	public boolean evaluatePredicates(ActiveState prevState,ActiveState newState)
	throws QueryException {
		SList failedPredicates = 
				newState.evaluatePredicatesAt(_input,prevState.dot());
		SList failedConstraints = 
				newState.evaluateNonExistenceConstraintsAt(_input,prevState.dot());
		if (failedPredicates.isEmpty() && failedConstraints.isEmpty())
			return true;
		if (!failedPredicates.isEmpty() && failedConstraints.isEmpty())
			prevState.errorDescription(
				new FailedPredicates(
					prevState,failedPredicates
				)
			);
		else  // !failedConstraints.isEmpty()
			prevState.errorDescription(
				new FailedNotXConstraints(
					prevState,failedConstraints
				)
			);
			
		//System.err.println("prevState.errorDescription() = "+prevState.errorDescription());
		return false;
	}


	// This is way too simple! Need a more sophisticated algorithm.
	// Filter also based on the exptected terminal type?
	// or then the logic of getSameConstraints needs to be reversed
	protected int objectsExpected(ActiveState state) {
		SList cons = 
			state.productionVariant()
				.getSameConstraints(state.queryConstraint(),state.dot());
		HashSet xp = expectedTerminals(state);
		int anchor = 
			state.queryConstraint().argumentBefore(state.dot()).position();
		SList result = new SList();
		for (int i=0;i<cons.size();i++) {
			Constraint c = (Constraint)cons.at(i);
			if (c == state.queryConstraint())
				result.pushBack(c);
			else {
				int fwdPos = 
					c.firstArg().position() == anchor ? 
						c.secondArg().position() : c.firstArg().position(); 
				if (xp.equals(expectedTerminals(state,c,fwdPos)))
					if (state.productionVariant().isRepetitive(fwdPos))
						return _input.objects().size(); // max possible value
					else
						result.pushBack(c);
			}
		}
		//System.err.println("#### result "+result);
		return result.size();
		/*	
		return 
			state.productionVariant()
				.getSameConstraints(state.queryConstraint(),state.dot())
					.size();
		*/
	}
	
	private HashSet expectedTerminals(ActiveState state,Constraint cnstr,int pos) {
		HashSet expectedTerms;
 		if (_grammar.isNonTerminal(state.symbolAt(pos))) 
			expectedTerms = 
				_predictionTable.firstTerminalsFor(
					((InDirectReference)cnstr.argumentAt(pos)).attribute(),
					state.symbolAt(pos)
				);
		else {
			expectedTerms = new HashSet();
			expectedTerms.add(state.symbolAt(pos));
		}
		return expectedTerms;
	}

	private HashSet expectedTerminals(ActiveState state) {
		return expectedTerminals(state,state.queryConstraint(),state.dot());
		/*
		HashSet expectedTerms;
 		if (_grammar.isNonTerminal(state.symbolAtDot())) 
			expectedTerms = 
				_predictionTable.firstTerminalsFor(
					((InDirectReference)
						state.queryConstraint().argumentAt(state.dot())
					).attribute(),
					state.symbolAtDot()
				);
		else {
			expectedTerms = new HashSet();
			expectedTerms.add(state.symbolAtDot());
		}
		return expectedTerms;
		*/
	}

	protected Sequence sortCandidates(ActiveState state, HashSet candidates) {
		DList list = Utils.asDList(candidates);
		BinaryPredicate orderBy = 
			state.productionVariant().getOrderingPredicateAt(state.dot());
		if (orderBy != null) { 
			Sorting.sort(list,orderBy);
			return list;
		} else
			return null;	
	}
	
	protected HashSet removeCoverConflictObjects(ActiveState state,
	        									 HashSet candidates) {
		coverConflictSet.clear();
		HashSet tmp = candidates.difference(state.cover());
		if (tmp.isEmpty()) {
			Copying.copy(
				candidates.intersection(state.cover()),
				coverConflictSet
			);
			//System.err.println("coverConflictSet == "+coverConflictSet);
		}
		return tmp;	
	}

	protected HashSet filterCandidates(HashSet candidates,ActiveState new_state) {
		if (candidates.isEmpty()) {
			coverConflictSet.clear();
			return candidates;
		}
		if (candidates.size()>1) {
 			//System.err.println("before ordering: "+candidates);
 			Sequence list = sortCandidates(new_state,candidates);
 			if (list != null) {
				candidates = new HashSet();
				candidates.add(
					list.at(
						new_state.productionVariant().
						getPositionAtOrderingFor(new_state.dot()).intValue()
					)
				);
			}
 			//System.err.println("finally: "+candidates);
		}
		return removeCoverConflictObjects(new_state,candidates);
			
		/*
		HashSet finalPick = new HashSet();
		HashSet expectedTerms = expectedTerminals(new_state);
		//System.err.println("expected: "+expectedTerms.toString());
		for (HashSetIterator obj= tmpPick.begin();!obj.atEnd();obj.advance())
			if (expectedTerms.get(_input.getTerminalType(obj.get()))!=null)
				finalPick.add(obj.get());

 		//System.err.println("after: "+finalPick);
		return finalPick;
		*/
	}

	protected HashSet peek(HashSet candidates,ActiveState newState) {
		HashSet peekSet = new HashSet();
		HashSet expectedTerms = expectedTerminals(newState);
		for (HashSetIterator obj= candidates.begin();!obj.atEnd();obj.advance())
			if (expectedTerms.get(_input.getTerminalType(obj.get()))!=null)
				peekSet.add(obj.get());

 		if (peekSet.size() < candidates.size() ) {
			//System.err.println("before peek: "+candidates);
 			//System.err.println("after: "+peekSet);
 		} 

		return peekSet;	
	}

	protected HashSet processQueryResult(ActiveState previousState,
	                                     InActiveState advancedWith,
	                                     ActiveState newState,
	                                     HashSet originalResult)
		throws QueryException,QueryResultException {
		HashSet queryResult;
		//System.err.println("originalResult: "+originalResult);
		//System.err.println("query at: "+newState);
		if (newState.productionVariant().isRepetitive(newState.dot())) {
			queryResult = 
				peek(
					removeCoverConflictObjects(newState,originalResult),
					newState
				); 
			//System.err.println("after peek: "+queryResult);
			if (queryResult.isEmpty()) {
				//if (!coverConflictSet.isEmpty())
				//	createCoverError(previousState,newState,coverConflictSet);
				//else 
				if (!newState.productionVariant().isOptional(newState.dot())){
					//System.err.println("### MissingRelation 1");
					createMissingRelationError(previousState,advancedWith,newState);
				}
				return queryResult;
			}
			Sequence list = sortCandidates(newState,queryResult);
			if (list == null)
				list = Utils.asSList(queryResult);
			queryResult = new HashSet();
			queryResult.add(list.popFront());
			newState.setPendingInput(Utils.asSList(list));		
		}
		else { 
			queryResult = 
				peek(
					filterCandidates(originalResult,newState),
					newState
				);
			//System.err.println("+++ queryResult "+queryResult);
			int xpCount = objectsExpected(newState);
			if (queryResult.size() > xpCount) {
			//if (queryResult.size() > 1) {
				//System.err.println("extra relations");
				createExtraRelationsError(previousState,advancedWith,newState,queryResult,xpCount);
				throw new QueryResultException("More than the expected number of objects in query result");
				//return new HashSet();
			} 
			/* else if (originalResult.difference(newState.cover()).size() < objectsExpected(newState)) {
				//createMissingRelationError(previousState,advancedWith,newState,originalResult);
				if (!coverConflictSet.isEmpty() &&
					originalResult.difference(newState.cover()).isEmpty()) {
					System.err.println("cover error 1");
					createCoverError(previousState,newState,coverConflictSet);
				} else { 
					System.err.println("missing relation 1");
					createMissingRelationError(previousState,advancedWith,newState);
				}
				return new HashSet();
			} */
			// check for empty queryresult and non-empty cover conflict set
			else if (queryResult.isEmpty() && !newState.productionVariant().isOptional(newState.dot())) {
				//if (!coverConflictSet.isEmpty()) {
				//	System.err.println("cover error 2");
				//	createCoverError(previousState,newState,coverConflictSet);
				//} else {
				
				//System.err.println("%%% missing relation 2");
				createMissingRelationError(previousState,advancedWith,newState);
				
				//}
			}
		}
		return queryResult;
	}                 
		

	
	protected HashSet findRoots() {
		HashSet roots = new HashSet();
		for (SListIterator state=_parseTable.inActiveStatesAt(startObject).begin();
			!state.atEnd();state.advance()) {
			
			if (((InActiveState)state.get()).category().equals(_grammar.startSymbol()) &&
				((InActiveState)state.get()).originator()==null)
				
				roots.add(state.get());
		}
		return roots;

/* 		HashSet roots = new HashSet();
		for (HashSetIterator object=_input.objects().begin();
		     !object.atEnd();object.advance())
			for (SListIterator state=_parseTable.inActiveStatesAt(object.get()).begin();
				 !state.atEnd();state.advance()) {
				 if (((InActiveState)state.get()).category().equals(_grammar.startSymbol()) &&
				     ((InActiveState)state.get()).originator()==null &&
				     ((InActiveState)state.get()).cover().size() == _input.objects().size())
				 	roots.add(state.get());
			}
		return roots;
 */	
 	}	

	protected SList findInits() {
		SList inits = new SList();
		for (SListIterator state=_parseTable.activeStatesAt(startObject).begin();
			 !state.atEnd();state.advance()) {
			if (((ActiveState)state.get()).productionVariant().production().lhs().equals(_grammar.startSymbol()) &&
				((ActiveState)state.get()).originator()==null &&
				((ActiveState)state.get()).dot()==1)
				inits.add(state.get());
		}
		return inits;
	}	
	
	protected void handlePotentialAmbiguity(ParseState newState,SList slot) {
		SListIterator duplicate = slot.find(newState);
		if (!newState.previous().isEmpty()) {
			createPathAmbiguityError((ParseState)duplicate.get());
		}
	}
	
	protected void initParse(IndexedMDSet input,
						 	 RelationalGrammarImplementation grammar)
	throws InvalidGrammarException { 
		super.initParse(input,grammar);
		_input_ambiguities = new SList();
		_grammar_ambiguities = new HashMap();
		_island_objects = new HashSet();
		oldErrors = new HashSet();
		allDeadEnds = new HashSet();
		coverConflictSet = new HashSet();
	}

 	protected Array initStates() {
		Array states = new Array();
		for (HashSetIterator attr=_grammar.fixedStartAttributes().begin();
		     !attr.atEnd();attr.advance())
		    Copying.copy(
		    	_predictionTable.statesFor(
		    		_grammar.startSymbol(),
		    		(String)attr.get()
		    	),
		    	states
		    );
		for (ArrayIterator i_state=states.begin();!i_state.atEnd();i_state.advance()) {
			((ActiveState)i_state.get()).creator("init");
			((ActiveState)i_state.get()).surrogate(new Integer(0));
		}
		_inits = states; 
		return states;
	}
	 
	protected Container filterErrorDescriptions(Array dead_ends) {
		return 
			Transforming.collect(
				Filtering.select(
					dead_ends,
					new UnaryPredicate () {
						public boolean execute(Object dead_end) {
							return ((DeadEnd)dead_end).state().errorDescription() != null;
						}
					}
				),
				new UnaryFunction () {
					public Object execute(Object dead_end) {
						return ((DeadEnd)dead_end).state().errorDescription();
					}					
				}		
			);
	}
	
	private boolean checkRecursion(SList complementList,InActiveState iState) {
		boolean result = true;
		//System.err.println(complementList);
		for (SListIterator complement=complementList.begin();
			 !complement.atEnd();complement.advance()) {
			 result = result && iState.contains((InActiveState)complement.get());
		}
		return result;
	}
	
	protected void newItem(ActiveState a_state, InActiveState i_state,
	                       Sequence new_items,String creator)
	throws AnalysisException {
		AgendaItem item = advance(a_state,i_state);
		if (item != null) {	
			if (!checkRecursion(a_state.completions(),i_state))
				createCompletionAmbiguityError(a_state);
			a_state.addToCompletions(i_state);
			new_items.pushBack(item);
			item.state().creator(creator);
		}	
	}	
	
	private SList extraneousRelations() {
		HashSet tuples = 
			(new RelationCoverChecker()).check(_input);
		SList descriptors = new SList();
		for (HashSetIterator t = tuples.begin();!t.atEnd();t.advance())
			descriptors.pushBack(
				new ExtraRelation((Tuple)t.get())
			);
		return descriptors;
	}
	
	public ParseResult parse(IndexedMDSet input,
						 	 RelationalGrammarImplementation grammar,
	                     	 Object first_object,PrintWriter log,
	                     	 int max_iterations)
	throws InvalidGrammarException,AnalysisException {
		int counter = 0;
		_comparisons = 0;
		int max_error_count = 10;
		int error_count = 0;
		ParseResult result,tmp;
		MultiFailure multi;
		startObject = first_object;
		initParse(input,grammar);


		initializeAgenda(initStates(),first_object);

		if (log != null) {
			log.println(_predictionTable.toString());
			log.println(agendaToString());
		}
		
		result = null;
		multi = null;
		
		do {
			counter = processInput(max_iterations);
			tmp = getResult();
			//System.err.println(tmp);
			if (!tmp.success()) {
				if (multi == null) {
					multi = new MultiFailure();
					result = multi;
				}
				multi.addFailure((SingleFailure)tmp);
				error_count = multi.errorDescriptors().size();
			}
			else 
				if (multi == null)
					result = tmp;
		} while (!tmp.success() &&  error_count<= max_error_count && 
		         recover((SingleFailure)tmp));
		System.err.println("parser done with errorcount "+new Integer(error_count));
		if (error_count > max_error_count)
			System.err.println("too many errors");
		if (log != null) {
			log.println(_parseTable.toString());
			log.println("** "+(new Integer(counter)).toString()+" iterations **");
			log.println("** "+(new Integer(_comparisons)).toString()+" comparisons **");
			if (detourTaken)
				log.println("!! Subsequent ambiguities may be due to error recovery actions !!");
			if (!recoveryDiagnostics.equals(""))
				log.println("%%%\n"+recoveryDiagnostics+"\n%%%");
 		}
 		return result;
	}	

/* 	private AgendaItem patchAttributes(ErrorDescriptor error)
	throws QueryException {
		RecoveryItem new_item = null;
		BranchingState new_state = error.patchedState(_parseTable,this);
		if (new_state != null) {
			new_item = new RecoveryItem(new_state);
			new_item.keys(new HashSet(new_state.indices()));
			System.err.println(new_item);
		}
		return new_item;
	}
 */	
	private BranchingState patchParse(ActiveState state)
	throws QueryException  {
		BranchingState new_state =
			new BranchingState(state.productionVariant().production().lhs());
		new_state.productionVariant(state.productionVariant());
		new_state.inversePermuteDescendants(state.subTrees());
		new_state.cover(new HashSet(state.cover()));
		try {
			new_state.computeAttributeValues(state);
		} catch (IndexOutOfBoundsException e) {
			//System.err.println(e);
			return null;
		}
		for (HashSetIterator index=new_state.indices().begin();!index.atEnd();index.advance())
			if (_parseTable.inActiveStatesAt(index.get()).contains(new_state))
				return null;
		if (!tryEvaluatePredicates(state,new_state))
			return null; 
		new_state.addRelationTuples(state.relationTuples());
		new_state.previous(state.getTruePrevious());
		//state.next(new_state);
		new_state.creator("patch");
		return new_state;	
	}

	private RecoveryItem patchAttributes(ActiveState state)
	throws QueryException {
		RecoveryItem new_item = null;
		//System.err.println("patching: "+state);
		BranchingState new_state = patchParse(state);
		//if (new_state != null)
	 	//	System.err.println("patctAttributes:\n"+new_state.toString());
		if (new_state != null) {
			new_item = new RecoveryItem(new_state);
			new_item.keys(new HashSet(new_state.indices()));
			//System.err.println(new_item);
		}
		else
			state.discard();
		//System.err.println("(patchAttributes) got: "+new_item);		
		return new_item;
	}

	
	private RecoveryItem findDetour(ActiveState errorState)
	throws QueryException {
		RecoveryItem new_item = null;
		if (!errorState.dotAtEnd()) {
			//System.err.println("(findDetour) errorState: "+errorState);
			int maskPos;
			maskPos = errorState.dot();
			for (;maskPos<errorState.productionVariant().production().rhs().size() &&
			  	 new_item == null && !errorState.bindsAttribute(maskPos); maskPos++) {
			  	errorState.maskSymbolAt(maskPos);   
				SList constraints = errorState.constraintsAt(maskPos+1);
				//System.err.println("in Detour:\n"+constraints.toString());
				if (!constraints.isEmpty()) {
					ActiveState new_state = new ActiveState(errorState.productionVariant());
					new_state.dot(maskPos + 1);
					DescendantList sub_trees = new DescendantList(errorState.subTrees());
					for (int i=errorState.dot();i<new_state.dot();i++) {
						EmptyLeaf leaf = new EmptyLeaf("##empty##");
						leaf.cover(new HashSet());
						sub_trees.addDescendant(i,leaf);
					}
					new_state.subTrees(sub_trees);
					new_state.previous(errorState.getTruePrevious());
					new_state.originator(errorState.originator());
					new_state.creator("detour");
					new_state.addRelationTuples(errorState.relationTuples());
					new_state.addToMask(errorState.mask());
					HashSet result_set;
					try {
						new_state.queryConstraint((Constraint)constraints.front());
						result_set = launchQuery(new_state);
					}
					catch (InvalidQueryArgumentsException e) {
						throw(new QueryException(e.getMessage()));
					}
					try {
						result_set = processQueryResult(errorState,null,new_state,result_set);
						if (result_set.size()==1) {
			    		//errorState.next(new_state);
							new_item = new RecoveryItem(new_state);
							new_item.keys(result_set);
						}
					} catch (QueryResultException e) {
						new_item = null;
					}
				}
			}
			if (new_item == null) 
				for (int i = errorState.dot();i<maskPos;i++)
					errorState.unMaskSymbolAt(i);	
		}
		if (new_item == null)
			errorState.discard();
				
		//System.err.println("(findDetour) got: "+new_item);		
		return new_item;
	}

	protected class IsRecoverable implements UnaryPredicate {
		public boolean execute(Object error) {
			return ((ErrorDescriptor)error).isRecoverable();
		}
	}

	private HashMap mapStateToError(HashSet errors) {
		//System.err.println(errors);
		HashMap map = new HashMap();
 		for (ForwardIterator error=errors.start();!error.atEnd();error.advance()) {
 			ActiveState loc = ((ErrorDescriptor)error.get()).incompleteParse();
 			if (loc != null)
 				map.put(loc,error.get());
 		}
 		return map;
	}


	private SList statesFromErrors(HashSet errors) {
		SList states = new SList();
		for (Enumeration e = errors.elements();e.hasMoreElements();)
			states.pushBack(((ErrorDescriptor)e.nextElement()).incompleteParse());
		return states;
	}

	private SList originatorStates(SList states) {
		HashSet origins = new HashSet();
		for (Enumeration elts=states.elements();elts.hasMoreElements();)
			Copying.copy(((ActiveState)elts.nextElement()).wherePredicted(),origins);
		return Utils.asSList(origins);
	}
	
	private boolean equivalentAttributeValues(Enumeration items) {
		SList recItems = Utils.asSList(Utils.enumerationAsContainer(items));
		InActiveState rep = (InActiveState)((RecoveryItem)recItems.front()).state();
		for (SListIterator item=recItems.begin();!item.atEnd();item.advance())
			if (!rep.attributeValues().equals(
					((InActiveState)((RecoveryItem)item.get()).state()).attributeValues()	
				)
			   )
			 	return false;	
		return true;
	}
	
	private Enumeration getPatchedItems(HashMap eqMap) {
		for (Enumeration key=eqMap.keys();key.hasMoreElements();) {
			Object currentKey = key.nextElement();
			if (equivalentAttributeValues(eqMap.values(currentKey)))
				return eqMap.values(currentKey);
		}
		return (new SList()).elements(); // no items
	}
		
	protected boolean recover(SingleFailure fault) throws QueryException {
 		SList items = new SList();
		HashSet errors =
			Utils.asHashSet(fault.errorDescriptors()).difference(oldErrors);
		if (errors.isEmpty())
			return false;
		//System.err.println("In recover");
		//System.err.println(Utils.contentsToString(fault.errorDescriptors(),"\n"));
		//System.err.println("errors: "+Utils.contentsToString(errors,"\n"));
		
		if (findRoots().isEmpty() && Finding.every(errors,new IsRecoverable())) {
 			SList states1 = statesFromErrors(errors);
 			HashMap stateToErrorMap = mapStateToError(errors);
 			SList states2 = 
 				Utils.asSList(Utils.enumerationAsContainer(stateToErrorMap.keys()));

  			while (!states1.isEmpty() && items.isEmpty()) {
  				//System.err.println("states1: "+Utils.contentsToString(states1,"\n"));
  				//System.err.println("states2: "+Utils.contentsToString(states1,"\n"));
 				SList originators1 = originatorStates(states1);
 				SList originators2 = originatorStates(states2);
				HashMap equiClasses = new HashMap(true);
 				for (ForwardIterator state=states1.begin();!state.atEnd();state.advance()) {
					RecoveryItem item =
						patchAttributes((ActiveState)state.get());
					if (item != null) {
						if (!((ActiveState)state.get()).wherePredicted().isEmpty())
							for (ForwardIterator predictedAt = ((ActiveState)state.get()).wherePredicted().begin();
						     	!predictedAt.atEnd();predictedAt.advance())
						     	equiClasses.add(predictedAt.get(),item);
						else
				     		equiClasses.add("no key",item);
						//System.err.println("equiClasses: "+equiClasses.toString());
				    }
				}
				if (!equiClasses.isEmpty()) {
					//System.err.println("After patch: "+equiClasses.toString());
					Enumeration patchedItems = getPatchedItems(equiClasses);
					if (patchedItems.hasMoreElements())
						items.pushBack(
							Utils.asSList(Utils.enumerationAsContainer(patchedItems)).front()						
						);
				}
				if (items.isEmpty()) {
					equiClasses.clear();
 					for (ForwardIterator state=states2.begin();!state.atEnd();state.advance()) {
						RecoveryItem item = findDetour((ActiveState)state.get());
						if (item != null) {
							item.setError((ErrorDescriptor)stateToErrorMap.get(state.get()));
							if (!((ActiveState)state.get()).wherePredicted().isEmpty())
								for (ForwardIterator predictedAt = ((ActiveState)state.get()).wherePredicted().begin();
						     		!predictedAt.atEnd();predictedAt.advance())
						     		equiClasses.add(predictedAt.get(),item);
						    else
						    	equiClasses.add("no key",item);
						}
					}
					if (!equiClasses.isEmpty()) {
							//System.err.println("After detour: "+equiClasses.toString());
							Copying.copy(
							Utils.enumerationAsContainer(
								equiClasses.values(equiClasses.keys().nextElement())
							),
							items
						);
						for (Enumeration it=items.elements();it.hasMoreElements();)
							((RecoveryItem)it.nextElement()).clearError();
						detourTaken = true;
					}
				}
				if (items.isEmpty()) {
					states1 = originators1;
					states2 = originators2;
					stateToErrorMap.clear();
				}
			}
			
			for (Enumeration it=items.elements();it.hasMoreElements();)
				_agenda.pushBack(it.nextElement());
		} 
		if (items.isEmpty())
			recoveryDiagnostics =
				"cannot recover from:\n"+
				Utils.listContents(errors,"\n","\nand\n");
			
		Copying.copy(errors,oldErrors);
		//System.err.println("oldErrors: "+Utils.contentsToString(oldErrors,"\n"));
		//System.err.println("Out recover");

		return !items.isEmpty();
	}


	protected ParseResult getResult() {
		HashSet parse_trees = findRoots();
		SList extras = extraneousRelations();
		if (!_input_ambiguities.isEmpty()) {
			//System.err.println("Parse Error--ambiguities detected:\n"+
			//            Utils.contentsToString(_input_ambiguities,"\n"));
			return new SingleFailure(_input_ambiguities,"Ambiguities detected");

		}
		if (!_grammar_ambiguities.isEmpty()) {
			// System.err.println("Parse Error--ambiguities detected:\n"+
			//            Utils.contentsToString(_grammar_ambiguities,"\n")); 
			SList ambList = new SList();
			for (Enumeration v=_grammar_ambiguities.elements();v.hasMoreElements();)
				ambList.add(v.nextElement());
			return new SingleFailure(ambList,"Ambiguities detected");

		}
		if (parse_trees.isEmpty()) {
			Array sub_parses = selectLongestSubParses();

			/* System.err.println(
				"Errors:\n"+
				Utils.contentsToString(sub_parses,"\n")
			); */

			Container error_descriptors = filterErrorDescriptions(sub_parses);
			if (!error_descriptors.isEmpty())
				return new SingleFailure(error_descriptors);
			
			SingleFailure faults=new SingleFailure("Other failure");
			if (allDeadEnds.isEmpty()) {
				HashSet unprocessed = _input.objects().difference(_parseTable.coveredInput());
				if (!unprocessed.isEmpty()) {
					faults.addError(new UnprocessedInput(unprocessed));
				}
			}
			return faults;
			
			//return new Failure(sub_parses);
		}
		else { // tentative success
			if (parse_trees.size()>1) {
				//System.err.println(parse_trees);
				SList inclEquivs = CompletionAmbiguity.inclusionEquivalences(Utils.asSList(parse_trees));
				//System.err.println("inclEquivs: " + inclEquivs);
				if (inclEquivs.size() > 1) {
					SingleFailure fault=
						new SingleFailure("Ambiguous parse result");
					fault.addError(
						new GrammarAmbiguity(
							(ActiveState)((InActiveState)inclEquivs.start().get()).previous().begin().get(),
							inclEquivs
						)
					);
					return fault;
				}
			}			
			SingleFailure faults=new SingleFailure("Unprocessed input");
			HashSet unprocessed = _input.objects().difference(_parseTable.coveredInput());
			if (!unprocessed.isEmpty() && allDeadEnds.isEmpty()) {
				faults.addError(new UnprocessedInput(unprocessed));
			} 
			if (!extras.isEmpty() && allDeadEnds.isEmpty()) {
				faults.addErrors(extras);
			} 
			if (!faults.errorDescriptors().isEmpty())
				return faults;
				
			return new Success(parse_trees);
		}
	}

	protected boolean satisfiesCurrentConstraints(Object input_object,
	                                              ActiveState state)
	throws QueryException {
		SList constraints = state.currentConstraints();
			//state.productionVariant().constraintsAt(state.dot());   
		if (constraints.isEmpty())
			return true;
		else {
			boolean result = true;
			HashSet tuples;
			try {
				for (SListIterator c=constraints.begin();!c.atEnd();c.advance()) {
					Array qArgs = 
						buildQueryArguments(
							state,input_object,(Constraint)c.get()
						);
					tuples = 
		            	_input.getRelation(((Constraint)c.get()).relation())
		            		.getTuples(qArgs);
		            Enumeration tuple = tuples.elements();
		            while (tuple.hasMoreElements())
		            	state.addRelationTuple((Tuple)tuple.nextElement());
		            // if (tuples.size()>0)
		            //	((Tuple)tuples.begin().get()).markTuple();
 					result = result && tuples.size()==Utils.joinSize(qArgs);
				}
		  	}
		  	catch (InvalidQueryArgumentsException e) {
		  		throw(new QueryException(e.getMessage()));
		  	}
			return result;
		 }                             
	                                                                
	}

	protected boolean satisfiesCurrentConstraints(InActiveState i_state,
	                                              ActiveState state)
	throws QueryException {
		SList constraints = state.currentConstraints();
			//state.productionVariant().constraintsAt(state.dot());   
		if (constraints.isEmpty())
			return true;
		else {
			boolean result = true;
			HashSet tuples;
			try {
				for (SListIterator c=constraints.begin();!c.atEnd();c.advance()) {
					Array qArgs = 
						buildQueryArguments(
							state,i_state,(Constraint)c.get()
						);
					tuples = 
		            	_input.getRelation(((Constraint)c.get()).relation())
		            		.getTuples(qArgs);
		            Enumeration tuple = tuples.elements();
		            while (tuple.hasMoreElements())
		            	state.addRelationTuple((Tuple)tuple.nextElement());
		            //if (tuples.size()>0)
		            //	((Tuple)tuples.begin().get()).markTuple();
					result = result && tuples.size()==Utils.joinSize(qArgs);
				}
		  	}
		  	catch (InvalidQueryArgumentsException e) {
		  		throw(new QueryException(e.getMessage()));
		  	}
			return result;
		 }                             
	                                                                
	}

	protected Array selectLongestSubParses() {
		Array sub_parses = new Array();
		HashSet deadEnds = new HashSet();
		for (ForwardIterator state=findInits().begin();!state.atEnd();state.advance())
			Copying.copy(
				collectDeadEnds((ParseState)state.get(),new HashSet(),new ParsePath()),
				deadEnds
			);
			// filter out those dead ends that have been previously considered
		//System.err.println(deadEnds);	
		HashSet newDeadEnds = deadEnds.difference(allDeadEnds);
		//System.err.println("****************\n"+newDeadEnds);
		//System.err.println("****************");
		Copying.copy(newDeadEnds,allDeadEnds);
		Copying.copy(newDeadEnds,sub_parses);
		Sorting.sort(
			sub_parses,
			new BinaryPredicate() {
				public boolean execute(Object arg1,Object arg2) {
					return ((DeadEnd)arg1).accCover().size() >
						   ((DeadEnd)arg2).accCover().size();
				}
			}
		);
		//System.err.println(sub_parses);
		if (sub_parses.isEmpty()) 
			return sub_parses;
		else
			return 
			(Array)Filtering.select(
				sub_parses,
				new BindFirstPredicate(
					new BinaryPredicate() {
						public boolean execute(Object arg1,Object arg2) {
							return ((Integer)arg1).equals(
						   				new Integer(((DeadEnd)arg2).accCover().size())
						   		   );
						}
					},
					new Integer(((DeadEnd)sub_parses.front()).accCover().size())
				)
			);
	}
	
	protected HashSet getPredictionAttributes(ActiveState state) {
		if (state.dot() == 1)
			return state.toBeBoundAttrs().intersection(_grammar.fixedStartAttributes());
		else
			return state.toBeBoundAttrs();
	}

}