package relap.Analyzer;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Finding;
import com.objectspace.jgl.predicates.BindSecondPredicate;
import relap.LanguageModel.ObjectReference;
import relap.LanguageModel.AttributeAssignment;
import relap.LanguageModel.ExistenceConstraint;
import relap.LanguageModel.Constraint;
import relap.LanguageModel.Verifiable;
import relap.LanguageModel.Tuple;
import relap.LanguageModel.InvalidQueryArgumentsException;
import java.util.Enumeration;

abstract public class InActiveState extends ParseState {


	protected HashSet _cover;
	protected String _category;
	protected HashMap _attributeValues;
	protected InActiveState _parent;
	protected SyntaxTreeNode _semantics;
	
	public InActiveState(String symbol_category) {
		_category = symbol_category;
		_cover = null;
		_attributeValues = null;
		_parent = null;
		_semantics = null;
	}
	 	
	public InActiveState() {
		_cover = null;
		_attributeValues = null;
		_parent = null;
		_semantics = null;
	}
	
	public void setAttributeValue(String attribute,Object value) {
    	if (_attributeValues == null)
    		_attributeValues = new HashMap();
    	_attributeValues.add(attribute,value);
   	}

	public HashMap attributeValues() {
		return _attributeValues;
	}
   
	public void attributeValues(HashMap map) {
		_attributeValues = map;
	}

	public String category() {
		return _category;
	}
   
	public void category(String symbol_category) {
		_category = symbol_category;
	}
  
  	public void setSemantics(SyntaxTreeNode node) {
  		_semantics = node;
  	}
  	
  	public SyntaxTreeNode getSemantics() {
  		return _semantics;
  	}
  	
	public void computeAttributeValues(ActiveState a_state,InActiveState i_state) {
		for (HashMapIterator assgn=a_state.productionVariant().attributeAssignments().begin();
		 	 !assgn.atEnd();assgn.advance()) {
			ObjectReference ref = ((AttributeAssignment)assgn.value()).value();
			if (ref.position()>a_state.subTrees().size())
				setAttributeValue((String)assgn.key(),i_state.referencedValue(ref));
			else 
				setAttributeValue((String)assgn.key(),a_state.dereference(ref).start().get());	 
		}
	}

	public void computeAttributeValues(ActiveState a_state)
	throws IndexOutOfBoundsException {
		for (HashMapIterator assgn=a_state.productionVariant().attributeAssignments().begin();
		 	 !assgn.atEnd();assgn.advance()) {
			ObjectReference ref = ((AttributeAssignment)assgn.value()).value();
			if (ref.position() > a_state.subTrees().size())
				throw new IndexOutOfBoundsException();
			setAttributeValue((String)assgn.key(),a_state.dereference(ref).start().get());	 
		}
	}

	// answer true if the receiver contains the argument object as
	// a subtree (leaf or branch)
	abstract public boolean contains(InActiveState state);
	
	public HashSet cover() {
		return _cover;
	}
   
	public void cover(HashSet set) {
		_cover = set;
	}
 
	public String describe() {
		StringBuffer buf = new StringBuffer("#( "+_category+" "+")");
		return buf.toString();
	}

	public boolean equals(Object state) {
		return
			super.equals(state) || 
			( getClass()==state.getClass() &&
		  	_category==((InActiveState)state).category() &&
		  	_cover.equals(((InActiveState)state).cover()) );
	}

	public SList evaluateNonExistenceConstraints(IndexedMDSet input)
	throws QueryException {
		SList failedPredicates = new SList();
		for (ArrayIterator neC = _productionVariant.production().notExistsConstraints().begin();
		 	 !neC.atEnd();neC.advance()) {
		 	try {
		 		if (Finding.every(mask(),new BindSecondPredicate(new IsNotMasked(),(ExistenceConstraint)neC.get())))
		 			if (!((ExistenceConstraint)neC.get()).verify(this,input)) {
		 				//System.err.println("failed: "+neC.get());
		 				Enumeration fData = ((Verifiable)neC.get()).getFailureData().elements();
		 				while (fData.hasMoreElements()) {
		 					failedPredicates.pushBack(
		 						new NotExistsError(
		 							this,
		 							(Verifiable)neC.get(),
		 							(FailureData)fData.nextElement()
		 						)
		 					);
		 					//System.err.println("failedPredicates.back(): "+failedPredicates.back());
		 				}	
		 			}
		 	} 
		 	catch (InvalidQueryArgumentsException e) {
		 		throw(new QueryException(e.getMessage()));
		 	}
		}
		return failedPredicates;
	
	}

	public SList evaluatePredicates(IndexedMDSet input)
	throws QueryException {
		SList failedPredicates = new SList();
		for (ArrayIterator c = _productionVariant.production().predicates().begin();
		 	 !c.atEnd();c.advance()) {
		 	try {
		 		//System.err.println(")))) mask() "+mask());
		 		//System.err.println("c "+c.get());
		 		if (Finding.every(mask(),new BindSecondPredicate(new IsNotMasked(),(Verifiable)c.get())))
		 			if (!((Verifiable)c.get()).verify(this,input)) {
		 				//System.err.println(">>> Failed: "+c.get());
		 				Enumeration fData = ((Verifiable)c.get()).getFailureData().elements();
		 				//System.err.println("with "+((Verifiable)c.get()).getFailureData());
		 				while (fData.hasMoreElements())
		 					failedPredicates.pushBack(
		 						new PredicateEvaluationError(
		 							this,
		 							(Verifiable)c.get(),
		 							(FailureData)fData.nextElement()
		 						)
		 					);		
		 			}
		 	} 
		 	catch (InvalidQueryArgumentsException e) {
		 		throw(new QueryException(e.getMessage()));
		 	}
		}
		//System.err.println(">>> Failed: "+failedPredicates);
		return failedPredicates;
	
	}
	
 	public int hashCode() {
		return _category.hashCode();
	}
 
	public HashSet indices() {
        //Return the set of InputObjects bound to the expander attributes of the receiver.
    	HashSet values = new HashSet();
    	for (HashMapIterator it = _attributeValues.begin();!it.atEnd();it.advance())
        	values.add(it.value());
		return values;
	}
  
	public void initCopy(ParseState new_state) {
		super.initCopy(new_state);
    	((InActiveState)new_state).category(_category);
    	((InActiveState)new_state).parent(_parent);
    	((InActiveState)new_state).cover(new HashSet(_cover));
    	((InActiveState)new_state).attributeValues(new HashMap(_attributeValues));
    	((InActiveState)new_state).setSemantics(_semantics);
    }
	public boolean isActive() {
		return false;
	}
   
	public InActiveState parent() {
		return _parent;
	}
  
	public void parent(InActiveState parent) {
		_parent = parent;
	}

 	public void setAncestry(InActiveState parent) {
 		parent(parent);
 	}
 	
 
/*
performSemantics
    "Perform the semantic action stored in the production that the receiver represents"
     | block |
        (block := productionVariant production semantics) notNil ifTrue: [ block value: self ]!
*/
  
	abstract public Object referencedValue(ObjectReference reference);
	
	
	public String toString() {
		StringBuffer buf = new StringBuffer("#( "+_category+" ");
    	if (_attributeValues != null) 
    		for (HashMapIterator it = _attributeValues.begin();!it.atEnd();it.advance())
        		buf.append(" :"+it.key().toString()+" "+it.value().toString());
    	buf.append(")\n   (");
    	for (Enumeration it = cover().elements();it.hasMoreElements();)
    		 buf.append(" "+it.nextElement().toString());
    	buf.append(")\n   creator: "+_creator.toString()+" from: "+previousToString());
    	buf.append("\n   "+originatorToString());
    	return buf.toString();
    }
  
	abstract public String toString(int steps,int indent);

	public Object valueOfAttribute(String attribute) {
		return _attributeValues.get(attribute);
	}

}