package relap.Analyzer;

import com.objectspace.jgl.*;
import relap.LanguageModel.Relation;
import java.util.Enumeration;
import relap.Editor.ObjectToSymbolMap;

public class IndexedMDSet {

	private HashSet _objects;
	private HashMap _relations;
	private HashMap _object_to_terminal;
	private ObjectToSymbolMap _object_class_to_terminal;
	
	public IndexedMDSet() {
		 _objects = new HashSet();
		 _relations = new HashMap();
		 _object_to_terminal = new HashMap();
		 _object_class_to_terminal = null;
	}

	public IndexedMDSet(ObjectToSymbolMap map) {
		 _objects = new HashSet();
		 _relations = new HashMap();
		 _object_to_terminal = new HashMap();
		 _object_class_to_terminal = map;
	}
	
	public void setObjectClassToTerminalMap(ObjectToSymbolMap map) {
		 _object_class_to_terminal = map;	
	}
	
	public String describeTerminal(String terminal) {
		if (_object_class_to_terminal == null)
			return terminal;
		else
			return _object_class_to_terminal.describeSymbol(terminal);
	}

	public HashSet describeTerminals(Container terminals) {
		HashSet result = new HashSet();
		Enumeration terminal = terminals.elements();
		while (terminal.hasMoreElements())
			result.add(describeTerminal((String)terminal.nextElement()));
		return result;
	}
	
	public void addObject(Object input_object) {
		_objects.add(input_object);	
	}
 
	public void addRelation(Relation relation) {
		if (_relations.add(relation.name(),relation)==null)
			_objects = _objects.union(relation.objects());
	}
	
	public void setTerminalType(Object input_object,String terminal) {
		_object_to_terminal.put(input_object,terminal);
	}
	
	public String getTerminalType(Object input_object) {
		return (String)_object_to_terminal.get(input_object);
	}
		
	public Relation getRelation(String relation) {
		return (Relation)_relations.get(relation);
	}
	
	public Enumeration getRelationNames() {
		return _relations.keys();
	}
	
	public HashSet objects() {		
		return _objects;
	}
	
	public void updateObjects() {
    	for (HashMapIterator it=_relations.begin();!it.atEnd();it.advance())
    		_objects = _objects.union(((Relation)it.value()).objects());	
    }
    
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("\n-- MD-set --");
    	buf.append("\nObjects:");
    	for (ForwardIterator it = _objects.begin();!it.atEnd();it.advance())
    		buf.append("\n    "+it.get().toString());
    	buf.append("\nRelations:");
    	for (HashMapIterator it=_relations.begin();!it.atEnd();it.advance()){
    		buf.append("\n"+((Relation)it.value()).name()+":");
    		buf.append(((Relation)it.value()).tuplesToString());
    	}
    	buf.append("\n-- end --");
		return buf.toString();
	}
}