package relap.Analyzer;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.*;

public class MiddleNodeIterator implements SyntaxTreeIterator {

	MiddleNode fCurrentRoot;
	SyntaxTreeNode fCurrentNode;
	ForwardIterator fChild;
	Stack fNodes;
	
	public MiddleNodeIterator(MiddleNode root) {
		fNodes = new Stack();
		newRoot(root);
	}
	
	public int depth() {
		return fNodes.size();
	}
	
	protected ForwardIterator linearizeDescendants(SyntaxTreeNode node) {
		SList tmp = new SList();
		for (ForwardIterator child = node.getDescendants().start();!child.atEnd();
		     child.advance()) 
			if (child.get() instanceof SList)
				Copying.copy((SList)child.get(),tmp);
			else
				tmp.pushBack(child.get());
		return tmp.start();
	}
	
	protected void newRoot(MiddleNode newRoot) {
		if (fCurrentRoot != null) 
			fNodes.push(new Pair(fCurrentRoot,fChild));
		fCurrentRoot = (MiddleNode)newRoot;
		fCurrentNode = fCurrentRoot;
		fChild = linearizeDescendants(fCurrentRoot);
	}

	protected boolean restoreRoot() {
		if (fNodes.isEmpty())
			return false;
		Pair oldState = (Pair)fNodes.pop();
		fCurrentRoot = (MiddleNode)oldState.first;
		fChild = (ForwardIterator)oldState.second;
		return true;
	}

	public boolean atEnd() {
		return fCurrentNode == null;
	}
	
	public SyntaxTreeNode current() {
		return fCurrentNode;
	}
	
	public void advance() {
		if (!fChild.atEnd()) {
			fCurrentNode = (SyntaxTreeNode)fChild.get();
			fChild.advance();
			if (fCurrentNode instanceof MiddleNode)
				newRoot((MiddleNode)fCurrentNode);
		}
		else if (restoreRoot())
			advance();
		else
			fCurrentNode = null;
	}
}
