package relap.Analyzer;

import relap.LanguageModel.OrderingVariant;
import relap.LanguageModel.ObjectReference;
import relap.LanguageModel.RelationalGrammarImplementation;
import relap.LanguageModel.Tuple;
import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Transforming;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Applying;
import com.objectspace.jgl.algorithms.Removing;
import com.objectspace.jgl.predicates.BindFirstPredicate;
import com.objectspace.jgl.predicates.IdenticalTo;
import java.lang.Class;

public abstract class ParseState 
	implements SymbolSequenceHolder {
	protected String _creator; 
	protected SList _up;
	protected SList _previous; 
	protected ActiveState _originator; 
	protected Integer _surrogate;
	protected ErrorDescriptor _errorDescriptor;
	protected OrderingVariant _productionVariant;
	protected SList relationTuples;
	protected SList mask;
	
	public ParseState() {
		_surrogate = new Integer(-1);
		relationTuples = new SList();
		mask = new SList();
	}

    //Answer a copy of the receiver

	public ParseState copy() {
		//System.err.println(this);
    	ParseState instance;
    	try {
    		instance = (ParseState)getClass().newInstance();
    	}
    	catch (IllegalAccessException e) {
    		return null;
    	}
    	catch (InstantiationException e) {
    		return null;
    	}
    	initCopy(instance);
    	return instance;
    }
    
    abstract public HashSet cover();
    
    public String creator() { return _creator; }

	public void creator(String creator) {
	    _creator = creator;
	}

	abstract public String describe();
	
	public void discard() {
		// a hook; do nothing (another case of caseless programming)
		// discard() in IntermediateState unlinks the state from the parse
	}
	
	public ErrorDescriptor errorDescription() {
    	return _errorDescriptor;
    }

	public void errorDescription(ErrorDescriptor descriptor) {
    	_errorDescriptor = descriptor;
    }
  
	public void clearError() {
    	_errorDescriptor = null;
    }
  
	public RelationalGrammarImplementation grammar() {
        return _productionVariant.production().grammar();
    }
       
	protected void initCopy(ParseState newParseState) {
		newParseState.productionVariant(_productionVariant);
    	newParseState.creator(_creator);
    	newParseState.up(new SList(up()));
    	newParseState.originator(_originator);
    	newParseState.previous(new SList(previous()));
    	newParseState.surrogate(_surrogate);
    	newParseState.errorDescription(_errorDescriptor);
    	newParseState.addRelationTuples(relationTuples);
    	newParseState.mask(new SList(mask));
    }
    
	public HashSet intersectionOfCovers(ParseState state) {
		return cover().intersection(state.cover());
	}
	
	
	abstract public boolean isActive();
	
	public SList mask() {
		return mask;
	}
	
	public void mask(SList list) {
		mask = list;
	}
	
	public void addToMask(Container items) {
		Copying.copy(items,mask);
	}
		
	public void unLinkFromParse() {
		_previous = null;
		_up = null;
	}	
	
	
	public ActiveState originator() {
    	return _originator;
    }

	public void originator(ActiveState state) {
    	_originator = state;
    }

	public String originatorToString() {
    	StringBuffer origin = new StringBuffer();
    	if (_originator != null) {
        	if(_originator.surrogate() != null)
                origin.append("originator: "+_originator.surrogate().toString());
            else 
            	origin.append("originator: <null>");
        }
        else origin.append("no originator");
    	return origin.toString();
    }

	public SList previous() {
		if (_previous == null)
			_previous = new SList();
		return _previous;
	}
  
	public SList getTruePrevious() {
		SList list = new SList();
		list.pushBack(this);
		return list;
	}
  
	public void previous(ParseState prev) {
		if (_previous==null)
			_previous = new SList();
		_previous.pushBack(prev);
	}

	public void previous(SList prevs) {
		_previous = prevs;
	}

	public String previousToString() {
		StringBuffer prev  = new StringBuffer();
    	if (_previous != null) {
    		prev.append("previous:");
    		for (SListIterator p=_previous.begin();!p.atEnd();p.advance())
        		if(((ActiveState)p.get()).surrogate() != null)
                	prev.append(" "+((ActiveState)p.get()).surrogate().toString());
            	else 
            		prev.append(" <null>");
        }
        else prev.append("no previous");
    	return prev.toString();
    }
 
	public OrderingVariant productionVariant() {
		return _productionVariant;
	}
   
	public void productionVariant(OrderingVariant variant) {
		_productionVariant = variant;
	}
	
	public void addRelationTuples(Container tuples) {
		Copying.copy(tuples,relationTuples);
	}
	
	public void addRelationTuple(Tuple tuple) {
		relationTuples.pushBack(tuple);
	}
	
	public SList relationTuples() {
		return relationTuples;
	}
	
	public void markRelationTuples() {
		Applying.forEach(relationTuples,
			new UnaryFunction () {
				public Object execute(Object tuple) {
					((Tuple)tuple).markTuple();
					return tuple;
				}
			}
		);
	}
   
    public static UnaryFunction getSurrogate() {
    	return new UnaryFunction() {
        	public Object execute(Object obj) {
				return ((ParseState)obj).surrogate();
            }
        };
    }
    	
    public Integer surrogate() {
		return _surrogate;
	}
  
	public void surrogate(Integer number) {
		_surrogate = number;
	}
  
	public SList up() {
		if (_up==null)
			_up = new SList();
		return _up;
	}

	public SList upAsSurrogate() {
		return (SList)Transforming.collect(_up,getSurrogate());
	}
	
	public void up(ParseState upState) {
		if (_up==null)
			_up = new SList();
		_up.pushBack(upState);
	}

	public void up(SList ups) {
		_up = ups;
	}
  
	public String upToString() {
		StringBuffer up  = new StringBuffer();
    	if (_up != null) {
    		up.append("up:");
    		for (SListIterator u=_up.begin();!u.atEnd();u.advance())
        		if(((ActiveState)u.get()).surrogate() != null)
                	up.append(" "+((ActiveState)u.get()).surrogate().toString());
            	else 
            		up.append(" <null>");
        }
        else up.append("no up-state");
    	return up.toString();
    }    

	public HashSet wherePredicted() {
		HashSet locations = new HashSet();
		if (!previous().isEmpty())
			for (ForwardIterator prev=previous().begin();!prev.atEnd();prev.advance())
				Copying.copy(((ActiveState)prev.get()).wherePredicted(),locations);
		else
			Copying.copy(up(),locations);
		return locations;	
	}
	
	abstract public Sequence dereference(ObjectReference reference);

}
