package relap.Analyzer;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Finding;
import com.objectspace.jgl.predicates.*;

public class ParseTable {

	private HashMap _table;
	private HashSet _entries;
	private Array _entry_index;
	
	public ParseTable(HashSet objects) {
		_table=new HashMap();
		_entries=new HashSet();
		_entry_index=new Array();
		for (HashSetIterator o=objects.begin();!o.atEnd();o.advance())
			_table.add(o.get(),new Pair(new HashSet(),new SList()));
	}

	class IsActive implements UnaryPredicate {
		public boolean execute(Object state) {
			return !(state instanceof SkipState) && ((ParseState)state).isActive();
		}
	}

	class IsInActive implements UnaryPredicate {
		public boolean execute(Object state) {
			return !(state instanceof SkipState) && !((ParseState)state).isActive();
		}
	}

	public SList activeStatesAt(Object input_object) {
		return (SList)Filtering.select((SList)((Pair)_table.get(input_object)).second,
									   new IsActive());
	}
	
	public ParseState add(Object input_object, ParseState state) {
	    ParseState newState = state.copy();
	    ((HashSet)((Pair)_table.get(input_object)).first).add(newState);
	    ((SList)((Pair)_table.get(input_object)).second).add(newState);
    	//if (_entries.add(newState)==null) {
    		newState.surrogate(new Integer(_entry_index.size()+1));
    		_entry_index.pushBack(newState);
    	//}
    	return newState;
    }	
   
  public ParseState addSkipState(Object input_object, ActiveState state) {
  	SkipState s = new SkipState();
  	s.copyStateFrom(state);
		return add(input_object,s);
  }
    
    public HashSet coveredInput() {
    	HashSet covered = new HashSet();
    	for (java.util.Enumeration key=_table.keys();key.hasMoreElements();)
    		for (java.util.Enumeration state=inActiveStatesAt(key.nextElement()).begin();
    		     state.hasMoreElements();)
    		     Copying.copy(
    		     	((InActiveState)state.nextElement()).cover(),
    		     	covered
    		     );
    	return covered;
    }

	public SList downStatesFor(ParseState state) {
        //Answer a collection containing all the state entries 
        //in the parse table having aParseState as the 'up''-state.
        	
		return entriesFor(
			new BinaryPredicate() {
				public boolean execute(Object state,Object up_state) {
					return 
						!Finding.findIf(
							((ParseState)state).up(),
							new BindFirstPredicate(new IdenticalTo(),up_state)
						).atEnd();
				}
			},
			state
		); 
	}
	
	public SList entriesFor(BinaryPredicate condition,Object argument) {
        //Answer a collection containing all the states in the table satisfying
        //the argument
    	SList array = new SList();
	    for (HashMapIterator slot=_table.begin();!slot.atEnd();slot.advance()) {
	    	Copying.copy(
	    		Filtering.select((SList)((Pair)slot.value()).second,
	    		                 new BindSecondPredicate(condition,argument)),
	    		array
	    	);
	    }
	    return array;
	}

	public SList inActiveStatesAt(Object input_object) {
		return (SList)Filtering.select(
			(SList)((Pair)_table.get(input_object)).second,
			new IsInActive()
		);
	}
	
	public Object inputObjectForSurrogate(Integer surrogate) {
        //Answer the input object indexing the parse state having anObject as its surrogate.
    	Object input_object = null;
    	HashMapIterator slot=_table.begin();
    	while (input_object == null && !slot.atEnd()) {
    		for (SListIterator state=((SList)((Pair)slot.value()).second).begin();
    			 !state.atEnd();state.advance())
				if (((ParseState)state.get()).surrogate().equals(surrogate)) {
					input_object = slot.key();   
					break;
				}
    		slot.advance();
		} 	
		return input_object;
	}

	public SList nextStatesFor(ParseState state) {
        //Answer a collection containing all the state entries 
        //in the parse table having aParseState as the 'previous'-state.
        	
		return entriesFor(
			new BinaryPredicate() {
				public boolean execute(Object state,Object prev_state) {
					return 
						!Finding.findIf(
							((ParseState)state).previous(),
							new BindFirstPredicate(new IdenticalTo(),prev_state)
						).atEnd();
				}
			},
			state
		);
	}

	public ParseState stateForSurrogate(Integer surrogate) {
        //Answer the state stored to the parse table having surrogate anObject.
		return (ParseState)_entry_index.at(surrogate.intValue()-1);
	}

   
	public SList statesAt(Object input_object) {
		return (SList)((Pair)_table.get(input_object)).second;
	}
   
	public boolean statesAtIncludes(Object input_object,ParseState state) {
		return 
			((HashSet)((Pair)_table.get(input_object)).first).get(state)!=null;
	}
   
	public SList statesCreatedBy(String creator) {
        //Answer a collection containing all the state entries 
        //in the parse table having creator as their creator.
		return entriesFor(
			new BinaryPredicate() {
				public boolean execute(Object state,Object creator) {
					return ((ParseState)state).creator().equals((String)creator);
				}
			},
			creator
		);
	}

	public String toString() {
		StringBuffer buf = new StringBuffer("\n-- Parse Table --");
		for (HashMapIterator slot=_table.begin();!slot.atEnd();slot.advance()) {
		
			buf.append("\n<"+slot.key().toString()+">");
			for (SListIterator state=((SList)((Pair)slot.value()).second).begin();
    			 !state.atEnd();state.advance()) {
    			buf.append("\n"+((ParseState)state.get()).surrogate().toString()+
    			           " "+state.get().toString());
     			buf.append("\n   downs: ");
    			for (SListIterator down=downStatesFor((ParseState)state.get()).begin();
    			     !down.atEnd();down.advance())
    				buf.append(((ParseState)down.get()).surrogate().toString()+" ");
    			buf.append(" next: ");
    			for (SListIterator next=nextStatesFor((ParseState)state.get()).begin();
    			     !next.atEnd();next.advance())
    				buf.append(((ParseState)next.get()).surrogate().toString()+" ");
    			}
   		}
		buf.append("\n-- end table --");
		return buf.toString();
	} 
	
}