package relap.Analyzer;

import relap.LanguageModel.Verifiable;
import relap.LanguageModel.ImplicitPredicate;
import relap.LanguageModel.Constraint;
import com.objectspace.jgl.HashSet;
import relap.Editor.ErrorInterpreter;

public class PredicateEvaluationError extends ErrorDescriptor {

	Verifiable fPredicate;
	FailureData fData;
	
	public PredicateEvaluationError(ParseState state,Verifiable predicate,FailureData data) {
		super(state,"Failed predicate");
		fPredicate = predicate;
		fData = data;
	}
	
	public boolean isRecoverable() {
		return false;
	}
	
	public HashSet cover() {
		return super.cover().union(fData.objects());
	}
	
	public Object index() {
		return null;
	}
	
	public HashSet indices() {
		return fData.objects();
	}
	
	public FailureData getFailureData() {
		return fData;
	}
	
	public Verifiable getPredicate() {
		return fPredicate;
	}
	
	public Verifiable getFailedCondition() {
		return getPredicate();
	}
	
	public String toString() {
		return "Predicate "+fPredicate+" failed with arguments "+fData;
	}
	
	public String describeWith(String msg) {
		return 
			msg+"\nin a "
			+location().productionVariant().production().description();
	}
	
	public String describeWith(ErrorInterpreter interpreter) {
		if (fPredicate instanceof ImplicitPredicate)
			return 
				interpreter.describePredicate(fPredicate.getName())
				+"\nin a "
				+location().productionVariant().production().description();
		else if (fPredicate instanceof Constraint)
			return 
				interpreter.describeMissingRelationTuple(
					fPredicate.getName(),fData
				);
		else
			return describe();
	}
	
	public String describe() {
		return 
			toString()+
			"\nin a "
			+location().productionVariant().production().description();
	}
	
	public void setLocation(ParseState state) {
		errorLoc = state;
	}
} 
