package relap.Analyzer;

import relap.LanguageModel.*;
import relap.Util.*;
import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.*;
import com.objectspace.jgl.predicates.*;

public class PredictionTable {
	private HashMap _table;
	private RelationalGrammarImplementation _grammar;
	private Array _initStates;
	private HashMap _firstTerminals;
	
	public PredictionTable() {
		_table = null;
		_grammar = null;
		_initStates = null;
		_firstTerminals = null;
	}
	
	class VBindsFirst implements BinaryPredicate {
		public boolean execute(Object attribute,Object variant) {
			return ((OrderingVariant)variant).bindsFirst((String)attribute);
		}
	}
	
	private void traverseAndBuildTerminalSets(ProductionDependency node,
											  Deque path)
	throws InvalidGrammarException {
		if (!path.contains(node)) {
			HashSet terminals = new HashSet();
			path.pushBack(node);
			for (ForwardIterator depNode=node.neighbours().start();
			     !depNode.atEnd();depNode.advance()) 
				if (!depNode.get().equals(node)) {
					traverseAndBuildTerminalSets(
						(ProductionDependency)depNode.get(),path
					);
					Copying.copy(
						(Container)
							((HashMap)
								_firstTerminals.get(((ProductionDependency)depNode.get()).attribute())
							).get(((ProductionDependency)depNode.get()).nonTerminal()),
						terminals
					);
				}
			for (ForwardIterator variant=node.variants().start();!variant.atEnd();
			     variant.advance()) 
				if (_grammar.isTerminal(((OrderingVariant)variant.get()).symbolAt(1)))
					terminals.add(((OrderingVariant)variant.get()).symbolAt(1));	
			Copying.copy(
				terminals,
				(Container)((HashMap)_firstTerminals.get(node.attribute())).get(node.nonTerminal())
			);
			path.popBack();		     
		}
		else 
			throw(
				new InvalidGrammarException(
					"in "+_grammar.getClass().getName()+":\n"+"cyclic dependency "+
					Utils.contentsToString(path,"\n->")+"\n->"+node.toString()
			    )
			);
	}
	
	public HashSet firstTerminalsFor(String attribute,String nonTerminal) {
		return
			(HashSet)
				((HashMap)_firstTerminals.get(attribute)).get(nonTerminal);
	
	}

	public void calculateFirstTerminals() 
	throws InvalidGrammarException {
		_firstTerminals = new HashMap();
    	for (HashSetIterator attr=_grammar.expanderAttributes().begin();
        	 !attr.atEnd();attr.advance()) {
        	 HashMap ntLists = new HashMap();
        	 for (HashSetIterator nt=_grammar.nonTerminals().begin();
        	      !nt.atEnd();nt.advance()) 
        	 	ntLists.put(nt.get(),new HashSet());	     
        	 _firstTerminals.put(attr.get(),ntLists);
        }
    	for (HashSetIterator attr=_grammar.expanderAttributes().begin();
        	 !attr.atEnd();attr.advance()) {
       		for (HashSetIterator nt=_grammar.nonTerminals().begin();
        	      !nt.atEnd();nt.advance()) {
        		DependencyGraph dependencies = 
        			new DependencyGraph(_grammar,(String)attr.get(),(String)nt.get());
        		dependencies.buildGraph(this);
        		//System.err.println(dependencies);
        		traverseAndBuildTerminalSets(dependencies.root(),new Deque());
        	} 
		}
/*     	for (HashMapIterator attr=_firstTerminals.begin();
        	 !attr.atEnd();attr.advance())
        	 for (HashMapIterator nt=((HashMap)attr.value()).begin();
        	      !nt.atEnd();nt.advance()) 
        	 	System.err.println(
        	 		"\n"+attr.key().toString()+"."+nt.key().toString()+" "+
        	 		nt.value().toString()
        	 	);
 */        
	}
 
	public void buildWith(RelationalGrammarImplementation grammar)
	throws InvalidGrammarException {
        //Populate the receivers internal dictionary with the starts-by-binding tuples."
    	HashMap attributes;
    	Array variants;
    	_grammar = grammar;
    	_table = new HashMap();
    	for (HashSetIterator nt=_grammar.nonTerminals().begin();
    		 !nt.atEnd();nt.advance()) {
        	attributes = new HashMap();
        	variants = new Array();
        	for(ArrayIterator p=((Array)_grammar.productionsExpanding((String)nt.get())).begin();
        	    !p.atEnd();p.advance())
            	Copying.copy(((GrammarProduction)p.get()).orderingVariants(),variants);
        	
        	for (HashSetIterator attr=_grammar.expanderAttributes().begin();
        	     !attr.atEnd();attr.advance()) 
            	attributes.add(attr.get(),
            				   Filtering.select(
            				   		variants,
            				   		new BindFirstPredicate(new VBindsFirst(),attr.get())
            				   )
            	);
        
        	attributes.add(grammar.startAttribute(),variants);
        	_table.add(nt.get(),attributes);
        }
        calculateFirstTerminals();
    }
 

	class IsInitState implements BinaryPredicate {
		public boolean execute(Object state,Object new_states) {
			return (_initStates.contains((ActiveState)state) ||
			        ((Array)new_states).contains((ActiveState)state));
		}
	}

	public Array initStates() {
		Array new_states;
		_initStates = new Array();
		new_states = statesFor(_grammar.startSymbol(),_grammar.startAttribute());
		while (!new_states.isEmpty()) {
			Copying.copy(new_states,_initStates);
			new_states = new Array();
			for (ArrayIterator i_state=_initStates.begin();!i_state.atEnd();i_state.advance()) {
				((ActiveState)i_state.get()).creator("init");
				((ActiveState)i_state.get()).surrogate(new Integer(0));
				if(_grammar.isNonTerminal(((ActiveState)i_state.get()).symbolAtDot())) {
					Array temp_states = new Array();
					Copying.copy(statesFor(((ActiveState)i_state.get()).symbolAtDot(),
											_grammar.startAttribute()),
								 temp_states);
					Copying.copy(
						Filtering.reject(temp_states,
					                	 new BindSecondPredicate(new IsInitState(),new_states)),
					    new_states
					);
					
					for (ArrayIterator it=new_states.begin();!it.atEnd();it.advance())
						((ActiveState)it.get()).up((ActiveState)i_state.get()); 
				}
			}
		}
		return _initStates;
	}

	public Array statesFor(String non_terminal,String attribute_to_bind) {
		Array states = new Array();
		for (ArrayIterator v=((Array)((HashMap)_table.get(non_terminal)).get(attribute_to_bind)).begin(); 
			 !v.atEnd();v.advance())
			 states.add(new ActiveState((OrderingVariant)v.get()));
		return states;
	}

	public Array variantsFor(String non_terminal,String attribute_to_bind) {
		return ((Array)((HashMap)_table.get(non_terminal)).get(attribute_to_bind)); 
	}

	public String toString() {
		StringBuffer buf = new StringBuffer("\n-- Prediction Table --");
		for (HashMapIterator entry=_table.begin();!entry.atEnd();entry.advance()) 
			for (HashMapIterator slot=((HashMap)entry.value()).begin();
			     !slot.atEnd();slot.advance())
				for(ArrayIterator v=((Array)slot.value()).begin();
				    !v.atEnd();v.advance())
					buf.append("\n"+(String)entry.key()+" "+(String)slot.key()+" "+
				               v.get().toString());
		buf.append("\n-- end table --");
		return buf.toString();
	}
}
