package relap.Editor;

import relap.LanguageModel.ExpanderRelation;
import relap.LanguageModel.Tuple;
import relap.LanguageModel.InvalidTupleException;
import relap.LanguageModel.ExistenceConstraint;
import relap.Analyzer.IndexedMDSet;
import relap.Analyzer.FailureData;
import relap.Analyzer.ExistenceFailureData;
import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.Array;
import com.objectspace.jgl.Container;
import com.objectspace.jgl.BinaryFunction;
import com.objectspace.jgl.algorithms.*;
import com.objectspace.jgl.functions.*;
import relap.Util.Utils;

public class ImplicitExpanderRelation extends ExpanderRelation {

	protected String fLeftToRightDescription;
	protected String fRightToLeftDescription;

	public ImplicitExpanderRelation(String name,
									String leftToRight,String rightToLeft) {
		super(name);
		fLeftToRightDescription = leftToRight;
		fRightToLeftDescription = rightToLeft;
	}

	public Tuple addTuple(Array tupleData)
		throws InvalidTupleException {
		Tuple newTuple = super.addTuple(tupleData);
		newTuple.markTuple();
		return newTuple;
	}
	
	public String describeInvalidTuple(Array tuple) {
		StringBuffer buf = new StringBuffer("Not expecting ");
		buf.append(tuple.at(0)+" ");
		buf.append(fLeftToRightDescription);
		buf.append(" "+tuple.at(1)+" here");
		return buf.toString();
	}

	public String describeMissingTuple(HashSet expectedObjects,Object object,
	                                   int pos,String kind) {
		if (pos == 0) 
			return
				"expecting a "+
				Utils.listContents(expectedObjects,", "," or ")+" "+
				fLeftToRightDescription+" this "+kind;
			
		else 
			return
				"expecting a "+
				Utils.listContents(expectedObjects,", "," or ")+" "+
				fRightToLeftDescription+" this "+kind;
	}
	
	public String describeExistenceFailure(ExistenceFailureData data,IndexedMDSet input) {
		Object anchor = data.getBoundArgument();
		//System.err.println("## "+data.objects());
		//System.err.println("## "+anchor);
		ExistenceConstraint c = data.getConstraint();
		Container expected = 
			Transforming.collect(
				data.terminals(),
				new BindFirst(
					new BinaryFunction () {
						public Object execute(Object input,Object terminal) {
							return
								((IndexedMDSet)input).describeTerminal((String)terminal);
						}
					},
					input
				)
			);
		if ( c.nullRefPos() == 1) 
			return
				"expecting "+data.count()+" "+
				Utils.listContents(expected,", "," or ")+" "+
				fLeftToRightDescription+" this "+input.describeTerminal(input.getTerminalType(anchor));
			
		else 
			return
				"expecting "+data.count()+" "+
				Utils.listContents(expected,", "," or ")+" "+
				fRightToLeftDescription+" this "+input.describeTerminal(input.getTerminalType(anchor));
	}
		
		
		
	public String describeFailure(FailureData data,IndexedMDSet input) {
		if (data instanceof ExistenceFailureData)
			return describeExistenceFailure((ExistenceFailureData)data,input);		
		return 
			"expecting the "+
			input.describeTerminal(input.getTerminalType(data.tuple().at(0)))+" to be "+
			fLeftToRightDescription+" the "+
			input.describeTerminal(input.getTerminalType(data.tuple().at(1)));
	}

	public String describeExtras(Container expectedObjects,Object object,
	                             int expectedCount, int pos,String kind) {
		if (pos == 0) 
			return
				"expecting only "+expectedCount+" "+
				Utils.listContents(expectedObjects,", "," or ")+" "+
				fLeftToRightDescription+" this "+kind;
			
		else 
			return
				"expecting only "+expectedCount+" "+
				Utils.listContents(expectedObjects,", "," or ")+" "+
				fRightToLeftDescription+" this "+kind;	                                  
	}

}