package relap.Editor;


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import relap.Analyzer.*;
import relap.LanguageModel.*;
import com.objectspace.jgl.*;
import com.objectspace.jgl.adapters.*;
import com.objectspace.jgl.Container;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.predicates.LessNumber;
import java.util.Enumeration;
import java.util.Vector;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.figures.PolyLineFigure;
import CH.ifa.draw.framework.DrawingEditor;

public class ParseFailurePanel 
	extends Panel implements ActionListener {    
    private TextArea textArea;
    Analyzer analyzer;
    Button dismissButton;
    Button prevButton;
    Button nextButton;
    Button browseButton;
    Array errors;
    String newline;
    ProcessingError currentError;
    ParseGraphBrowser browser;
		OrderedMap lineToError;
		
    public ParseFailurePanel(Frame dw,Analyzer analyzer,ProcessingResult failure) {
    	this(dw,analyzer,failure,"Error:");
    }

    public ParseFailurePanel(Frame dw,Analyzer analyzer,ProcessingResult failure,String title) {
      this.analyzer = analyzer;
			lineToError = new OrderedMap(new LessNumber());
			SList allErrors = new SList();
			Copying.copy(new VectorArray(failure.errors()),allErrors);
			Copying.copy(new VectorArray(failure.warnings()),allErrors);
			errors = relap.Util.Utils.asArray(filterErrors(allErrors));
			setLayout(new BorderLayout());
      Label label = new Label(title);
      add("North",label);
      textArea = new TextArea(9, 60);
      textArea.setEditable(false);
      newline = System.getProperty("line.separator");
      browser = new ParseGraphBrowser(dw,"Browse",analyzer.getParser().getParseTable());
      browser.setVisible(false);
      add("Center", textArea);
      textArea.addMouseListener(
        new MouseAdapter() {
        	public void mouseClicked(MouseEvent e) {
        		if (e.getClickCount() > 1) {
       				//System.err.println(
       				//	"start at: "+textArea.getSelectionStart()+" end at: "+
       				//	textArea.getSelectionEnd()
       				//);
       				updateError();
       			}
       			else {
       				//System.err.println(">> CLICK <<");
       				textArea.select(
       					textArea.getCaretPosition(),textArea.getCaretPosition()+1
       				);
       				updateError();
       			}
       		}
       	}
      );
      setTextContents(errors);
      textArea.select(0,1);
      updateError();
      
      Panel p2 = new Panel();
      browseButton = new Button("Show parse");
      browseButton.addActionListener(this);
      
			if (currentError == null || currentError.location() == null)
				browseButton.setEnabled(false);
      p2.add(browseButton);
      
      add("South", p2);
    }

	private void setTextContents(Container errorList) {
		Enumeration error = errorList.elements();
		int pos = 0;
		while (error.hasMoreElements()) {
			pos = textArea.getText().length();
			ProcessingError current = (ProcessingError)error.nextElement();
			lineToError.add(new Integer(pos),current);
			textArea.insert(
				current.describeWith(analyzer)+newline,
				pos
			);
			if (error.hasMoreElements()) {
				for (int i = 0;i<80;i++)
					textArea.append("-");
				textArea.append(newline);
			}
		}	
	}

	private Container filterErrors(Container descriptors) {
		//System.err.println("errors: "+descriptors);
		Array a = new Array();
		Copying.copy(descriptors,a);
		Enumeration errors =
			Filtering.select(
				a,
				new UnaryPredicate() {
					public boolean execute(Object error) {
						if (error instanceof ExtraRelation)
							if (((ExtraRelation)error).extraTuple().relation()
									instanceof ImplicitExpanderRelation)
								return false;
						return true;
					}
				}
			).elements();
		SList result = new SList();
		while (errors.hasMoreElements()) 			
				Copying.copy(((ProcessingError)errors.nextElement()).unPack(),result);
		
		return result;
	}	

	
    public void actionPerformed(ActionEvent event) {
    	Object source = event.getSource();
      if (source == browseButton) {
        browser.clearContents();
        if (currentError.location() != null)
         	browser.setContents(currentError.location());
        Point p = getParent().getLocation();
        p.translate(0,getParent().getSize().height);
        browser.setLocation(p);
        browser.setVisible(true);
        
      }
    }
    
    
    private HashSet getFigures(ProcessingError error) {
    	if (error instanceof relap.Analyzer.ExtraRelation) {
			HashSet figures = new HashSet();    		
			relap.LanguageModel.Tuple extra = 
    			((ExtraRelation)error).extraTuple();
    		if (extra.relation() instanceof ExplicitExpanderRelation)
    			figures.add(
    				((ExplicitExpanderRelation)extra.relation()).getFigure(extra)
    			);
    		// else there is no explicit figure representing this relation
    		return figures;
    	}
    	else if (error != null)
    		return error.indices();
    	else 
    		return new HashSet();
    }
    
    public void cleanUp() {
    	analyzer.clearSelection();
    }
    
   private ProcessingError getSelectedError() {
   	ProcessingError selected = null;
   	if (!textArea.getSelectedText().equals(""))	 {
   		int pos = textArea.getSelectionStart();
			for (OrderedMapIterator e = lineToError.begin();!e.atEnd();e.advance()) 
				if (((Integer)e.key()).intValue() <= pos)
					selected = (ProcessingError)e.value();
   	}
   	return selected;
   }
   
  private void selectCurrentError() {
   	if (currentError != null) {
   		int start = 
   			((Integer)lineToError.keys(currentError).nextElement()).intValue();
   		int stop = start;
   		OrderedMapIterator it = lineToError.find(new Integer(start));
   		if (!it.atEnd())
   			it.advance();
   		if (it.atEnd())
   			stop = textArea.getText().length();
   		else
   			stop = ((Integer)it.key()).intValue()-82;
   		textArea.select(start,stop);
   	} 			
  }
   
 	private void updateError() {
   	cleanUp();
		currentError = getSelectedError();
		//System.err.println("currentError: "+currentError);
		selectCurrentError();
		analyzer.setSelection(getFigures(currentError));
		if (currentError != null)
    	if (currentError.location() != null) { 
        if (browseButton != null)
        	browseButton.setEnabled(true);
        if (browser.isVisible()) {
        	browser.clearContents();
        	browser.setContents(currentError.location());
       		browser.setVisible(true);
       	}
      }
      else 
        if (browseButton != null)
        	browseButton.setEnabled(false);
  }

}