package relap.Editor;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import relap.Analyzer.*;
import relap.LanguageModel.*;
import com.objectspace.jgl.DList;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.ForwardIterator;
import com.objectspace.jgl.HashMap;
import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.UnaryPredicate;
import com.objectspace.jgl.Container;
import com.objectspace.jgl.DListIterator;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Filtering;
import java.util.Enumeration;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.figures.PolyLineFigure;
import CH.ifa.draw.framework.DrawingEditor;

public class ParseGraphBrowser extends Dialog implements ActionListener { 
    protected SList fUpButtons;
    protected SList fDownButtons;
    protected SList fPrevButtons;
    protected SList fNextButtons;
    FramePanel fUp;
    FramePanel fDown;
    FramePanel fPrev;
    FramePanel fNext;
    Button dismissButton;
    ParseState fCurrent; 
    ParseTable fParseTable;

    protected void addpanel(Panel panel,
     						GridBagLayout gridbag,
                            GridBagConstraints c) {
       gridbag.setConstraints(panel, c);
       add(panel);
   }
   
   	class DirectionFlags {
   		public boolean up;	
   		public boolean down;	
   		public boolean prev;	
   		public boolean next;	
   		
   		public DirectionFlags(boolean up,boolean down, boolean prev, boolean next) {
   			this.up = up;
   			this.down = down;
   			this.prev = prev;
   			this.next = next;
   		}

   		public DirectionFlags() {
   			this.up = false;
   			this.down = false;
   			this.prev = false;
   			this.next = false;
   		}
   	}
    
   class Centroid extends FramePanel {
   
		DirectionFlags dirFlags;
   		int halfBase = 16;
   		int height = 13;
   		public Centroid(LayoutManager m) {
   			super(m);
   			dirFlags = new DirectionFlags();
   		}
   		
   		public Centroid(LayoutManager m, DirectionFlags flags) {
   			super(m);
   			dirFlags = flags;
   		}
   		
   		public Centroid() {
   			super();
   			dirFlags = new DirectionFlags();
   		}
   		
   		public Centroid(DirectionFlags flags) {
   			super();
   			dirFlags = flags;
   		}
   		
   		public Insets getInsets() {
   			return new Insets(16,16,16,16);
   		}
   		
   		public void drawRightArrow(Graphics g, Point baseMid) {
   			int x[] = new int[3];
   			int y[] = new int[3];
   			x[0] = baseMid.x;          y[0] = baseMid.y - halfBase;
   			x[1] = baseMid.x + height; y[1] = baseMid.y;
   			x[2] = baseMid.x;          y[2] = baseMid.y + halfBase;
   			
   			Color col = g.getColor();
   			g.setColor(Color.black);
    		g.fillPolygon(x, y, 3);	
   			g.setColor(col);
   		
   		}

   		public void drawDownArrow(Graphics g, Point baseMid) {
   			int x[] = new int[3];
   			int y[] = new int[3];
   			x[0] = baseMid.x - halfBase; y[0] = baseMid.y;
   			x[1] = baseMid.x + halfBase; y[1] = baseMid.y;
   			x[2] = baseMid.x;            y[2] = baseMid.y + height;
   			
   			Color col = g.getColor();
   			g.setColor(Color.white);
    		g.fillPolygon(x, y, 3);	
    		g.setColor(Color.darkGray);
    		g.drawPolygon(x, y, 3);	
   			g.setColor(col);
   		
   		}
   		
   		public void paint(Graphics g) {
   			super.paint(g);
   			Dimension size = getSize();
   			int centerX = size.width / 2;
   			int centerY = size.height / 2;
   			Point north = new Point(centerX, 0);
   			north.translate(0,2);
   			Point south = new Point(centerX, size.height);
   			south.translate(0,-height-2);
   			Point west = new Point(0, centerY);
   			west.translate(2,0);
   			Point east = new Point(size.width, centerY);
   			east.translate(-height-2,0);
   			if (dirFlags.prev)
  				drawRightArrow(g,west);
   			if (dirFlags.next)
  				drawRightArrow(g,east);
   			if (dirFlags.up)
  				drawDownArrow(g,north);
   			if (dirFlags.down)
  				drawDownArrow(g,south);
   		}
   	}
   	
   class FramePanel extends Panel {
   
   		Dimension fMinSize;
   		public FramePanel(LayoutManager m) {
   			super(m);
   			fMinSize =null;
   		}
   		
   		public FramePanel() {
   			super();
   		}
   		
   		public Insets getInsets() {
   			return new Insets(10,10,10,10);
   		}
   		
   		public void drawFrame(Graphics g) {
   			Dimension size = getSize();
  			Color col = g.getColor();
   			g.setColor(Color.darkGray);
    		g.drawRect(0, 0, size.width-1, size.height-1);	
   			g.setColor(col);
   		
   		}
   		
   		public void paint(Graphics g) {
   			super.paint(g);
  			drawFrame(g);
   		}
   		
   		/* public Dimension getMinimumSize() {
   			if (fMinSize == null)
   				return super.getMinimumSize();
   			return fMinSize;
   		}

   		public void setMinimumSize(Dimension dim) {
   			fMinSize = new Dimension(dim);
   		}

   		public Dimension getPreferredSize() {
   			if (getComponentCount() ==0)
   				return getMinimumSize();
   			return super.getPreferredSize(); 
   		}*/
   	}
   	
    public ParseGraphBrowser(Frame dw, String title,ParseTable parseTable) {
        super(dw, title, false);
   		fUpButtons = new SList();
   		fPrevButtons = new SList();
   		fDownButtons = new SList();
   		fNextButtons = new SList();
   		fParseTable = parseTable;
    }
    
    public void clearContents() {
    	removeAll();
    	fUpButtons.clear();
    	fPrevButtons.clear();
   		fDownButtons.clear();
   		fNextButtons.clear();
    }
    
    protected ForwardIterator getUpLabels() {
    	SList labels = new SList();
    	for (ForwardIterator state = fCurrent.up().start();
    	     !state.atEnd();state.advance())
    		labels.pushBack(
    			((ParseState)state.get()).surrogate().toString()
    		);
     	return labels.start();
     }
    
    protected ForwardIterator getPrevLabels() {
    	SList labels = new SList();
    	for (ForwardIterator state = fCurrent.previous().start();
    	     !state.atEnd();state.advance())
    		labels.pushBack(
    			((ParseState)state.get()).surrogate().toString()
    		);
     	return labels.start();
     }
    
    protected ForwardIterator getNextLabels() {
    	SList labels = new SList();
    	for (ForwardIterator state = fParseTable.nextStatesFor(fCurrent).start();
    	     !state.atEnd();state.advance())
    		labels.pushBack(
    			((ParseState)state.get()).surrogate().toString()
    		);
     	return labels.start();
     }

    protected ForwardIterator getDownLabels() {
    	SList labels = new SList();
    	for (ForwardIterator state = fParseTable.downStatesFor(fCurrent).start();
    	     !state.atEnd();state.advance())
    		labels.pushBack(
    			((ParseState)state.get()).surrogate().toString()
    		);
     	return labels.start();
     }

    protected Panel getUpButtons() {
		fUp = new FramePanel();
		for (ForwardIterator label = getUpLabels();!label.atEnd();label.advance()) {
			Button button = new Button((String)label.get());
			fUpButtons.pushBack(button);
			fUp.add(button);
        	button.addActionListener(this);
   		}
   		return fUp;
    }
    
    protected Panel getPrevButtons() {
		fPrev = new FramePanel();
		for (ForwardIterator label = getPrevLabels();!label.atEnd();label.advance()) {
			Button button = new Button((String)label.get());
			fPrevButtons.pushBack(button);
			fPrev.add(button);
        	button.addActionListener(this);
   		}
   		return fPrev;
    }

    protected Panel getNextButtons() {
		fNext = new FramePanel();
		for (ForwardIterator label = getNextLabels();!label.atEnd();label.advance()) {
			Button button = new Button((String)label.get());
			fNextButtons.pushBack(button);
			fNext.add(button);
        	button.addActionListener(this);
   		}
   		return fNext;
    }

    protected Panel getDownButtons() {
		fDown = new FramePanel();
		for (ForwardIterator label = getDownLabels();!label.atEnd();label.advance()) {
			Button button = new Button((String)label.get());
			fDownButtons.pushBack(button);
			fDown.add(button);
        	button.addActionListener(this);
   		}
   		//if (fDownButtons.isEmpty())
   		//	fDown.add(new Label("no predictions"));
   		return fDown;
    }
    
    protected void uniWide(FramePanel first, FramePanel second) {
    	System.err.println("pre: "+first.getSize()+" "+second.getSize());
     	if (first.getSize().equals(second.getSize()))
    		return;
	   	if (first.getSize().width < second.getSize().width) {
    		first.setSize(
    			new Dimension(
    				second.getSize().width,first.getSize().height
    			)
    		);
    		first.invalidate();
    	}
    	else {
    		second.setSize(
    			new Dimension(
    				first.getSize().width,second.getSize().height
    			)
    		);
    		second.invalidate();
    	}
    	System.err.println("post: "+first.getSize()+" "+second.getSize());
    }

    public void setContents(ParseState newCurrent) {
        fCurrent = newCurrent;
 		//System.err.println(fCurrent.describe());
        setFont(new Font("Helvetica", Font.PLAIN, 14));
		Panel navPanel = new Panel();
        navPanel.setLayout(new BorderLayout(2,2));
        
   		Panel panel = getUpButtons();
        navPanel.add("North",panel);
        //addpanel(panel, gridbag, c);
        
		panel = getPrevButtons();
		panel.setLayout(new GridLayout(0,1));
        navPanel.add("West",panel);

		panel = getNextButtons();
		panel.setLayout(new GridLayout(0,1));
		navPanel.add("East",panel);

		panel = getDownButtons();
        navPanel.add("South",panel);
		
		panel = 
			new Centroid(
				new DirectionFlags(
					!fUpButtons.isEmpty(),
					!fDownButtons.isEmpty(),
					!fPrevButtons.isEmpty(),
					!fNextButtons.isEmpty()
				)
			);
		panel.setLayout(new GridLayout(0,1));
		fillCenter(panel);
        navPanel.add("Center",panel);
        
		add("North",navPanel);

        dismissButton = new Button("Dismiss");
        dismissButton.addActionListener(this);
        panel = new Panel();
        panel.add(dismissButton);

        add("South",panel);

		pack();
    }
    
    private void fillCenter(Panel panel) {
   		if (fCurrent != null) {
				Label label = 
					new Label(fCurrent.surrogate().toString()+": "+fCurrent.describe());
				label.setAlignment(Label.CENTER);
        panel.add(label);
        label = new Label("created by "+fCurrent.creator());
				label.setAlignment(Label.CENTER);
        panel.add(label);
        if (fCurrent instanceof ActiveState)
        	if (fCurrent.errorDescription() != null)
        		for (ForwardIterator err = fCurrent.errorDescription().unPack().start();
        		     !err.atEnd();err.advance()) {
        		     System.err.println(err.get());
        			if (((ErrorDescriptor)err.get()).getFailedCondition() != null) {
        				label = new Label("error: "+((ErrorDescriptor)err.get()).getFailedCondition().toString());
						label.setAlignment(Label.LEFT);
        				panel.add(label);
        			}
        		}
    	}
    	else {
				Label label = 
					new Label("Transient state due to error recovery or\nmissing optional symbols");
				label.setAlignment(Label.CENTER);
        panel.add(label);    		
    	}
    }

    public void actionPerformed(ActionEvent event) {
    	Object source = event.getSource();
    	if (source == dismissButton) {
    		clearContents();
        	setVisible(false);
        }
        else {
        	if (fUpButtons.contains(source)   ||
        	    fDownButtons.contains(source) ||
        	    fPrevButtons.contains(source) ||
        	    fNextButtons.contains(source)) 
        	{
        		fCurrent = 
        			fParseTable.stateForSurrogate(
        				new Integer(((Button)source).getLabel())
        			);
        		clearContents();
        		setContents(fCurrent);
        	}
        }
    }
}