package relap.LanguageModel;

public class AttributeAssignment {

	private ObjectReference _attributeExpr;
	private ObjectReference _assignedExpr;

	public AttributeAssignment(ObjectReference attributeExpr,
					  		   ObjectReference assignedExpr) {
		_attributeExpr = attributeExpr;
		_assignedExpr = assignedExpr;
	}
	
	public AttributeAssignment copy() {
		return new AttributeAssignment(_attributeExpr.copy(),_assignedExpr.copy());
	}

	public AttributeAssignment copy(String attribute) {
		InDirectReference expr= (InDirectReference)_attributeExpr.copy();
		expr.attribute(attribute);
		return new AttributeAssignment(expr,_assignedExpr.copy());
	}

	public String attribute() {
		return ((InDirectReference)_attributeExpr).attribute();
	}

	public void check(RelationalGrammarImplementation grammar) 
	throws InvalidAttributeAssignmentException {
		try {
			_attributeExpr.checkLhsReference(grammar);
			_assignedExpr.checkRhsReference(grammar);
		}
		catch(InvalidObjectReferenceException e) {
			throw(new InvalidAttributeAssignmentException(this,e.getMessage()));
		}
		if (value().production().isRepetitive(value().position()))
			throw(new InvalidAttributeAssignmentException(
						this,
						"Cannot assign a set of values from a symbol sequence")
			);
		if (value().production().isOptional(value().position()))
			throw(new InvalidAttributeAssignmentException(
						this,
						"Cannot assign a value from an optional right-hand-side symbol")
			);
	}

	public ObjectReference value() {
		return _assignedExpr;
	}
	
	public ObjectReference attributeExpr() {
		return _attributeExpr;
	}
	
	public void value(ObjectReference objectRef) {
		_assignedExpr = objectRef;
	}
	
	public String toString() {
		return _attributeExpr.toString()+" = "+
		       _assignedExpr.toString();
	}
}