package relap.LanguageModel;

import com.objectspace.jgl.Pair;
import com.objectspace.jgl.Array;
import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.SList;
import relap.Analyzer.FailureData;
import relap.Analyzer.ConstraintFailureData;
import java.util.Enumeration;

public class Constraint
	implements Verifiable {

	private String _relation;
	private Pair _objects;
	private SList fLatestFailures;
	
	public Constraint(String relation) {
		_relation = relation;
		fLatestFailures = new SList();
	}
	
	public Constraint(String relation,Pair objects) {
		_relation = relation;
		_objects = objects;
		fLatestFailures = new SList();
	}	

	

	public ObjectReference argumentAt(int position) {
    	if (firstArg().position() == position)
        	return firstArg();
        else
        	return secondArg();
    }
   
	public ObjectReference argumentBefore(int position) {
    	if (firstArg().position() < position)
        	return firstArg();
        else
        	return secondArg();
    }

	public void check(RelationalGrammarImplementation grammar) 
	throws InvalidConstraintException {
		if (grammar.relations().get(_relation) == null)
			throw(new InvalidConstraintException(this,
												 "Undefined relation \""+
												 _relation+"\""));
		try {
			firstArg().checkRhsReference(grammar);
			secondArg().checkRhsReference(grammar);			
		}
		catch(InvalidObjectReferenceException e) {
			throw(new InvalidConstraintException(this,e.getMessage()));
		}
		if (firstArg().production().isRepetitive(firstArg().position()) &&
			secondArg().production().isRepetitive(secondArg().position()))
			throw(new InvalidConstraintException(
							this,
							"Both arguments cannot be repetitive")
			);			
	}
	
	public ObjectReference firstArg() {
    	return (ObjectReference)_objects.first;
    }

   
	public Pair objects() { return _objects; }
 

	public String relation() { return _relation; }
  
	public String getName() { return _relation; }
  
	public ObjectReference secondArg() {
    	return (ObjectReference)_objects.second;
    }
   
	public String toString() {
		if (_objects == null || _relation == null)
			return super.toString();
    	return _relation+"("+_objects.first.toString()+","+
      		   _objects.second.toString()+")";
    }
    
	public boolean verify(relap.Analyzer.ParseState state,
				          relap.Analyzer.IndexedMDSet input)
		throws InvalidQueryArgumentsException {
		Array arguments = new Array(input.getRelation(relation()).arity());
		fLatestFailures.clear();
//		System.err.println(constraint);
		arguments.put(0,state.dereference((ObjectReference)objects().first));
		arguments.put(1,state.dereference((ObjectReference)objects().second));
		HashSet tuples = 
			input.getRelation(relation()).getTuples(arguments);
		//System.err.println(tuples);
		if (!tuples.isEmpty())
			state.addRelationTuple((Tuple)tuples.begin().get());
		if (tuples.isEmpty()) {
			//System.err.println("getFailedArgs(): "+input.getRelation(relation()).getFailedArgs());
			Enumeration args = input.getRelation(relation()).getFailedArgs().elements();
			while (args.hasMoreElements())
				fLatestFailures.pushBack(
					new ConstraintFailureData((Array)args.nextElement())
				);
			//System.err.println("arguments: "+arguments);
			return false;
		}
		else
			return true;				          
	}
	
	public SList getFailureData() {
		return new SList(fLatestFailures);
	}

	public SList rhsReferences() {
		SList refs = new SList();
		refs.pushBack(firstArg());
		refs.pushBack(secondArg());
		return refs;
	}
}
