package relap.LanguageModel;

import relap.Util.Utils;
import relap.Analyzer.ParseState;
import relap.Analyzer.IndexedMDSet;
import relap.Analyzer.FailureData;
import relap.Analyzer.ExistenceFailureData;
import com.objectspace.jgl.Pair;
import com.objectspace.jgl.Array;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.HashSetIterator;
import com.objectspace.jgl.SListIterator;
import com.objectspace.jgl.algorithms.Copying;
import java.util.Enumeration;

public class ExistenceConstraint extends Constraint {

	protected HashSet constrainedTerminals;
	protected SList terminalsAsList;
	protected SList fLatestFailures;
	protected int fExpectedNumber;
	protected Object fBoundArgument;
	
	public ExistenceConstraint(String relation,SList terminals) {
		super(relation);
		terminalsAsList = terminals;
		constrainedTerminals = new HashSet();
		Copying.copy(terminals,constrainedTerminals);
		fLatestFailures = new SList();
		fExpectedNumber = 1;
		fBoundArgument = null;
	}
	
	public ExistenceConstraint(String relation,Pair objects,SList terminals) {
		this(relation,objects,terminals,1);
	}

	public ExistenceConstraint(String relation,Pair objects,SList terminals,
	                           int exactly) {
		super(relation,objects);
		terminalsAsList = terminals;
		constrainedTerminals = new HashSet();
		Copying.copy(terminals,constrainedTerminals);
		fExpectedNumber = exactly;
		fLatestFailures = new SList();
		fBoundArgument = null;
	}	
	
	public void check(RelationalGrammarImplementation grammar) 
	throws InvalidConstraintException {
		check(grammar,true);
	}

	public void check(RelationalGrammarImplementation grammar,boolean optionalArgumentsOK) 
	throws InvalidConstraintException {
		if (grammar.relations().get(relation()) == null)
			throw(new InvalidConstraintException(this,
												 "Undefined relation \""+
												 relation()+"\""));
		if (constrainedTerminals.isEmpty())
			constrainedTerminals = new HashSet(grammar.terminals());
			
		for (SListIterator term=terminalsAsList.begin();!term.atEnd();
		     term.advance())
	    	if (grammar.terminals().get(term.get())==null)
				throw(new InvalidConstraintException(this,
													 "Undefined terminal \""+
													 term.get().toString()+"\""));
		if (nullReferenceCount() == 0)
			throw(new InvalidConstraintException(this,"A null-reference \"_\" expected"));									

		if (nullReferenceCount() > 1)
			throw(new InvalidConstraintException(this,"Only one null-reference \"_\" expected"));									

		if (nullRefPos() == 1) {
			try {
				secondArg().checkRhsReference(grammar);			
			}
			catch(InvalidObjectReferenceException e) {
				throw(new InvalidConstraintException(this,e.getMessage()));
			}
			if (!optionalArgumentsOK && secondArg().production().isOptional(secondArg().position()))
				throw(new InvalidConstraintException(this,"Optional right-hand-side symbol not allowed in the argument expression"));
		}
		else {
			try {
				firstArg().checkRhsReference(grammar);			
			}
			catch(InvalidObjectReferenceException e) {
				throw(new InvalidConstraintException(this,e.getMessage()));
			}
			if (!optionalArgumentsOK && firstArg().production().isOptional(firstArg().position()))
				throw(new InvalidConstraintException(this,"Optional right-hand-side symbol not allowed in the argument expression"));
		}
	}

	public boolean verify(ParseState state,IndexedMDSet input) 
	throws InvalidQueryArgumentsException {
		//System.err.println("-----------");
		//System.err.println(state);
		boolean result = true;
		fLatestFailures.clear();
		Array arguments = new Array(input.getRelation(relation()).arity());
//		System.err.println(neConstraint);
		arguments.put(0,state.dereference((ObjectReference)objects().first));
		arguments.put(1,state.dereference((ObjectReference)objects().second));
		fBoundArgument = null;
		Enumeration arg = ((SList)arguments.at(nullRefPos()%2)).elements();
		while (arg.hasMoreElements()) {
			fBoundArgument = arg.nextElement();
			arguments.put(nullRefPos()%2,new SList(1,fBoundArgument));
			//Copying.copy((SList)arguments.at(nullRefPos()%2),fBoundArgument);
		//HashSet tmpResult = 
		//	input.getRelation(neConstraint.relation()).query(arguments);
		//	System.err.println("BBBB "+fBoundArgument);
		//	HashSet tmpSet = 
				//.difference(state.cover());	
		//System.err.println("arguments: "+arguments);
		//System.err.println("query result: "+tmpResult);	
			HashSet querySet = 
				filterQueryResult(input.getRelation(relation()).query(arguments),input);	
			//for (HashSetIterator obj=tmpSet.begin();!obj.atEnd();obj.advance())
			//	if (constrainedTerminals().get(input.getTerminalType(obj.get()))!=null)
			//		finalSet.add(obj.get());
			//System.err.println(this);
			//System.err.println("fExpectedNumber == "+fExpectedNumber);
		//System.err.println("finalResult.size() == "+finalResult.size());
			boolean subResult = evaluate(querySet);
			if (!subResult) {
				fLatestFailures.pushBack(createFailureData(querySet));	
				result = false;
			}
		}
		return result;
	}
	
	protected HashSet filterQueryResult(HashSet tmpSet,IndexedMDSet input) {
		HashSet resultSet = new HashSet();
		for (HashSetIterator obj=tmpSet.begin();!obj.atEnd();obj.advance())
			if (constrainedTerminals().get(input.getTerminalType(obj.get()))!=null)
				resultSet.add(obj.get());
		return resultSet;
	}
	
	protected Object createFailureData(HashSet querySet) {
		return
			new ExistenceFailureData(
					new Pair(fBoundArgument,querySet),
						getQuantity(),
						constrainedTerminals,
						this
					);
	}
		 
	protected boolean evaluate(HashSet result) {
		//System.err.println("ExistenceConstraint::evaluate "+result);
		if (result.size() != fExpectedNumber) 
			return false;
		else
			return true;
	}

	public SList getFailureData() {
		return fLatestFailures;
	}

	public HashSet constrainedTerminals() {
		return constrainedTerminals;
	}
	
	public SList copyConstrainedTerminals() {
		return Utils.asSList(constrainedTerminals);
	}

	public int getExpectedNumber() {
		return fExpectedNumber;
	}
		
	private int nullReferenceCount() {
		int count = 0;
		if (firstArg().position() == -1) count++;
		if (secondArg().position() == -1) count++;
		return count;
	}
	
	public int nullRefPos() {
		if (firstArg().position() == -1) return 1;
		else return 2;
	}
	
	public ObjectReference nonNullRef() {
		if (firstArg().position() == -1)
			return secondArg();
		else
			return firstArg();
	}	
	
	protected String getQuantity() {
		switch (fExpectedNumber) {
			case 1 : return "one";
			case 2 : return "two";
			case 3 : return "three";
			case 4 : return "four";
			case 5 : return "five";
			default : 
				return (new Integer(fExpectedNumber)).toString();
		}
	}
	
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(getQuantity()+" ");
		if (!terminalsAsList.isEmpty()) 
			buf.append(" {"+Utils.listContents(terminalsAsList,",",",")+"}");
		buf.append(super.toString());
		return buf.toString();
	}
}