package relap.LanguageModel;

import java.util.Enumeration;
import com.objectspace.jgl.*;
import relap.Util.Utils;
import relap.Analyzer.FailureData;
import relap.Analyzer.IndexedMDSet;

public class ExpanderRelation extends Relation {

	public ExpanderRelation(String name) {
		super(name,2);
	}
	
	protected HashSet closureTerminators(int nullPos) {
		HashSet result = new HashSet();
		for (ForwardIterator ob = objects().start();!ob.atEnd();ob.advance()) {
			Array args = new Array(2);
			SList nullArg = new SList();
			nullArg.pushBack(null);
			SList obArg = new SList();
			obArg.pushBack(ob.get());
			args.put(nullPos,nullArg);
			args.put((nullPos+1)%2,obArg);
			try {
				if (query(args).isEmpty())
					result.add(ob.get());
			} catch (InvalidQueryArgumentsException e) { 
				System.err.println(e); // shouldn't really happen 
			}
		}
		return result;
	}
	
	public Enumeration minimalObjects() {
		return closureTerminators(0).elements();
	}

	public Enumeration maximalObjects() {
		return closureTerminators(1).elements();
	}

	public HashSet minimalObjectSet() {
		return closureTerminators(0);
	}

	public HashSet maximalObjectSet() {
		return closureTerminators(1);
	}
	
	public String describeExtras(Container expectedObjects,Object object,
	                             int expectedCount, int pos,String kind) {
	    StringBuffer buf = new StringBuffer();
	    if (pos == 0) {
			buf.append("expecting only "+expectedCount+" objects x of type "+
			           Utils.listContents(expectedObjects,", "," or ")
			           +" such that ");
			buf.append("'x "+name());
			buf.append(" "+object+"'");
		} else {
			buf.append("expecting only "+expectedCount+" objects x of type "+
			           Utils.listContents(expectedObjects,", "," or ")
			           +" such that ");
			buf.append("'"+object+" ");
			buf.append(name()+" x'\n");
		}
		return buf.toString();
	}

	public String describeInvalidTuple(Tuple tuple) {
		return "invalid: "+name()+"("+tuple.data().at(0)+","+tuple.data().at(1)+")";
	}

	public String describeInvalidTuple(Array tuple) {
		return "invalid: "+name()+"("+tuple.at(0)+","+tuple.at(1)+")";
	}

	public String describeExtraTuple(Tuple tuple) {
		return "extra: "+name()+"("+tuple.data().at(0)+","+tuple.data().at(1)+")";
	}
	
	public String describeMissingTuple(HashSet expectedObjects,Object object,
	                                   int pos,String kind) {
	    StringBuffer buf = new StringBuffer();
	    if (pos == 0) {
			buf.append("expecting an object x of type "+
			           Utils.listContents(expectedObjects,", "," or ")
			           +" such that ");
			buf.append("'x "+name());
			buf.append(" "+object+"'");
		} else {
			buf.append("expecting an object x of type "+
			           Utils.listContents(expectedObjects,", "," or ")
			           +" such that ");
			buf.append("'"+object+" ");
			buf.append(name()+" x'\n");
		}
		return buf.toString();
	}
	
	public String describeFailure(FailureData data,IndexedMDSet input) {
		return "expecting "+name()+"("+data.tuple().at(0)+","+data.tuple().at(1)+")";
	}
	
	public String describeMissingTuple() {
		return "expecting a "+name();
	}
}
