package relap.LanguageModel;

import com.objectspace.jgl.Array;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Transforming;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.ArrayIterator;
import com.objectspace.jgl.Pair;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.DList;
import com.objectspace.jgl.Sequence;
import com.objectspace.jgl.SListIterator;
import com.objectspace.jgl.UnaryFunction;
import com.objectspace.jgl.BinaryPredicate;
import com.objectspace.jgl.predicates.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;
import relap.Analyzer.SyntaxTreeNode;

public class GrammarProduction 
	implements SymbolSequence {

	private String _lhs;
	private Array _rhs;
	private Array _constraints;
	private Array _attrAssigns;
	private Array _cNet;
	private Array _neConstraints;
	private Array _necNet;
	private RelationalGrammarImplementation _grammar;
	private Class _inactiveStateClass;
	private Array _predicates;
	private Array _predNet;
	private Array _orderingPredicates;
	private String _orderingExpression;
	private String _description;
	private RhsOrderingStrategy _strategy;
	private Sequence isRepetitive; 
	private Sequence isOptional;
	private String _semanticsClassName;
	private Class _semanticsClass;
	
	public GrammarProduction(RelationalGrammarImplementation g,String lhs) {
		_grammar = g;
		_lhs = lhs;
		_rhs = new Array();
		_constraints = new Array();
		_neConstraints = new Array();
		_attrAssigns = new Array();
		_cNet = null;
		_necNet = null;
		_inactiveStateClass = null;
		_predicates = new Array();
		_predNet = null;
		_description = "";
		//_strategy = new ARGRhsOrderingStrategy(g,this);
		_strategy = new RhsOrderingWithOptionals(g,this);

		isRepetitive = null;
		isOptional = null;
	}
	
	public void check() throws InvalidProductionException {
		try {
			checkRhs();
			checkConstraints();
			checkNotExistsConstraints();
			checkPredicates();
			checkAssignments();
			checkOrderingPredicates();
			checkSemantics();
		}
		catch (InvalidConstraintException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
		catch (InvalidImplicitPredicateException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
		catch (InvalidAttributeAssignmentException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
		catch (InvalidOrderingPredicateException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
	}
		
	public String lhs() { return _lhs; }
	
	public Array rhs() { return _rhs; }
	
	public void description(String d) { _description = d; }
	
	public String description() { return _description; }
	
	public Array cNet() { return _cNet; }

	public Array necNet() { return _necNet; }
	
	public Array predNet() { return _predNet; }

	public Array notExistsConstraints() { return _neConstraints; }

	public Array predicates() { return _predicates; }

	public Array attributeAssignments() { return _attrAssigns; }

	public RelationalGrammarImplementation grammar() { return _grammar; }
	
	public void addToRhs(String symbol) {
    		_rhs.add(symbol);
	}
		
	public void rightHandSide(String rhs) throws InvalidProductionException {
    	try {
			GrammarReader.parseRightHandSide(rhs,this);
    	}
    	catch (IOException e) {
			throw(new InvalidProductionException(this,
												 "internal error - read failed in GrammarProduction.constraints()"));
		}
		catch (ExpressionSyntaxException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
    	if (_rhs.size() == 0) {
    		throw(new InvalidProductionException(this,"Right-hand-side is empty"));
    	}
		_orderingPredicates = new Array(_rhs.size());
    }	
    
    public boolean isRepetitive(int pos) {
    	return ((Boolean)isRepetitive.at(pos-1)).booleanValue();
    }
    
    public boolean isOptional(int pos) {
    	return ((Boolean)isOptional.at(pos-1)).booleanValue();
    }
    
    public int sequenceLength() {
    	return _rhs.size();
    }
    
    public void setRepetitionFlags(Sequence flags) {
    	isRepetitive = flags;
    }
    
    public void setOptionalityFlags(Sequence flags) {
    	isOptional = flags;
    }
    
    public Array orderingPredicates() {
    	return _orderingPredicates;
    }
    
    public BinaryPredicate getOrderingPredicateAt(int pos) {
    	Pair pair = (Pair)_orderingPredicates.at(pos-1);
    	if (pair != null)
    		return (BinaryPredicate)pair.second;
    	else
    		return null;
    }
	
	public Integer getPositionAtOrderingFor(int pos) {
    	Pair pair = (Pair)_orderingPredicates.at(pos-1);
    	//System.err.println("Pair at "+(pos-1)+": "+_orderingPredicates.at(pos-1));
    	if (pair != null)
    		return (Integer)pair.first;
    	else
    		return null;
    }
	
	public void order(String description) throws InvalidProductionException {
		_orderingExpression = description;
    	try {
			GrammarReader.parseOrderingExpression(description,this);
    	}
    	catch (IOException e) {
			throw(new InvalidProductionException(this,
												 "internal error - read failed in GrammarProduction.constraints()"));
		}
		catch (ExpressionSyntaxException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
   	}
	
	private void checkOrderingPredicates() 
		throws InvalidOrderingPredicateException {
		for (ArrayIterator pred = _orderingPredicates.begin();!pred.atEnd();
		     pred.advance()) {
		    if (pred.get() != null) { 
			String predName = (String)((Pair)pred.get()).second;
			try {
				Method predMethod = 
					_grammar.getPredicateEvaluator().getClass().getMethod(predName,new Class[0]);
				//System.err.println("predMethod.getReturnType().getName(): "+predMethod.getReturnType().getName());
				if (!predMethod.getReturnType().getName().equals("com.objectspace.jgl.BinaryPredicate"))
					throw(
						new InvalidOrderingPredicateException(
							"no ordering predicate named "+predName+" found"
						)
					);
				else
					((Pair)pred.get()).second = 
						predMethod.invoke(_grammar.getPredicateEvaluator(),new Object[0]);
			} catch (NoSuchMethodException e) {
				throw(
					new InvalidOrderingPredicateException(
						"no ordering predicate named "+predName+" found"
					)
				);
			
			} catch (IllegalAccessException e) {
				System.err.println(e);
		
			} catch (IllegalArgumentException e) {
				System.err.println(e);
		
			} catch (InvocationTargetException e) {
				System.err.println(e);
		
			}
			}
		}	
	}
	
	public void constraints(String description) throws InvalidProductionException {
    	try {
			Copying.copy(GrammarReader.parseConstraintList(description,this),
						 _constraints);
    	}
    	catch (IOException e) {
			throw(new InvalidProductionException(this,
												 "internal error - read failed in GrammarProduction.constraints()"));
		}
		catch (ExpressionSyntaxException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
   	}
	
	public void notExistsConstraints(String description) throws InvalidProductionException {
    	try {
			Copying.copy(GrammarReader.parseNotExistsConstraintList(description,this),
						 _neConstraints);
    	}
    	catch (IOException e) {
			throw(new InvalidProductionException(this,
												 "internal error - read failed in GrammarProduction.constraints()"));
		}
		catch (ExpressionSyntaxException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
  }

	public void disambiguate(String description) throws InvalidProductionException {
		notExistsConstraints(description);
	}
	
	public void predicates(String description) throws InvalidProductionException {
    	try {
			Copying.copy(GrammarReader.parsePredicateList(description,this),
						 _predicates);
    	}
    	catch (IOException e) {
			throw(new InvalidProductionException(this,
												 "internal error - read failed in GrammarProduction.constraints()"));
		}
		catch (ExpressionSyntaxException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
   	}

	private void installConstraints() {
		Constraint c;
		if (_cNet == null && !_constraints.isEmpty()) {
			_cNet = new Array(_rhs.size());
			for (ArrayIterator slots = _cNet.begin();!slots.atEnd();slots.advance())
				slots.put(new Array());
		}
		for (Enumeration constr = _constraints.elements();
		     constr.hasMoreElements();) {
		     c = (Constraint)constr.nextElement();
			((Array)_cNet.at(c.firstArg().position()-1)).
				add(new Pair(new Integer(c.secondArg().position()),c));
			((Array)_cNet.at(c.secondArg().position()-1)).
				add(new Pair(new Integer(c.firstArg().position()),c));
		}
	}

	private void installNEConstraints() {
		//NotExistsConstraint c;
		ExistenceConstraint c;
		if (_necNet == null && !_neConstraints.isEmpty()) {
			_necNet = new Array(_rhs.size());
			for (ArrayIterator slot = _necNet.begin();!slot.atEnd();slot.advance())
				slot.put(new Array());
		}
		for (Enumeration constr = _neConstraints.elements();
		     constr.hasMoreElements();) {
		     //c = (NotExistsConstraint)constr.nextElement();
		     c = (ExistenceConstraint)constr.nextElement();
			((Array)_necNet.at(c.nonNullRef().position()-1)).add(c);
		}
	}

	private void installPredicates() {
		Verifiable v;
		if (_predNet == null && !_predicates.isEmpty()) {
			_predNet = new Array(_rhs.size());
			for (ArrayIterator slot = _predNet.begin();!slot.atEnd();slot.advance())
				slot.put(new Array());
		}
		for (Enumeration pred = _predicates.elements();
		     pred.hasMoreElements();) {
		     v = (Verifiable)pred.nextElement();
		     for (SListIterator ref = v.rhsReferences().begin();!ref.atEnd();
		          ref.advance()) {
		         SList restRefs = new SList(v.rhsReferences());
		         restRefs.remove(ref.get());
		     //System.err.println("#### reference: "+ref.get());
		     if (((ObjectReference)ref.get()).position()>0)
					((Array)_predNet.at(((ObjectReference)ref.get()).position()-1)).
						add(new Pair(
								Filtering.select(
									Transforming.collect(
										restRefs,
										new UnaryFunction () {
											public Object execute(Object oRef) {
												return new Integer(((ObjectReference)oRef).position());
											}
										}
									),
									new BindSecondPredicate(new GreaterNumber(),new Integer(0))	
								),
								v
							)
						);
				}
		}
	}

	private void checkConstraints() 
	throws InvalidConstraintException,InvalidProductionException {
		for (Enumeration constr = _constraints.elements();
		     constr.hasMoreElements();)
		     ((Constraint)constr.nextElement()).check(_grammar);
		installConstraints();

        if (_cNet == null) {
       		if (_rhs.size() > 1)
            	throw(new InvalidProductionException(this,
                      "No constraints defined between right-hand-side elements"));
        } else {
	        if (_rhs.size() == 1)
    	    	throw(new InvalidProductionException(this,
        	    	  "Only one symbol at right-hand-side: constraints are not needed"));
	        for (ArrayIterator slots = _cNet.begin();!slots.atEnd();slots.advance())
    	        if (((Array)slots.get()).isEmpty())
        	    	throw(new InvalidProductionException(this,
            	          "The symbol at position "+Integer.toString(slots.index()+1)+
                	      " should be referenced in some constraint"));
        }
	}
	

	private void checkNotExistsConstraints() 
	throws InvalidConstraintException,InvalidProductionException {
		for (Enumeration constr = _neConstraints.elements();
		     constr.hasMoreElements();)
		     //((NotExistsConstraint)constr.nextElement()).check(_grammar);
		     ((ExistenceConstraint)constr.nextElement()).check(_grammar);
		
		installNEConstraints();
	}

	private void checkPredicates() 
		throws InvalidConstraintException,InvalidImplicitPredicateException,
		       InvalidProductionException {
		for (Enumeration p = _predicates.elements();p.hasMoreElements();) {
		     Object pred = p.nextElement();
		     if (pred instanceof Constraint)
		     	((Constraint)pred).check(_grammar);
		     else
		     	((ImplicitPredicate)pred).check(_grammar);
		}
		installPredicates();
	}


	private void checkRhs() 
		throws InvalidProductionException {
		for (ArrayIterator sym = _rhs.begin();!sym.atEnd();sym.advance())
   			if (!_grammar.isTerminal((String)sym.get()))
    			if (!_grammar.isNonTerminal((String)sym.get()))
    				throw(
    					new InvalidProductionException(
    						this,
    						"Undefined terminal/non-terminal \""+
    						(String)sym.get()+"\" at right-hand-side"
    					)
    				);

	}

	public void assignments(String description) 
	throws InvalidProductionException {
    	try {
			Copying.copy(GrammarReader.parseAssignmentList(description,this),
						 _attrAssigns);
    	}
    	catch (IOException e) {
			throw(new InvalidProductionException(this,
			"internal error - read failed in GrammarReader.parseAssignmentList()"));
		}
		catch (ExpressionSyntaxException e) {
			throw(new InvalidProductionException(this,e.getMessage()));
		}
   	}
   	
   	private boolean mapAttribute(String attribute) {
   		if (_grammar.attributeMap().get(attribute)==null)
   			return false;
   		else {
   			String other_attribute = 
   				(String)_grammar.attributeMap().get(attribute);
   			AttributeAssignment assign = null;
   			for (ArrayIterator assgn=_attrAssigns.begin();!assgn.atEnd();assgn.advance())
				if (((AttributeAssignment)assgn.get()).attribute().equals(other_attribute)) {
					assign = (AttributeAssignment)assgn.get();
					break;
				}
			if (assign==null)
				return false;
			else 
				_attrAssigns.pushBack(assign.copy(attribute));	
			return true;
   		}
   	
   	}

	private void checkAssignments() 
	throws InvalidAttributeAssignmentException,
		   InvalidProductionException {
		for (Enumeration assgns = _attrAssigns.elements();
		     assgns.hasMoreElements();)
		     ((AttributeAssignment)assgns.nextElement()).check(_grammar);
		Array attributes = 
			(Array)Transforming.collect(_attrAssigns,new GetAttribute());
		for (Enumeration attrs = _grammar.expanderAttributes().elements();
			 attrs.hasMoreElements();) {
			 String attribute = (String)attrs.nextElement();
			 if (!attributes.contains(attribute))
			 	if (!mapAttribute(attribute))
			 		throw(new InvalidProductionException(this,
			 			  "No assignment for attribute \""+attribute+"\""));
		}
	}
	
	public Array orderingVariants() throws InvalidProductionException {
		try {
			return _strategy.orderingVariants();
		}
		catch (RhsOrderingException e) {
			throw(e);
		}
	}
	
	public void semantics(String className) {
		_semanticsClassName = className;
	}

	public void checkSemantics()
		throws InvalidProductionException {
		if (_semanticsClassName != null)
			try {
				_semanticsClass = Class.forName(_semanticsClassName);
			} 
			catch (ClassNotFoundException e) {
				throw new InvalidProductionException(
					this,
					"Semantics class "+_semanticsClassName+" not found"
				);
			}
	}
	
	public SyntaxTreeNode getSemantics() {
		if (_semanticsClass != null) {
			Constructor cons;
        	try {
        		cons = _semanticsClass.getConstructor(new Class[0]);
        		//System.err.println(cons);
        		return (SyntaxTreeNode)cons.newInstance(new Object[0]);
        	} catch (IllegalAccessException e) { 
        		System.err.println(e);
        	} catch (InstantiationException e) { 
        		System.err.println(e);
        	} catch (InvocationTargetException e) { 
        		System.err.println(e);
        	} catch (NoSuchMethodException e) {
        		System.err.println(e);
        	}		
        	return null;
		}
		else
			return getDefaultSemantics();
	}
	
	protected SyntaxTreeNode getDefaultSemantics() {
		return new relap.Analyzer.MiddleNode();
	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("Description: "+_description+"\n");
		buf.append(_lhs+" -> ");
		int counter = 0;
		for (Enumeration arIt = _rhs.elements();arIt.hasMoreElements();) {
			buf.append((String)arIt.nextElement());
			counter++;
			if (isRepetitive(counter) && !isOptional(counter))
				buf.append("+");
			else if (isRepetitive(counter) && isOptional(counter))
				buf.append("*");
			else if (!isRepetitive(counter) && isOptional(counter))
				buf.append("?");
			buf.append(" ");
		}
		buf.append("\n");
		for (Enumeration arIt = _constraints.elements();arIt.hasMoreElements();)
			buf.append(arIt.nextElement().toString()+"\n");
		for (Enumeration arIt = _neConstraints.elements();arIt.hasMoreElements();)
			buf.append(arIt.nextElement().toString()+"\n");
		for (Enumeration arIt = _predicates.elements();arIt.hasMoreElements();)
			buf.append(arIt.nextElement().toString()+"\n");
		if (_orderingExpression != null)
			buf.append(_orderingExpression.trim()+"\n");
		for (Enumeration arIt = _attrAssigns.elements();arIt.hasMoreElements();)
			buf.append(arIt.nextElement().toString()+" "+"\n");
		if (_semanticsClassName == null)
			buf.append("Default semantics");
		else
			buf.append("Syntax-tree node class: '"+_semanticsClassName+"'");
		
		return buf.toString();
	}
}