package relap.LanguageModel;

import com.objectspace.jgl.Array;
import com.objectspace.jgl.Pair;
import com.objectspace.jgl.Container;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.DList;
import com.objectspace.jgl.SListIterator;
import java.util.*;
import java.io.*;


public class GrammarReader {
	
	public static void parseRightHandSide(String description,
						 				  GrammarProduction production) 
	throws IOException,ExpressionSyntaxException {
		StreamTokenizer tokens = new StreamTokenizer(new StringReader(description));
		tokens.ordinaryChar('+');
		tokens.ordinaryChar('?');
		tokens.ordinaryChar('*');
    	DList isRepetitive = new DList();
    	DList isOptional = new DList();
    	while (tokens.nextToken() != StreamTokenizer.TT_EOF) {
    		if (tokens.ttype != StreamTokenizer.TT_WORD)
				throw(new ExpressionSyntaxException(
							"non-terminal or terminal symbol",
							tokens.toString(),
							description)
				);
    		production.addToRhs(tokens.sval);
   			if (tokens.nextToken() != StreamTokenizer.TT_EOF) {
   				if (tokens.ttype == '+') {
   					isRepetitive.pushBack(new Boolean(true));
   					isOptional.pushBack(new Boolean(false));
   				} else if (tokens.ttype == '?') {
   					isRepetitive.pushBack(new Boolean(false));
   					isOptional.pushBack(new Boolean(true));
   				} else if (tokens.ttype == '*') {
   					isRepetitive.pushBack(new Boolean(true));
   					isOptional.pushBack(new Boolean(true));
   				} else {
   					isRepetitive.pushBack(new Boolean(false));
		    		isOptional.pushBack(new Boolean(false));
   					tokens.pushBack();
   				}
   			} else {
   				isRepetitive.pushBack(new Boolean(false));
    			isOptional.pushBack(new Boolean(false));
   			}
    	}
    	production.setRepetitionFlags(isRepetitive);
    	production.setOptionalityFlags(isOptional);
    }	
    
	public static void parseOrderingExpression(String description,
						 					   GrammarProduction production) 
	throws IOException,ExpressionSyntaxException {
		SList positions = new SList();
		StreamTokenizer tokens = new StreamTokenizer(new StringReader(description));
		tokens.ordinaryChar(',');
    	do { 
			tokens.nextToken();
    		if (tokens.ttype != StreamTokenizer.TT_NUMBER)
				throw(
					new ExpressionSyntaxException(
						"integer",tokens.toString(),description
					)
				);  
			if ((int)tokens.nval <= 0 || (int)tokens.nval > production.rhs().size())
				throw(
					new ExpressionSyntaxException(
						"integer in range 1.."+production.rhs().size(),
						tokens.toString(),
						description
					)
				);  
			positions.pushBack(new Integer((int)tokens.nval));
    		tokens.nextToken();
    	} while (tokens.ttype == ',');
    	if (tokens.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException(
				"ordering predicate name",tokens.toString(),description)
			);    		
    	for	(SListIterator pos = positions.begin();!pos.atEnd();pos.advance()) {
    		if (production.orderingPredicates()
    				.at(((Integer)pos.get()).intValue()-1) != null)
    			throw(
    				new ExpressionSyntaxException(
						"more than one ordering predicate for position "+(Integer)pos.get(),
						description
					)
    			); 
    		production.orderingPredicates()
    			.put(
    				((Integer)pos.get()).intValue()-1,
    				new Pair(new Integer(pos.index()),tokens.sval)
    			);
    	}
   	}
	
	public static Container parseConstraintList(String description,
						 					 GrammarProduction production) 
	throws IOException,ExpressionSyntaxException {
		Array constraints = new Array();
		StreamTokenizer tokens = new StreamTokenizer(new StringReader(description));
		tokens.ordinaryChar('(');
		tokens.ordinaryChar(')');
		tokens.ordinaryChar(':');
		tokens.ordinaryChar(',');
		tokens.ordinaryChar('_');
    	while (tokens.nextToken() != StreamTokenizer.TT_EOF) 
    		constraints.add(GrammarReader.parseExpanderConstraint(tokens,description,production));
		return constraints;
   	}
	
	public static Container parseNotExistsConstraintList(String description,
						 					 			 GrammarProduction production) 
	throws IOException,ExpressionSyntaxException {
		Array constraints = new Array();
		StreamTokenizer tokens = new StreamTokenizer(new StringReader(description));
		tokens.ordinaryChar('(');
		tokens.ordinaryChar(')');
		tokens.ordinaryChar(':');
		tokens.ordinaryChar(',');
		tokens.ordinaryChar('_');
		tokens.ordinaryChar('~');
		tokens.ordinaryChar('{');
		tokens.ordinaryChar('}');
    	while (tokens.nextToken() != StreamTokenizer.TT_EOF) 
    		//constraints.add(GrammarReader.parseNotExistsConstraint(tokens,description,production));
    		constraints.add(GrammarReader.parseExistenceConstraint(tokens,description,production));
		return constraints;
   	}

	public static Container parsePredicateList(String description,
						 					   GrammarProduction production) 
	throws IOException,ExpressionSyntaxException {
		Array constraints = new Array();
		StreamTokenizer tokens = new StreamTokenizer(new StringReader(description));
		tokens.ordinaryChar('(');
		tokens.ordinaryChar(')');
		tokens.ordinaryChar(':');
		tokens.ordinaryChar(',');
		tokens.ordinaryChar('_');
    	while (tokens.nextToken() != StreamTokenizer.TT_EOF) 
    		constraints.add(GrammarReader.parsePredicate(tokens,description,production));
		return constraints;
   	}
	
	private static Object parsePredicate(StreamTokenizer descriptor,
											  String descriptorString,
						 					  GrammarProduction production)
	throws ExpressionSyntaxException,IOException {
		if (descriptor.ttype == StreamTokenizer.TT_WORD &&
			  production.grammar().relations().get(descriptor.sval)==null) {							
		//if (descriptor.sval.equals("implicit")) {
			String method;
			SList arguments = new SList();
			//descriptor.nextToken();
			//if (descriptor.ttype != StreamTokenizer.TT_WORD)
			//	throw(new ExpressionSyntaxException("predicate method name",
			//										descriptor.toString(),
			//										descriptorString));
			method = descriptor.sval;
			descriptor.nextToken();
			if (descriptor.ttype != '(')
				throw(new ExpressionSyntaxException("(", 
													descriptor.toString(),
													descriptorString));
			descriptor.nextToken();	
			do {
				arguments.pushBack(
					GrammarReader.parseReference(
						descriptor,descriptorString,production
					)
				);
				if (descriptor.ttype == ',')
					descriptor.nextToken();
				else if (descriptor.ttype != ')')
					throw(new ExpressionSyntaxException(")", 
														descriptor.toString(),
														descriptorString));
			} while (descriptor.ttype != ')');
			return
				new ImplicitPredicate(method,relap.Util.Utils.asArray(arguments));							  
		}
		else if (descriptor.ttype == '~')
			return GrammarReader.parseExistenceConstraint(descriptor,descriptorString,production);
		else 
			return GrammarReader.parseAnyConstraint(descriptor,descriptorString,production);
	}

	
	private static Constraint parseAnyConstraint(StreamTokenizer descriptor,
																String descriptorString,
						 										GrammarProduction production) 
	throws ExpressionSyntaxException,IOException {
		String relation;
		ObjectReference firstArg;
		ObjectReference secondArg;
		SList terminalTypes = new SList();
		int exactly = 0;
		boolean isExact = false;
	
		if (descriptor.ttype == StreamTokenizer.TT_NUMBER) { // ExistenceConstraint
			exactly = (int)descriptor.nval;
			if (exactly < 0)
				throw(
					new ExpressionSyntaxException("nonnegative number",
					descriptor.toString(),
					descriptorString)
				);
			isExact = true;
			descriptor.nextToken();
		}
		if (descriptor.ttype == '{') {
			do {
				descriptor.nextToken();
				if (descriptor.ttype != StreamTokenizer.TT_WORD)
					throw(new ExpressionSyntaxException("terminal symbol",
														descriptor.toString(),
														descriptorString));
				terminalTypes.pushBack(descriptor.sval);
				descriptor.nextToken();
			} 
			while (descriptor.ttype == ',');
			if (descriptor.ttype != '}')
				throw(new ExpressionSyntaxException("}",
													descriptor.toString(),
													descriptorString));
			descriptor.nextToken();
		}
		if (descriptor.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("relation name",
												descriptor.toString(),
												descriptorString));
		relation = descriptor.sval;
		descriptor.nextToken();
		if (descriptor.ttype != '(')
			throw(new ExpressionSyntaxException("(", 
												descriptor.toString(),
												descriptorString));
		descriptor.nextToken();									  
		firstArg = GrammarReader.parseReference(descriptor,descriptorString,production);
		if (descriptor.ttype != ',')
			throw(new ExpressionSyntaxException(",",
												descriptor.toString(),
												descriptorString));
		descriptor.nextToken();									  
		secondArg = GrammarReader.parseReference(descriptor,descriptorString,production);
		if (descriptor.ttype != ')')
			throw(new ExpressionSyntaxException(")",
												descriptor.toString(),
												descriptorString));	
		if (isExact)
			return	
		 		new ExistenceConstraint(
					relation,new Pair(firstArg,secondArg),terminalTypes,exactly
				);
		else if (!terminalTypes.isEmpty() ||
			        (firstArg instanceof NullReference) || (secondArg instanceof NullReference))
			return 
				new Some(
					relation,new Pair(firstArg,secondArg),terminalTypes
				);

		return	new Constraint(relation,new Pair(firstArg,secondArg));
	}

private static ExistenceConstraint parseExistenceConstraint(StreamTokenizer descriptor,
																String descriptorString,
						 										GrammarProduction production) 
	throws ExpressionSyntaxException,IOException {
		String relation;
		ObjectReference firstArg;
		ObjectReference secondArg;
		SList terminalTypes = new SList();
		int exactly = 0;
		boolean isExact = false;
	
		if (descriptor.ttype == StreamTokenizer.TT_NUMBER) { // ExistenceConstraint
			exactly = (int)descriptor.nval;
			if (exactly < 0)
				throw(
					new ExpressionSyntaxException("nonnegative number",
					descriptor.toString(),
					descriptorString)
				);
			isExact = true;
			descriptor.nextToken();
		}
		if (descriptor.ttype == '~') { //not exists -constraint
			isExact = true;
			exactly = 0;
			descriptor.nextToken();
		}
		if (descriptor.ttype == '{') {
			do {
				descriptor.nextToken();
				if (descriptor.ttype != StreamTokenizer.TT_WORD)
					throw(new ExpressionSyntaxException("terminal symbol",
														descriptor.toString(),
														descriptorString));
				terminalTypes.pushBack(descriptor.sval);
				descriptor.nextToken();
			} 
			while (descriptor.ttype == ',');
			if (descriptor.ttype != '}')
				throw(new ExpressionSyntaxException("}",
													descriptor.toString(),
													descriptorString));
			descriptor.nextToken();
		}
		if (descriptor.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("relation name",
												descriptor.toString(),
												descriptorString));
		relation = descriptor.sval;
		descriptor.nextToken();
		if (descriptor.ttype != '(')
			throw(new ExpressionSyntaxException("(", 
												descriptor.toString(),
												descriptorString));
		descriptor.nextToken();									  
		firstArg = GrammarReader.parseReference(descriptor,descriptorString,production);
		if (descriptor.ttype != ',')
			throw(new ExpressionSyntaxException(",",
												descriptor.toString(),
												descriptorString));
		descriptor.nextToken();									  
		secondArg = GrammarReader.parseReference(descriptor,descriptorString,production);
		if (descriptor.ttype != ')')
			throw(new ExpressionSyntaxException(")",
												descriptor.toString(),
												descriptorString));	
		if (isExact) {
			if (exactly > 0)
				return	
		 			new ExistenceConstraint(
						relation,new Pair(firstArg,secondArg),terminalTypes,exactly
					);
			else
				return 
					new NotExists(
						relation,new Pair(firstArg,secondArg),terminalTypes
					);
		}
		return 
			new Some(relation,new Pair(firstArg,secondArg),terminalTypes);
	}


	private static Constraint parseExpanderConstraint(StreamTokenizer descriptor,
											  String descriptorString,
						 					  GrammarProduction production) 
	throws ExpressionSyntaxException,IOException {
		String relation;
		ObjectReference firstArg;
		ObjectReference secondArg;
		if (descriptor.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("relation name",
												descriptor.toString(),
												descriptorString));
		relation = descriptor.sval;
		descriptor.nextToken();
		if (descriptor.ttype != '(')
			throw(new ExpressionSyntaxException("(", 
												descriptor.toString(),
												descriptorString));
		descriptor.nextToken();									  
		firstArg = GrammarReader.parseReference(descriptor,descriptorString,production);
		if (descriptor.ttype != ',')
			throw(new ExpressionSyntaxException(",",
												descriptor.toString(),
												descriptorString));
		descriptor.nextToken();									  
		secondArg = GrammarReader.parseReference(descriptor,descriptorString,production);
		if (descriptor.ttype != ')')
			throw(new ExpressionSyntaxException(")",
												descriptor.toString(),
												descriptorString));		
		return new Constraint(relation,new Pair(firstArg,secondArg));
	}

	

	public static Container parseAssignmentList(String description,
						 					 	GrammarProduction production)
	throws IOException,ExpressionSyntaxException {
		Array assignments = new Array();
		StreamTokenizer tokens = new StreamTokenizer(new StringReader(description));
		tokens.ordinaryChar(':');
		tokens.ordinaryChar('=');
		tokens.nextToken();
    	while (tokens.ttype != StreamTokenizer.TT_EOF) 
    		assignments.add(GrammarReader.parseAssignment(tokens,description,production));
		return assignments;
   	}

	private static AttributeAssignment parseAssignment(StreamTokenizer descriptor,
													   String descriptorString,
						 							   GrammarProduction production) 
	throws ExpressionSyntaxException,IOException {
		String relation;
		ObjectReference firstArg;
		ObjectReference secondArg;

		firstArg = GrammarReader.parseReference(descriptor,descriptorString,production);
		if (descriptor.ttype != '=')
			throw(new ExpressionSyntaxException("=",
												descriptor.toString(),
												descriptorString));
		descriptor.nextToken();									  
		secondArg = GrammarReader.parseReference(descriptor,descriptorString,production);

		return new AttributeAssignment(firstArg,secondArg);
	}

	private static ObjectReference parseReference(StreamTokenizer descriptor,
										   		  String descriptorString,
												  GrammarProduction production) 
	throws ExpressionSyntaxException,IOException {
		int pos;
		switch (descriptor.ttype){
			case '_' : 
			{
				descriptor.nextToken();
				return new NullReference(production);
			}
			case StreamTokenizer.TT_NUMBER :
			{				
				pos = (int)descriptor.nval;
				descriptor.nextToken();
				if (descriptor.ttype == ':') {
					descriptor.nextToken();
					if (descriptor.ttype != StreamTokenizer.TT_WORD)
						throw(new ExpressionSyntaxException("attribute name",
															descriptor.toString(),
															descriptorString));
					String attribute = descriptor.sval;
					descriptor.nextToken();
					return new InDirectReference(production,attribute,pos);
				}
				return new DirectReference(production,pos);
			}
			default : throw(new ExpressionSyntaxException("integer or _",
														  descriptor.toString(),
													      descriptorString));
		}
	}

}