package relap.LanguageModel;

import com.objectspace.jgl.Pair;
import com.objectspace.jgl.Array;
import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.adapters.ObjectArray;
import com.objectspace.jgl.algorithms.Copying;
import relap.Analyzer.FailureData;
import relap.Analyzer.ImplicitPredicateFailureData;
import relap.Util.Utils;
import relap.Util.ArrayComparator;
import java.util.Enumeration;
import java.lang.reflect.*;

public class ImplicitPredicate
	implements Verifiable {

	private String fName;
	private Array fArgumentObjects;
	private Object fEvaluator;
	private Method fPredicate;
	private SList fLatestFailures;
	
	
	public ImplicitPredicate(String name,Array objects,Object evaluator,
	                         Method predicateMethod) {
		fName = name;
		fArgumentObjects = objects;
		fEvaluator = evaluator;
		fPredicate = predicateMethod; 
		fLatestFailures = new SList();
	}	

	public ImplicitPredicate(String name,Array objects,Object evaluator) {
		fName = name;
		fArgumentObjects = objects;
		fEvaluator = evaluator; 
		fLatestFailures = new SList();
	}	

	public ImplicitPredicate(String name,Array objects) {
		fName = name;
		fArgumentObjects = objects;
		fEvaluator = null; 
		fLatestFailures = new SList();
	}	
	
	public String getName() {
		return fName;
	}
	
	public void check(RelationalGrammarImplementation grammar) 
	throws InvalidImplicitPredicateException {
		boolean methodFound = false;
		if (fEvaluator == null)
			fEvaluator = grammar.getPredicateEvaluator();
			
		Method evaluatorMethods[] = fEvaluator.getClass().getMethods();
		Class methodArgs[] = new Class[fArgumentObjects.size()];
		try {
			for (int i=0;i<methodArgs.length;i++)
				methodArgs[i] = Class.forName("java.lang.Object");
			for (int i=0;i<evaluatorMethods.length;i++) {
				/*
				System.err.println(
					"evaluatorMethods[i].getName().equals(fName) == "+
					evaluatorMethods[i].getName().equals(fName)
				);
				*/
				if (evaluatorMethods[i].getName().equals(fName)) {
					/*
					System.err.println(
					"evaluatorMethods[i].getReturnType() == boolean.class == "+
					(evaluatorMethods[i].getReturnType() == 
				    	boolean.class)
					);
					*/
					if (evaluatorMethods[i].getReturnType() == 
				    	boolean.class) {
						/*
						System.err.println(evaluatorMethods[i]);
				    	System.err.println(
				    		"ArrayComparator.equalTo(evaluatorMethods[i].getParameterTypes(),methodArgs) == "+
				    		(ArrayComparator.equalTo(evaluatorMethods[i].getParameterTypes(),methodArgs))
				    	);
				    	*/
				    	if (ArrayComparator.equalTo(
				    			evaluatorMethods[i].getParameterTypes(),
				    			methodArgs)
				    	   )
						{
							methodFound = true;
							fPredicate = evaluatorMethods[i];
							break;
						}
					}
				}
				//System.err.println(evaluatorMethods[i]);
			}
		} catch (ClassNotFoundException e) {
			System.err.println(e);
		}
		if (!methodFound)
			throw(
				new InvalidImplicitPredicateException(
					this,
					"No evaluation method defined for this predicate in "+
					fEvaluator.getClass()
				)
			);
	
		try {
			Enumeration args = fArgumentObjects.elements();
			while (args.hasMoreElements()) 
				((ObjectReference)args.nextElement()).checkRhsReference(grammar);
		}
		catch(InvalidObjectReferenceException e) {
			throw(new InvalidImplicitPredicateException(this,e.getMessage()));
		}
	}
	
   
	public String toString() {
    	return 
    		"implicit "+fName+
    	    "("+Utils.listContents(fArgumentObjects,",",",")+")";
    }
    
	public boolean verify(relap.Analyzer.ParseState state,
				          relap.Analyzer.IndexedMDSet input)
		throws InvalidQueryArgumentsException {
		boolean result = true;
		Array argArray = new Array(fArgumentObjects.size());
		for (int i = 0;i<argArray.size();i++)
			argArray.put(
				i, 
				state.dereference((ObjectReference)fArgumentObjects.at(i))
			);
			
		Enumeration args = Utils.cartesianProduct(argArray).elements();
		while (args.hasMoreElements()) {
			Object arguments[] = new Object[fArgumentObjects.size()];
			Array array = (Array)args.nextElement();
			for (int i = 0;i<arguments.length;i++)
				arguments[i] = array.at(i);
			try {
				if (((Boolean)fPredicate.invoke(fEvaluator,arguments))
						.booleanValue() 
					== false) {
					fLatestFailures.pushBack(
						new ImplicitPredicateFailureData(new Array(arguments))
					);	
					result = false;
				} 
			} catch (IllegalAccessException e) {
				System.err.println(e);
		
			} catch (IllegalArgumentException e) {
				System.err.println(e);
		
			} catch (InvocationTargetException e) {
				System.err.println(e);
		
			}
		}
		return result;
	}

	public SList getFailureData() {
		return fLatestFailures;
	}

	public SList rhsReferences() {
		return Utils.asSList(fArgumentObjects);
	}
	
	public Object getEvaluator() {
		return fEvaluator;
	}

	public Method getPredicateMethod() {
		return fPredicate;
	}
}
