package relap.LanguageModel;


public class InDirectReference extends ObjectReference {

	private String _attribute;
	
	public InDirectReference(GrammarProduction production,int position) {
		super(production,position);
	}
	public InDirectReference(GrammarProduction production,String attribute,
							 int position) {
		super(production,position);
		_attribute = attribute;
	}

	public ObjectReference copy() { 
		return new InDirectReference(production(),_attribute,position());
	}

	public String attribute() { return _attribute; }
	public void attribute(String attr) { _attribute = attr; }

	public void checkLhsReference(RelationalGrammarImplementation grammar) 
		throws InvalidObjectReferenceException {
		if (position() != 0)
			throw(new InvalidObjectReferenceException(this,"reference with index 0 expected"));
		checkAttribute(grammar);
	}

	public void checkRhsReference(RelationalGrammarImplementation grammar) 
		throws InvalidObjectReferenceException {
		if (position() < 1 || position() > production().rhs().size())
			throw(new InvalidObjectReferenceException(this,
						"reference to a rhs element out of bounds"));
		String symbol = (String)(production().rhs().at(position()-1));
		if (!grammar.isNonTerminal(symbol))
			throw(new InvalidObjectReferenceException(this,
						"symbol at position "+Integer.toString(position())+
						" is not a non-terminal"));
		checkAttribute(grammar);
	}			

	private void checkAttribute(RelationalGrammarImplementation grammar)
		throws InvalidObjectReferenceException {
		if (grammar.expanderAttributes().get(_attribute) == null)
			throw(new InvalidObjectReferenceException(this,
				"Undefined attribute \""+_attribute+"\""));
	}

	public boolean isInDirectReference() { return true; }

	public String toString() {
		return super.toString()+":"+_attribute;
	}
}
