package relap.LanguageModel;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.algorithms.Transforming;
import com.objectspace.jgl.predicates.BindSecondPredicate;
import java.util.*;
import relap.Util.Utils;

public class OrderingVariant 
	implements SymbolSequence {

	private GrammarProduction _production;
	private Array _permutation;
	private Array _inverse_permutation;
	private Array _rhs;
	private Array _constraintNetwork;
	private Array _ne_constraint_network;
	private Array _predicate_network;
	private HashMap _attributeAssignments;
	private HashSet _attributesBoundByFirst;
	private HashSet _firstAttrs;
	
	public OrderingVariant(GrammarProduction production) {
		_production = production;
		_permutation = new Array();
		_rhs = new Array();
		_attributeAssignments = new HashMap();
	}


	public HashMap attributeAssignments() {
		return _attributeAssignments;
	}

	class FirstGivesValue implements UnaryPredicate {
    	public boolean execute(Object arg) {
    		return ((AttributeAssignment)arg).value().position() == 1;
   		}
   	} 
	class GetAttribute implements UnaryFunction {
    	public Object execute(Object arg) {
    		return ((AttributeAssignment)arg).attribute();
   		}
   	} 
  
	public boolean bindsFirst(String attribute) {
        //Answer true if the receiver's first rhs elements provides a binding for
        //anAttribute; otherwise answer false.
    	HashSet assignments;
    	if (_attributesBoundByFirst == null) {
        	assignments = new HashSet();
         	for (HashMapIterator assgns=_attributeAssignments.begin();
         	     !assgns.atEnd();assgns.advance())
         		if ((new FirstGivesValue()).execute(assgns.value()))
         	     	assignments.add(assgns.value());
        	_attributesBoundByFirst = new HashSet(); 
        	Copying.copy(Transforming.collect(assignments,new GetAttribute()),
        	 			 _attributesBoundByFirst);
		}
    	return _attributesBoundByFirst.get(attribute) != null;
	}
 
	public Constraint constraintAt(int pos) {
        //Answer the constraint pertaining to the rhs element at position anInteger and
        //some earlier element in this permutation
    	Constraint constraint = null;
    	if (_constraintNetwork != null)
	    	for (ArrayIterator it = ((Array)_constraintNetwork.at(pos-1)).begin();
    			 !it.atEnd();it.advance())
        	    if (((DirectReference)((Pair)it.get()).first).position() < pos) 
            		constraint = (Constraint)((Pair)it.get()).second;
		return constraint;
	}
   
	public SList constraintsAt(int pos) {
        //Answer the constraints pertaining to the rhs element at position anInteger and
        //all the earlier elements in this permutation
		return networkObjectsAt(_constraintNetwork,pos); 
	}

	public SList neConstraintsAt(int pos) {
    	SList constraints = new SList();
    	if (_ne_constraint_network != null)
    		Copying.copy(
    			(Array)_ne_constraint_network.at(pos-1),
            	constraints
            );
		return constraints;
	}

	public SList predicatesAt(int pos) {
    	SList predicates = new SList();
    	if (_predicate_network != null)
    		Copying.copy(
    			Filtering.select(
    				(Array)_predicate_network.at(pos-1),
    				new BindSecondPredicate(
    					new BinaryPredicate() {
    						public boolean execute(Object pair,Object anInt) {
    							boolean ok = true;
    							for (SListIterator ref = ((SList)((Pair)pair).first).begin();
    							     !ref.atEnd();ref.advance()) {
       								if (((DirectReference)ref.get()).position() 
    								      >= ((Integer)anInt).intValue()) {
    									ok = false;
    									break;
    								}
    							}
    							return ok;
    						}	
    					},
    					new Integer(pos)
    				)
    			),
            	predicates
            );
		return 
			Utils.asSList(
			Transforming.collect(
				predicates,
				new UnaryFunction() {
					public Object execute(Object pair) {
						return ((Pair)pair).second;
					}
				}
			)
			);
	}

	public String constraintsToString() {
		if (_constraintNetwork != null)
			return _constraintNetwork.toString();
		return "no constraints";
	}

	public String nEConstraintsToString() {
		if (_ne_constraint_network != null)
			return _ne_constraint_network.toString();
		return "no ne-constraints";
	}

	public String predicatesToString() {
		if (_predicate_network != null)
			return _predicate_network.toString();
		return "no predicates";
	}

	private SList networkObjectsAt(Array network,int pos) {
        //Answer the objects pertaining to the rhs element at position anInteger and
        //all the earlier elements in this permutation
    	SList constraints	= new SList();
    	if (network != null)
    		for (ArrayIterator it = ((Array)network.at(pos-1)).begin();
    			 !it.atEnd();it.advance())
            	if (((DirectReference)((Pair)it.get()).first).position() < pos) 
            		constraints.pushBack(((Pair)it.get()).second);
		return constraints;
	}


	class FirstGivesValueByAttribute implements UnaryPredicate {
    	public boolean execute(Object arg) {
    		return ((AttributeAssignment)((Pair)arg).second).value().position() == 1 &&
    				((AttributeAssignment)((Pair)arg).second).value().isInDirectReference();
    	}
    } 
	class GetAttributeOfValueExpr implements UnaryFunction {
    	public Object execute(Object arg) {
    		return ((InDirectReference)((AttributeAssignment)((Pair)arg).second).value()).attribute();
    	}
    } 
 
	public HashSet firstAttrs() {
        //Answer the a collection of expander attributes for which the first rhs element of this ordering
        //provides a binding.
    	if (_firstAttrs==null) {
    		_firstAttrs = new HashSet();
    		HashSet temp = new HashSet();
        	Copying.copy(
            	Filtering.select(_attributeAssignments,
            					 new FirstGivesValueByAttribute()),
            	temp
            ); 
			_firstAttrs = 
				(HashSet)Transforming.collect(temp,new GetAttributeOfValueExpr());
        	
    	}
    	return _firstAttrs;
    }


	
	public GrammarProduction production() { return _production; }
	
	public OrderingVariant permute(Container permutation) {
		Copying.copy(permutation,_permutation);
		inversePermute();
		permuteRhs();
    	if(_production.cNet() != null) {
    		_constraintNetwork = new Array(_production.cNet().size());
			permuteConstraints(_constraintNetwork,_production.cNet());
		}
    	if(_production.necNet() != null) {
    		_ne_constraint_network = new Array(_production.necNet().size());
			permuteNEConstraints(_ne_constraint_network,_production.necNet());
		}
    	if(_production.predNet() != null) {
    		_predicate_network = new Array(_production.predNet().size());
			permutePredicates(_predicate_network,_production.predNet());
		}
		permuteAttributeAssignments();
		/*
		System.err.println(
			_production+"\npermuted->"+
			toString()+
			"\nconstraints "+constraintsToString()+
			"\nne-constraints "+nEConstraintsToString()+
			"\npredicates "+predicatesToString()
		);
		*/
		return this;
	}
	
	private void inversePermute() {
		_inverse_permutation = new Array(_permutation.size());
        for (int i = 0; i<_inverse_permutation.size();i++) {
            _inverse_permutation.put(((ObjectReference)_permutation.at(i)).position()-1,
                                     new DirectReference(_production,i+1));
        }
	}		
	
	public Array inversePermutation() {
		return _inverse_permutation;
	}		
	
	private void permuteRhs() {
		for (int i=0;i<_production.rhs().size();i++)
			_rhs.add(_production.rhs().at(((ObjectReference)_permutation.at(i)).position()-1));	
	}
	
	public boolean isRepetitive(int pos) {
		return 
			_production.isRepetitive(
				((ObjectReference)_inverse_permutation.at(pos-1)).position()
			);
	}

	public boolean isOptional(int pos) {
		return 
			_production.isOptional(
				((ObjectReference)_inverse_permutation.at(pos-1)).position()
			);
	}
	
    public int sequenceLength() {
    	return _rhs.size();
    }

	private Constraint permuteConstraint(Constraint constraint) {
    	Pair references  =
    		new Pair(
    			constraint.firstArg().copy(),
    			constraint.secondArg().copy()
    		);
    	((ObjectReference)references.first).position(
			((ObjectReference)_inverse_permutation.at(constraint.firstArg().position()-1))
    		.position()    	
    	);
    	((ObjectReference)references.second).position(
    		((ObjectReference)_inverse_permutation.at(constraint.secondArg().position()-1))
    		.position()
    	);
    	return new Constraint(constraint.relation(),references);
    }
  
  
 	private ExistenceConstraint permuteExistenceConstraint(ExistenceConstraint constraint) {
    	Pair references  =
    		new Pair(
    			constraint.firstArg().copy(),
    			constraint.secondArg().copy()
    		);
    	if (constraint.nullRefPos()==2)
    		((ObjectReference)references.first).position(
				((ObjectReference)_inverse_permutation.at(constraint.nonNullRef().position()-1))
    			.position()    	
    		);
    	else
    		((ObjectReference)references.second).position(
    			((ObjectReference)_inverse_permutation.at(constraint.nonNullRef().position()-1))
    			.position()
    		);
    	// an ugly hack - make a constraint clone itself and then set permute references
    	// or, make a constraint permute itself
    	if (constraint instanceof Some)
    		return 
    			new Some(
    				constraint.relation(),
    				references,
    				constraint.copyConstrainedTerminals()
    			);	
    	else if (constraint instanceof ExistenceConstraint) 
    		return 
    			new ExistenceConstraint(
    				constraint.relation(),
    				references,
    				constraint.copyConstrainedTerminals(),
    				constraint.getExpectedNumber()
    			);
    	else if (constraint instanceof NotExists)
    		return 
    			new NotExists(
    				constraint.relation(),
    				references,
    				constraint.copyConstrainedTerminals()
    			);
    	else
    		return null; // throw some exception!
    		
    }

   private Verifiable permutePredicate(Verifiable verifiable) {
    	if (verifiable instanceof ImplicitPredicate) {
    		ImplicitPredicate predicate = (ImplicitPredicate)verifiable;
    		SList rhsRefs = new SList();
    		for (SListIterator ref = predicate.rhsReferences().begin();
    			 !ref.atEnd();ref.advance()) {
    			 ObjectReference newRef = ((ObjectReference)ref.get()).copy();
    			 newRef.position(
    			 	((ObjectReference)_inverse_permutation.at(newRef.position()-1))
    					.position()
    			 );
    			 rhsRefs.pushBack(newRef);
    		}
    		return 
    			new ImplicitPredicate(
    				predicate.getName(),
    				Utils.asArray(rhsRefs),
    				predicate.getEvaluator(),
    				predicate.getPredicateMethod()
    			);
       	}
      else if (verifiable instanceof ExistenceConstraint)
      	return permuteExistenceConstraint((ExistenceConstraint)verifiable);
      else
    		return permuteConstraint((Constraint)verifiable);
    }
    
   private void permutePredicates(Array local_network,Array production_network) {
    	for (int i=0;i<local_network.size();i++) {
    		local_network.put(i,new Array());
    		for (ArrayIterator companions = 
    			 ((Array)production_network.at(((ObjectReference)_permutation.at(i)).position()-1)).begin();
    			 !companions.atEnd();companions.advance()) {
    			 	SList refs = new SList();
    			 	for (SListIterator pos = ((SList)((Pair)companions.get()).first).begin();
    			 	     !pos.atEnd();pos.advance())
    			 	    refs.pushBack(
    			 	     	new DirectReference(
    			 	     		_production,
    							((ObjectReference)
    								_inverse_permutation.at(
    									((Integer)pos.get()).intValue()-1
    								)
   								).position()
    								
    						)
    					);
     			 	     		
    				((Array)local_network.at(i)).add(
    					new Pair(
    						refs,
    						permutePredicate((Verifiable)((Pair)companions.get()).second)
    					)
    					
    				);
    		}
    	}
    }


   private void permuteNEConstraints(Array local_network,Array production_network) {
    	for (int i=0;i<local_network.size();i++) {
    		local_network.put(i,new Array());
    		for (ArrayIterator companions = 
    			 ((Array)production_network.at(((ObjectReference)_permutation.at(i)).position()-1)).begin();
    			 !companions.atEnd();companions.advance())
    				((Array)local_network.at(i)).add(
    					/*permuteNotExistsConstraint(
    						(NotExistsConstraint)companions.get()
    					)*/
    					permuteExistenceConstraint(
    						(ExistenceConstraint)companions.get()
    					)
    					
    				);
    		
    	}
    }

   private void permuteConstraints(Array local_network,Array production_network) {
    	for (int i=0;i<local_network.size();i++) {
    		local_network.put(i,new Array());
    		for (ArrayIterator companions = 
    			 ((Array)production_network.at(((ObjectReference)_permutation.at(i)).position()-1)).begin();
    			 !companions.atEnd();companions.advance())
    				((Array)local_network.at(i)).add(
    					new Pair(
    						new DirectReference(
    							_production,
    							(
    								(ObjectReference)_inverse_permutation.at(
    									((Integer)
   											((Pair)companions.get()).first
   										).intValue()-1
    								)
   								).position()
    								
    						),
    						permuteConstraint((Constraint)((Pair)companions.get()).second)
    					)
    					
    				);
    		
    	}
    }

	private void permuteAttributeAssignments() {
		AttributeAssignment assignment;
		ObjectReference ref;
		
    	for (ArrayIterator assigns = _production.attributeAssignments().begin();
    		 !assigns.atEnd();assigns.advance()) {

    		ref = ((AttributeAssignment)assigns.get()).value().copy();
    		ref.position(
    			(
        			(ObjectReference)
        				_inverse_permutation.at(
        					(
        						(ObjectReference)
        							((AttributeAssignment)assigns.get()).value()
        					).position()-1
        				)
       			).position()
       		);
    		assignment = 
    			new AttributeAssignment(
    				((AttributeAssignment)assigns.get()).attributeExpr().copy(),
    				ref
    			); 
        	_attributeAssignments.add(assignment.attribute(),assignment);
        }		
	}

	public SList getSameConstraints(Constraint queryConstraint,int pos) {
		//System.err.println("@@@@pos: "+pos);
		//System.err.println("@@@@queryConstraint: "+queryConstraint);
		//System.err.println("@@@@constraintsAt(pos): "+constraintsAt(pos));
		SList sameConstraints = new SList();
		for (int i = pos;i<=_rhs.size();i++)
			for (SListIterator c = constraintsAt(i).begin();!c.atEnd();
			     c.advance()) {
				Constraint constraint = (Constraint)c.get();
				//System.err.println("@@@@constraint: "+constraint);
				if (constraint.relation().equals(queryConstraint.relation())
				     &&
				    constraint.argumentBefore(i).position() ==
				    queryConstraint.argumentBefore(pos).position())
					sameConstraints.pushBack(constraint);     
			}
		//System.err.println("@@@@sameConstraints: "+sameConstraints);
		return sameConstraints;	
	}

	public BinaryPredicate getOrderingPredicateAt(int pos) {
		return production().getOrderingPredicateAt(
			((ObjectReference)_permutation.at(pos-1)).position()
		);
	}
	
	public Integer getPositionAtOrderingFor(int pos) {
		return production().getPositionAtOrderingFor(
			((ObjectReference)_permutation.at(pos-1)).position()
		);
	}
	
	public String symbolAt(int pos) {
		return (String)_rhs.at(pos-1);
	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		//buf.append("Description: "+_description+"\n");
		buf.append(_production.lhs()+" -> ");
		int counter = 0;
		for (Enumeration arIt = _rhs.elements();arIt.hasMoreElements();) {
			buf.append((String)arIt.nextElement());
			counter++;
			if (isRepetitive(counter) && isOptional(counter))
				buf.append("*");
			else if (isRepetitive(counter) && !isOptional(counter))
				buf.append("+");
			else if (!isRepetitive(counter) && isOptional(counter))
				buf.append("?");
			buf.append(" ");
		}
		/* for (Enumeration arIt = _rhs.elements();arIt.hasMoreElements();)
			buf.append((String)arIt.nextElement()+" "); */
		
		buf.append("permutation: "+_permutation+" ");

		/*
		for (Enumeration arIt = _constraintNetwork.elements();arIt.hasMoreElements();)
			buf.append(arIt.nextElement().toString()+"\n");
		buf.append("\n");

		for (Enumeration arIt = _predicate_network.elements();arIt.hasMoreElements();)
			buf.append(arIt.nextElement().toString()+" ");
		buf.append("\n");
        */
		buf.append("( ");
		for (Enumeration arIt = _attributeAssignments.elements();arIt.hasMoreElements();)
			buf.append(arIt.nextElement().toString()+" ");

		buf.append(")");
		return buf.toString();
	
	}
}