package relap.LanguageModel;

import com.objectspace.jgl.*;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Transforming;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.functions.BindFirst;
import java.util.Enumeration;
import relap.Util.Utils;

public class Relation {

	private String _name;
	private int _arity;
	private Array _indices;
	private Array _tuples;
	private SList failedArgs;
	private SList foundArgs;
	
	protected Relation(String name,int arity) {
		_name = name;
		_arity = arity;
		_indices = new Array();
		_tuples = new Array();
		for (int i = 0;i<_arity;i++)
			_indices.add(new HashMap());
		failedArgs = new SList();
		foundArgs = new SList();
	}
	
	public int arity() { return _arity; }
	
	public String name() { return _name; }
	
	public SList getFailedArgs() { return failedArgs; }
	
	public SList getFoundArgs() { return foundArgs; }

	public String toString() {
		StringBuffer buf = new StringBuffer(_name);
		buf.append(": "+tuplesToString());		
		for (ArrayIterator it = _indices.begin();!it.atEnd();it.advance()) {
			buf.append("\nIndex "+Integer.toString(it.index()).toString());
			for (Enumeration enum = ((HashMap)it.get()).keys();
				 enum.hasMoreElements();) {
				 Object key = enum.nextElement();
				 buf.append("\n"+key.toString()+" -> "
				            +((HashMap)it.get()).get(key).toString());
			}
		}
		
		
		return buf.toString();
	}

	public String tuplesToString() {
		StringBuffer buf = new StringBuffer();
		for (ArrayIterator it = _tuples.begin();!it.atEnd();it.advance())
			buf.append("\n"+it.get().toString());
		return buf.toString();
	}
		
	public Tuple addTuple(Array tuple)
	throws InvalidTupleException {
		Array new_tuple_data = new Array();
		if (tuple.size() != _arity)
			throw(new InvalidTupleException(
				Integer.toString(_arity)+" objects expected in "+
								 tuple.toString()));
		Copying.copy(tuple,new_tuple_data);
		Tuple new_tuple = new Tuple(new_tuple_data,this);
		_tuples.add(new_tuple);
		for (ArrayIterator it = new_tuple_data.begin();!it.atEnd();it.advance()) {
			if (((HashMap)_indices.at(it.index())).get(it.get()) == null)
				((HashMap)_indices.at(it.index())).put(it.get(),new HashSet());	
		
			((HashSet)((HashMap)_indices.at(it.index())).get(it.get())).add(new_tuple);
		}
		return new_tuple;
	}
	
	public Array filterTuples(UnaryPredicate criterion) {
		return (Array)Filtering.select(_tuples,criterion);
	}
	
	private int nullCount(Array arr) {
		int count = 0;
		for (ArrayIterator it = arr.begin();!it.atEnd();it.advance())
			if (it.get() == null)
				count++;
		return count;
	}
			 
	private int nullPos(Array arr) {
		int pos = -1;
		for (ArrayIterator it = arr.begin();!it.atEnd();it.advance())
			if (it.get() == null)
				pos = it.index();
		return pos;
	}
			 
	public HashSet simpleQuery(Array arguments) 
	throws InvalidQueryArgumentsException {
		HashSet result = new HashSet();
		HashSet temp = null;
		if (arguments.size() != _arity)
			throw(new InvalidQueryArgumentsException(
				Integer.toString(_arity)+" arguments expected"));
		if (nullCount(arguments) != 1)
			throw(new InvalidQueryArgumentsException(
				"exactly one null argument expected"));
		for (ArrayIterator it = arguments.begin();!it.atEnd();it.advance()) 
			if (it.get()!=null) {
				temp = 
					(HashSet)((HashMap)_indices.at(it.index())).get(it.get());
				if (temp!=null) {
					foundArgs.pushBack(arguments);
					if (result.isEmpty())
						result = result.union(temp);
					else {
						result = result.intersection(temp);
						if (result.isEmpty()) {
							failedArgs.pushBack(arguments);
							return new HashSet(); 	
						}
					}
				} 
				else {
					failedArgs.pushBack(arguments);
					return new HashSet(); 	
								// current argument does not appear at current pos
								// in any tuple
				}
			}	
					
		return 
			(HashSet)Transforming.
				collect(result,
						new BindFirst(new ObjectAtPos(),
								 	  new Integer(nullPos(arguments))));
	}

	public HashSet query(Array arguments) 
		throws InvalidQueryArgumentsException {
		failedArgs.clear();
		foundArgs.clear();
		//System.err.println(arguments);
		SList queryArgs = Utils.cartesianProduct(arguments);
		HashSet result = new HashSet();
		Enumeration tuple = queryArgs.elements();
		while (tuple.hasMoreElements())
			result = result.union(simpleQuery((Array)tuple.nextElement()));
		return result;
	}
	
	public boolean predicate(Array arguments) 
		throws InvalidQueryArgumentsException {
		failedArgs.clear();
		foundArgs.clear();
		SList queryArgs = Utils.cartesianProduct(arguments);
		boolean result = true;
		Enumeration tuple = queryArgs.elements();
		while (tuple.hasMoreElements())
			result = simplePredicate((Array)tuple.nextElement()) && result;
		return result;
	}
	
	public boolean simplePredicate(Array arguments) 
	throws InvalidQueryArgumentsException {
		HashSet result = new HashSet();
		HashSet temp = null;
		if (arguments.size() != _arity)
			throw(new InvalidQueryArgumentsException(
				Integer.toString(_arity)+" arguments expected"));
		if (nullCount(arguments) == 1)
			return query(arguments).isEmpty() == false;
		if (nullCount(arguments) > 1)
			throw(new InvalidQueryArgumentsException(
				"at most one null argument expected in predicate constraint"));
		// nullCount == 0
		for (ArrayIterator it = arguments.begin();!it.atEnd();it.advance()) {
			temp = (HashSet)((HashMap)_indices.at(it.index())).get(it.get());
			if (temp!=null) {
				foundArgs.pushBack(arguments);
				if (result.isEmpty())
					result = result.union(temp);
				else {
					result = result.intersection(temp);
					if (result.isEmpty()) {
						failedArgs.pushBack(arguments);
						return false; 	
					}
				}
			} 
			else {
				failedArgs.pushBack(arguments); 
				return false; 	// current argument does not appear at current pos
								// in any tuple
			}
		}
			
		return result.isEmpty() == false;
	}
	
	public HashSet getTuples(Array arguments) 
		throws InvalidQueryArgumentsException {
		failedArgs.clear();
		foundArgs.clear();
		SList queryArgs = Utils.cartesianProduct(arguments);
		HashSet result = new HashSet();
		Enumeration tuple = queryArgs.elements();
		while (tuple.hasMoreElements())
			result = result.union(getTuplesForTuple((Array)tuple.nextElement()));
		return result;
	}
	
	public HashSet getTuplesForTuple(Array arguments) 
	throws InvalidQueryArgumentsException {
		HashSet result = new HashSet();
		HashSet temp = null;
		if (arguments.size() != _arity)
			throw(new InvalidQueryArgumentsException(
				Integer.toString(_arity)+" arguments expected"));
		if (nullCount(arguments) > 0)
			throw(new InvalidQueryArgumentsException(
				"no null arguments expected in tuple retrieval"));
		// nullCount == 0
		for (ArrayIterator it = arguments.begin();!it.atEnd();it.advance()) {
			temp = (HashSet)((HashMap)_indices.at(it.index())).get(it.get());
			if (temp!=null) {
				foundArgs.pushBack(arguments);
				if (result.isEmpty())
					result = result.union(temp);
				else {
					result = result.intersection(temp);
					if (result.isEmpty()) {
						failedArgs.pushBack(arguments);
						return new HashSet(); 	
					}
				}
			} 
			else {
				failedArgs.pushBack(arguments);
				return new HashSet(); 	
						// current argument does not appear at current pos
						// in any tuple
			}
		}
		return result;
	}
	
	public HashSet objects() {
		HashSet objs = new HashSet();
		for (ArrayIterator it1=_tuples.begin();!it1.atEnd();it1.advance())
			for (ArrayIterator it2 = ((Tuple)it1.get()).data().begin();
			     !it2.atEnd();it2.advance())
				objs.add(it2.get());
		return objs;
	}
}
