package relap.LanguageModel;

abstract public class RelationalGrammar implements Grammar {

	public void buildGrammar() throws InvalidGrammarException {
		buildSymbolSets();
		buildProductions();
		check();
	}
	

	protected void buildSymbolSets() throws InvalidGrammarException {
		buildTerminals();
		buildNonTerminals();
		buildRelations();
    		buildAttributes();
	}
	abstract protected void buildTerminals() throws InvalidGrammarException; 
	abstract protected void buildNonTerminals() throws InvalidGrammarException; 
	abstract protected void buildRelations() throws InvalidGrammarException; 
	abstract protected void buildAttributes() throws InvalidGrammarException; 
	abstract protected void buildProductions() throws InvalidGrammarException;
	abstract protected void check() throws InvalidGrammarException;

	public String toString() {
		StringBuffer buf = new StringBuffer("Grammar <"+getClass().getName()+">\n");

		buf.append("Terminals:\n"+terminalsToString()+"\n");
		buf.append("Non-Terminals:\n"+nonTerminalsToString()+"\n");
		buf.append("Start symbol:\n"+startSymbolToString()+"\n");
		buf.append("Relations:\n"+relationsToString()+"\n");
		buf.append("Attributes:\n"+attributesToString()+"\n");
		buf.append("Productions:\n"+productionsToString()+"\n");

		return buf.toString();
	}
	abstract protected String attributesToString();
	abstract protected String terminalsToString(); 
	abstract protected String nonTerminalsToString();
	abstract protected String startSymbolToString();
	abstract protected String relationsToString();
	abstract protected String productionsToString();
}
