package relap.LanguageModel;

import java.util.*;
import java.lang.Class;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.HashMap;
import com.objectspace.jgl.Array;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.Container;
import com.objectspace.jgl.BinaryPredicate;
import com.objectspace.jgl.predicates.BindFirstPredicate;
import com.objectspace.jgl.algorithms.Copying;
import com.objectspace.jgl.algorithms.Filtering;
import com.objectspace.jgl.adapters.ObjectArray;
import relap.Util.*;
import relap.Analyzer.SyntaxTreeNode;

abstract public class RelationalGrammarImplementation extends RelationalGrammar {

	private HashSet _symbols;
	private HashSet _terminals;
	private HashSet _nonTerminals;
	private String _startSymbol;
	private HashSet _expanderAttributes;
	private String _start_attribute;
	private HashSet _relations;
	private Array _productions;
	private Class _productionFactory;
	private HashMap _terminal_to_interface;
	private HashMap _attribute_map;
	private Object fPredicateEvaluator;
	
	public RelationalGrammarImplementation() {
		_symbols = new HashSet();
		_terminals = new HashSet();
		_nonTerminals = new HashSet();
		_startSymbol = null;
		_start_attribute = "@";
		_expanderAttributes = new HashSet();
		_relations = new HashSet();
		_productions = new Array();
		_productionFactory = null;
		_symbols.add(_start_attribute);
		_terminal_to_interface = new HashMap();
	}
	
	public void setPredicateEvaluator(Object object) {
		fPredicateEvaluator = object;
	}
	
	public Object getPredicateEvaluator() {
		return fPredicateEvaluator;
	}
	
	public HashSet nonTerminals() { return _nonTerminals; }
	
	public HashSet terminals() { return _terminals; }
	
	public HashSet relations() { return _relations; }
	
	public HashSet expanderAttributes() { return _expanderAttributes; }

	public String startAttribute() { return _start_attribute; }

	public String startSymbol() { return _startSymbol; }

	public Array productions() { return _productions; }
	
	protected void attributeMap(HashMap map) {
		_attribute_map = map;
	}
	
	public HashMap attributeMap() {
		return _attribute_map;
	}
	
	// hook
	protected void setAttributeMap() {
		attributeMap(new HashMap());
	}
	
	// hook
	public HashSet fixedStartAttributes() {
		return expanderAttributes();
	}
	
	public boolean isTerminal(String symbol) { 
		return _terminals.get(symbol) != null;
	}
	
	public boolean isNonTerminal(String symbol) { 
		return _nonTerminals.get(symbol) != null;
	}
	
	public boolean isRelation(String symbol) { 
		return _relations.get(symbol) != null;
	}
	
	public boolean isAttribute(String symbol) { 
		return _expanderAttributes.get(symbol) != null;
	}
	
	class ExpandsNT implements BinaryPredicate {
		public boolean execute(Object nt,Object prod) {
			return ((String)nt).equals(((GrammarProduction)prod).lhs());
		}
	}
	
	public Container productionsExpanding(String nt) {
		return Filtering.select(_productions,
								new BindFirstPredicate(new ExpandsNT(),nt));
	}
	
	protected void buildTerminals() throws InvalidGrammarException {
		parseTerminals(terminalDeclarations());
	} 
	protected void buildNonTerminals() throws InvalidGrammarException {
		parseNonTerminals(nonTerminalDeclarations());
		parseStartSymbol(startSymbolDeclaration());
	} 
 
	protected void buildRelations() throws InvalidGrammarException {
		parseRelations(relationDeclarations());
	} 
 
	protected void buildAttributes() throws InvalidGrammarException {
		parseAttributes(attributeDeclarations());
		setAttributeMap();
	} 
 
	protected void buildProductions() throws InvalidGrammarException {
		SList allMethods = new SList();
		Copying.copy(new ObjectArray(getClass().getMethods()),allMethods);
		for (Enumeration nts = _nonTerminals.elements();nts.hasMoreElements();) {
			String nonTerminal = (String)nts.nextElement();
			int p_count = 0;
			Container methods =
				Filtering.select(allMethods,
						 		 new BindFirstPredicate(new IsPrefixOfMethodName(),
						 								nonTerminal+"_"));
			for (Enumeration mIt = methods.elements();
				 mIt.hasMoreElements();) {
				Method method = (Method)mIt.nextElement();
				Object args[] = new Object[1]; 
				args[0] = new GrammarProduction(this,nonTerminal);	
				try {
					checkProductionSignature(method);
					if (method.invoke(this,args) != null) {	
						_productions.add(args[0]);
						p_count++;
					}
				}
				catch (IllegalAccessException e) {
					throw(new InvalidGrammarException(e.getMessage()));
				}
				catch (InvocationTargetException e) {
					e.getTargetException().printStackTrace(System.err);
					throw((InvalidProductionException)e.getTargetException());
				}
			}
			if (p_count == 0)
				throw(new InvalidGrammarException("No production defined for non-terminal \""+
												   nonTerminal+"\""));
		}
	} 

	private void checkProductionSignature(Method m) 
	throws InvalidGrammarException {
		try {
			if (m.getReturnType() != Class.forName("relap.LanguageModel.GrammarProduction"))
				throw(new InvalidReturnTypeException(m));
		}
		catch (ClassNotFoundException e) {
			System.err.println("Class "+e.getMessage()+" not found");
			throw(new InvalidGrammarException("System error"));
		} 

		Class pars[] = m.getParameterTypes();
		if (pars.length != 1)	
			throw(new InvalidParameterListException(m));
		try {
			if (pars[0] != Class.forName("relap.LanguageModel.GrammarProduction"))
				throw(new InvalidParameterListException(m));
		}
		catch (ClassNotFoundException e) {
			System.err.println("Class "+e.getMessage()+" not found");
			throw(new InvalidGrammarException("System error"));
		} 

		Class excs[] = m.getExceptionTypes();
		if (excs.length != 1)	
			throw(new InvalidExceptionListException(m));
		try {
			if (excs[0] != Class.forName("relap.LanguageModel.InvalidGrammarException"))
				throw(new InvalidExceptionListException(m));
		}
		catch (ClassNotFoundException e) {
			System.err.println("Class "+e.getMessage()+" not found");
			throw(new InvalidGrammarException("System error"));
		} 
		
			
	}

	protected void check() throws InvalidGrammarException{
		for (Enumeration pIt = _productions.elements();pIt.hasMoreElements();)
				((GrammarProduction)pIt.nextElement()).check(); 
	} 

	protected String attributesToString() {
		return attributeDeclarations();
	}

	protected String terminalsToString() {
		return terminalDeclarations();
	}
 
	protected String nonTerminalsToString() {
		return nonTerminalDeclarations();
	}

	protected String startSymbolToString() {
		return startSymbolDeclaration();
	}

	protected String relationsToString() {
		return relationDeclarations();
	}

	protected String productionsToString() {
		return Utils.contentsToString(_productions,"\n");
	}
	
	private void parseTerminals(String str) throws InvalidGrammarException {
		if (str.length() == 0)
			throw(new InvalidGrammarException("No terminals defined"));
		else {
			StringTokenizer tknzr = new StringTokenizer(str);
			String token;
			while (tknzr.hasMoreTokens()) {
				token = (String)tknzr.nextToken();
				if (_symbols.add(token) != null)
					throw(new DuplicateDefinitionException(token,"terminal"));				
/*
				try {
					if (!Class.forName(token).isInterface())
						throw(new InvalidGrammarException("Class "+token+" should be interface"));
					_terminal_to_interface.add(token,Class.forName(token));	
				}
				catch (ClassNotFoundException e) {
					throw(new InvalidGrammarException("Interface "+e.getMessage()+" not found"));
				} 
*/
				_terminals.add(token);
			} 
		} 
	}
	
	public Class classForTerminal(String terminal) {
		return (Class)_terminal_to_interface.get(terminal);
	}
	
	private void parseNonTerminals(String str) throws InvalidGrammarException {
		if (str.length() == 0)
			throw(new InvalidGrammarException("No non-terminals defined"));
		else {
			StringTokenizer tknzr = new StringTokenizer(str);
			String token;
			while (tknzr.hasMoreTokens()) {
				token = (String)tknzr.nextToken();
				if (_symbols.add(token) != null)
					throw(new DuplicateDefinitionException(token,"non-terminal"));
				_nonTerminals.add(token);
			} 
		} 
	}
	
	private void parseStartSymbol(String str) throws InvalidGrammarException {
		if (str.length() == 0)
			throw(new InvalidGrammarException("No start symbol defined"));
		else {
			StringTokenizer tknzr = new StringTokenizer(str);
			String token;
			if (tknzr.countTokens() > 1)
				throw(new InvalidGrammarException("More than one start symbol defined"));
			else {
				token = (String)tknzr.nextToken();
				if (_nonTerminals.get(token) == null)
					throw(new UndefinedSymbolException(token,"non-terminal"));
				_startSymbol = token;
			} 
		} 
	}
	
	private void parseRelations(String str) throws InvalidGrammarException {
		if (str.length() == 0)
			throw(new InvalidGrammarException("No relations defined"));
		else {
			StringTokenizer tknzr = new StringTokenizer(str);
			String token;
			while (tknzr.hasMoreTokens()) {
				token = (String)tknzr.nextToken();
				if (_symbols.add(token) != null)
					throw(new DuplicateDefinitionException(token,"relation"));
				_relations.add(token);
			} 
		} 
	}
	private void parseAttributes(String str) throws InvalidGrammarException {
		if (str.length() == 0)
			throw(new InvalidGrammarException("No attributes defined"));
		else {
			StringTokenizer tknzr = new StringTokenizer(str);
			String token;
			while (tknzr.hasMoreTokens()) {
				token = (String)tknzr.nextToken();
				if (_symbols.add(token) != null)
					throw(new DuplicateDefinitionException(token,"attribute"));
				_expanderAttributes.add(token);
			} 
		} 
	}
	abstract public String terminalDeclarations();
	abstract public String nonTerminalDeclarations();
	abstract public String startSymbolDeclaration();
	abstract public String attributeDeclarations();
	abstract public String relationDeclarations();

	public SyntaxTreeNode getDefaultLeafSemantics() {
		return new relap.Analyzer.LeafNode();
	}
}
