package relap.Util;

import java.io.*;
import com.objectspace.jgl.*;
import relap.Analyzer.IndexedMDSet;
import relap.LanguageModel.ExpressionSyntaxException;
import relap.LanguageModel.ExpanderRelation;
import relap.LanguageModel.InvalidTupleException;

public class InputReader {
	private HashMap _relations;
	private IndexedMDSet _input;
	private FileReader _file_reader;
	private String _file_name;
	
	public InputReader() {
	}

	private void parseRelation(StreamTokenizer tokens)
	throws IOException,ExpressionSyntaxException,
		   InvalidTupleException {
		ExpanderRelation rel;
		if (tokens.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("relation name",
			                                     tokens.toString(),
			                                     _file_name));
		rel = new ExpanderRelation(tokens.sval);
		_relations.put(tokens.sval,rel);
		tokens.nextToken();
		if (tokens.ttype != '{')
			throw(new ExpressionSyntaxException("{",
			                                     tokens.toString(),
			                                     _file_name));
		while (tokens.nextToken() != '}')
			rel.addTuple(parseTuple(tokens));

		_input.addRelation(rel);
	}
	
	private Array parseTuple(StreamTokenizer tokens)
	throws IOException,ExpressionSyntaxException {
		String first,first_type,second,second_type;
		if (tokens.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("object name",
			                                     tokens.toString(),
			                                     _file_name));
		first = tokens.sval;
		tokens.nextToken();
		if (tokens.ttype != ':')
			throw(new ExpressionSyntaxException(":",
			                                     tokens.toString(),
			                                     _file_name));
		tokens.nextToken();
		if (tokens.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("object type",
			                                     tokens.toString(),
			                                     _file_name));
		first_type = tokens.sval;
		tokens.nextToken();
		if (tokens.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("object name",
			                                     tokens.toString(),
			                                     _file_name));
		second = tokens.sval;
		tokens.nextToken();
		if (tokens.ttype != ':')
			throw(new ExpressionSyntaxException(":",
			                                     tokens.toString(),
			                                     _file_name));
		tokens.nextToken();
		if (tokens.ttype != StreamTokenizer.TT_WORD)
			throw(new ExpressionSyntaxException("object type",
			                                     tokens.toString(),
			                                     _file_name));
		second_type = tokens.sval;
		_input.setTerminalType(first,first_type);
		_input.setTerminalType(second,second_type);
		Array result = new Array(2);
		result.put(0,first);
		result.put(1,second);
		return result;
	}
	
	private void parseUnrelatedObjects(StreamTokenizer tokens)
	throws IOException,ExpressionSyntaxException {
		String object,type;

		while (tokens.nextToken() == '@') {
			tokens.nextToken();
			if (tokens.ttype != StreamTokenizer.TT_WORD)
				throw(new ExpressionSyntaxException("object name",
			                                     	tokens.toString(),
			        	                            _file_name));
			object = tokens.sval;
			tokens.nextToken();
			if (tokens.ttype != ':')
				throw(new ExpressionSyntaxException(":",
			                                     	tokens.toString(),
			                                     	_file_name));
			tokens.nextToken();	
			if (tokens.ttype != StreamTokenizer.TT_WORD)
				throw(new ExpressionSyntaxException("object type",
			                                     	tokens.toString(),
			                                     	_file_name));
			type = tokens.sval;	
			_input.addObject(object);
			_input.setTerminalType(object,type);
		}
		tokens.pushBack();
	}

	public IndexedMDSet readInput(String file)
	throws IOException,ExpressionSyntaxException,InvalidTupleException {
		_relations = new HashMap();
		_input = new IndexedMDSet();
		_file_name = file;
		_file_reader = new FileReader(file);
		StreamTokenizer tokens = new StreamTokenizer(_file_reader);
		tokens.ordinaryChar('{');
		tokens.ordinaryChar('}');
		tokens.ordinaryChar(':');
		parseUnrelatedObjects(tokens);
		while (tokens.nextToken() != StreamTokenizer.TT_EOF)
			parseRelation(tokens);
		return _input;
	}
}
